;; this file contains all template definitions 
;; (including class definitions for shadow facts)
;; plus some global initialisation stuff like
;; declaration of modules for the rules and
;; loading user functions

(provide deftemplates)

;; --- modules --------------------------------------

(defmodule INIT)
(defmodule CONTENT)
(defmodule FINAL)
(defmodule CLEANUP)

(set-current-module MAIN)
(set-strategy breadth)


;; --- class definitions -------------------------------

(import org.cfpm.caves.za.*)

(defclass person Person)
(defclass household Household)
(defclass model SouthAfricaModel)
(defclass node NetworkNode)
(defclass link NetworkLink)
(defclass church Church)
(defclass burialSociety BurialSociety)
(defclass stokvel Stokvel)
(defclass endorsementScheme EndorsementScheme)
(defclass mine Mine)


;; --- global variables --------------------------------

;; most of these are defined via the RulebaseCreator
;; so that they can be set as model parameters

;; ?*standard-wage*
;; ?*shack-proportion*
;; ?*burial-cost*
;; ?*random-partner-probability*
;; ?*similarity-age-range*
;; ?*min-remittance*
;; ?*max-remittance*
;; ?*remittance-probability*
;; ?*school-fee*
;; ?*college-fee*
;; ?*min-lobola*
;; ?*max-lobola*
;; ?*house-cost*
;; ?*min-shack-rent*
;; ?*max-shack-rent*

(defglobal ?*min-travel-costs* = 100)  ;; how much are travel costs for migrants??
(defglobal ?*max-travel-costs* = 300)


;; --- user functions ----------------------------------

(load-function EndorsementValue)
(load-function CommonAttributeCounter)
(load-function RandomListElement)
(load-function SortValues)
(load-function ElementWithHighestOccurrence)
(load-function MessageToFact)


;; --- fact templates --------------------------------

;; household economy related

(deftemplate help-needed
    "reminds household that they need to ask for help (food) in the next tick"
    (slot household)
    (slot amount)
    (slot tickStamp (type integer))
)

(deftemplate asked-for-help
    "serves as memory for a household so they know whom they have asked for help"
    (slot requester)
    (multislot hh-list)
    (slot tickStamp (type integer))
)

(deftemplate payment
    "record payments made"
    (slot payer)
    (slot recipient)
    (slot cause)
    (slot amount (type float))
    (slot payment-date (type integer)))

(deftemplate rent-a-shack
    "records a contract for renting a shack"
    (slot landlord)
    (slot tenant)
    (slot rent (type float))
    (slot start-date)
    (slot end-date))

;; endorsement related

(deftemplate endorsed
    "records an endorsement of one agent (clauseOwner) to another agent (endorsedObject)"
    (slot endorsement)
    (slot endorsed-object)
    (multislot uniquifiers)
    (slot clauseOwner)
    (slot tickStamp (type integer)))

(deftemplate object-endorsement-value
	"used in ranking agents by endorsements"
    (slot object-type)
	(slot object)
    (slot endorsement-value)
    (slot clauseOwner)
    (slot tickStamp (type integer)))

(deftemplate similarity-index
	"number of positions in tag with identical values"
	(slot similarity (type integer))
    (slot other-person)
    (slot clauseOwner)
    (slot tickStamp (type integer)))

(deftemplate known-similarities
	"corresponding lists of known others and their similarity indices"
	(multislot other-persons)
    (multislot similarity-indices)
    (slot clauseOwner)
    (slot tickStamp (type integer)))


;; group related

(deftemplate group
	"group could be a stokvel, a church, a burial society, etc."
	(slot group-type)
    (slot group-id)
    (multislot members))

(deftemplate group-wanted
    "a reminder that the person wants a group of the specified type"
    (slot type)
    (slot person)
	(slot tickStamp)
)

(deftemplate group-proposal
    "if an agent wants to join a group of a particular type which doesn't exist yet he proposes to found one to his friends "
    (slot type)
	(slot proposed-by)
	(slot proposed-to)
	(slot tickStamp)
)

;; agent communication related
    
(deftemplate message
	"components of message are components of facts"
	(slot sender)
    (slot recipient)
    (slot header)
    (multislot slot-names)
    (multislot slot-values)
    (slot tickStamp (type integer)))


;; employment related
    
(deftemplate employment
	"an artifact to provide non-grant money to households"
	(slot employee)
    (slot employer)
    (slot wage (type float))
    (slot start-date)
	(slot end-date))

(deftemplate job-performance
    "to record how the employer assesses the performance of an employee"
	(slot employee)
	(slot employer)
	(slot reliability)	;; number of times the employee was reliable in a job
	(slot capability)	;; number of times the employee was capable to fulfill the job
	(slot num-jobs)		;; number of jobs
	(slot tickStamp)
)

(deftemplate employees-demanded
	"number demanded by a household or mine"
	(slot employer) ;; a household or mine
    (slot number-demanded)
    (slot tickStamp))

(deftemplate job-application
    "application for an advertised job"
	(slot applicant)
	(slot potential-employer)
	(slot tickStamp)
)


;; death related

(deftemplate death
	"records the death of a person"
	(slot name)
    (slot household)
    (slot tickStamp))

(deftemplate death-in-hh
    "reminds hh to pay for burial in next tick"
	(slot household)
    (slot deceased)
    (slot tickStamp)
)

(deftemplate membership-reminder
    "reminds hh which group the deceased was a member of"
	(slot household)
    (slot deceased)
	(slot group)
	(slot type)
    (slot tickStamp)
)

;; birth related 

(deftemplate birth
	"records the birth of a person"
	(slot name)
    (slot household)
    (slot tickStamp))

;; marriage related 

(deftemplate marriage
    "records a marriage"
	(slot groom)
	(slot groom-household)
	(slot bride)
    (slot bride-household)
	(slot tickStamp)
)

(deftemplate upcoming-household
    "reminds person that they want to form their own household (once there's enough money)"
	(slot person)
	(slot tickStamp)
)

;; mining related

(deftemplate leave
    "records the fact that a miner left the area"
    (slot miner)
    (slot tickStamp))
