package org.cfpm.caves.za;

import java.util.ArrayList;
import java.util.HashMap;

import uchicago.src.sim.util.SimUtilities;
//import uchicago.src.sim.network.DefaultEdge;
import java.beans.*;

import jess.JessException;
import jess.RU;
import jess.Rete;
import jess.Value;
import jess.WatchConstants;

/** A burial society is a community insurance that pays out money when a relative of a member dies.
 *  Monthly fees have to be paid in order to get anything. If someone can't pay there's a grace period
 *  of 3-6 months. If they pull out of the society they don't get any money (i.e. it's NOT a savings club!)
 *  Members usually try to stick with the society as long as possible.
 *    
 * @author Ruth
 *
 */
public class BurialSociety extends SavingsClub {
	
	//--- burial society attributes --------------------------------

	/** the standard sum paid out for a burial. Default is 2000 Rand. */
	double payoutSum = 2000; 	// default: 2000 Rand
	
	//--- constructors ----------------------------------------------
	
	/** Creates a new burial society for the given members with the given holdings.
	 * @param holdings the holdings
	 * @param members the list of members
	 * @param model the model
	 */
	public BurialSociety(double holdings, ArrayList<Person> members, SouthAfricaModel model){
		super("burialSociety", holdings, members, model);
		// set monthly fee (draw random sample from uniform distribution (20-60 Rand)
		this.monthlyFee = model.randomInt(20,60);
		// put it into Jess (needs to be here as Jess can't handle inheritance with shadow facts)
		try {
			model.engine.definstance("burialSociety", this, true);
		} catch (JessException je) {
			System.err.println("An error occurred at line "	
					+ je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}
	}

	//--- methods -------------------------------------------------------
	
	/** lets the given person join this burial society
	 * @param member the new member
	 */
	public void addMember(Person member){
		super.addMember(member);
		// add burial society to member's societies
		member.joinBurialSociety(this);
	}
		
	/** lets the given person leave this burial society
	 * @param member the leaving member
	 */
	public void removeMember(Person member){
		super.removeMember(member);
		// remove burial society from member's societies
		member.leaveBurialSociety(this);
	}

	/** pay out to the requesting member for the burial of the deceased person
	 * 
	 * @param requestingMember the member requesting a payout
	 * @param deceased the person whose burial is covered by this burial society
	 * @return amount paid out (<= standard payout)
	 */
	public double payOut(Value requestingMember, Value deceased){
		// pay out agreed sum to requesting member 
		if (this.members.contains(model.namePersonMap.get(requestingMember))) {
			// TODO: pay more if requesting member = deceased?
			double old = this.holdings;
			double payout = Math.min(payoutSum, this.holdings);
			this.holdings -= payout;
			pcs.firePropertyChange("holdings", old, this.holdings);
			return payout;
		}
		return 0.0;
	}
	
	/** pay out to the requesting member for the burial of the deceased person
	 * 
	 * @param requestingMemberName name of the member requesting a payout
	 * @param deceasedName name of the person whose burial is covered by this burial society
	 * @return amount paid out (<= standard payout)
	 */
	public double payOut(String requestingMemberName, String deceasedName){
		try {
			return payOut(new Value(requestingMemberName, RU.SYMBOL), new Value(deceasedName, RU.SYMBOL));
		}
		catch (JessException je){
			System.err.println("An error occurred at line " + je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}
		return 0.0;
	}
	
	/** lets the given member pay in their monthly fee (or more in case they have to pay off
	 * a grace period)
	 * @param member the member paying in
	 * @param amount the amount paid in
	 */
	public void payIn(Person member, double amount){
		double old = this.holdings;
		this.holdings += amount;
		pcs.firePropertyChange("holdings", old, this.holdings);
		// check if member was on grace period
		if (onGrace.containsKey(member)){
			// assume member paid all outstanding debts
			// TODO: keep track of outstanding amounts so that members can pay off debt over several months
			Value[] oldMembers = getMembersOnGrace();
			int[] oldPeriods = getGracePeriods();
			onGrace.remove(member);
			pcs.firePropertyChange("membersOnGrace", oldMembers, getMembersOnGrace());
			pcs.firePropertyChange("gracePeriods", oldPeriods, getGracePeriods());
		}
	}
	

}
