package org.cfpm.caves.za;

import jess.JessException;
import jess.Value;
import jess.RU;
import jess.WatchConstants;

import java.beans.*;
import java.util.ArrayList;

/**
 * This class represents a church of a particular denomination in a village. We
 * assume that all members of a household are members of the same church, i.e.
 * membership is handled on the household level.
 * Used as shadow fact in Jess.
 * 
 * @author scott, ruth
 * 
 */
public class Church {

	// --- church attributes --------------------------------

	/** the list of households who are members of this church */
	ArrayList<Household> members = new ArrayList<Household>();

	/** list of member household names (for easier access in Jess) */
	public Value[] memberIDs;

	/** this church's name */
	public Value name;

	/** this church's denomination */
	public Value denomination;

	private PropertyChangeSupport pcs = new PropertyChangeSupport(this);

	// --- constructors ----------------------------------------------

	/**
	 * Creates a new church of the given name and denomination
	 * 
	 * @param nameString
	 *            the name
	 * @param denominationString
	 *            the denomination
	 * @param model
	 *            the model
	 */
	public Church(String nameString, String denominationString,
			SouthAfricaModel model) {
		try {
			name = new Value(nameString, RU.SYMBOL);
			denomination = new Value(denominationString, RU.SYMBOL);
			model.engine.definstance("church", this, true);
			model.dump("Created new church: " + nameString
					+ " of denomination " + denominationString);
		} catch (JessException je) {
			System.err.println("An error occurred at line "
					+ je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}
		memberIDs = new Value[0];
	}

	// --- methods -------------------------------------------------------

	/**
	 * Add the household to the <code>ArrayList members<\code> and
	 * replace the memberIDs array of <code>Value<\code> with the
	 * same array extended to include the id of the added household.
	 * @param h the new member household
	 */
	public void addHousehold(Household h) {
		members.add(h);
		Value[] idList = new Value[members.size()];
		for (int i = 0; i < members.size(); i++)
			idList[i] = members.get(i).id;
		setMemberIDs(idList);
		// set church slot in household
		h.setChurch(this.name);
	}

	/** removes the ghiven household from the list of members
	 * 
	 * @param h the household leaving the church
	 */
	public void removeHousehold(Household h) {
		members.remove(h);
		Value id = h.id;
		ArrayList<Value> idList = new ArrayList<Value>(members.size());
		for (Household hh : members) {
			if (!hh.id.equals(id))
				idList.add(h.id);
		}
		setMemberIDs(idList);
		// remove church from church slot in household
		h.setChurch(new Value(RU.NONE));
	}

	/** returns the member IDs
	 * 
	 * @return list of household IDs
	 */
	public Value[] getMemberIDs() {
		return memberIDs;
	}

	/** sets the member IDs to the given list of IDs
	 * 
	 * @param memberIDs list of household IDs as an Array of Value
	 */
	public void setMemberIDs(Value[] memberIDs) {
		Value[] old = this.memberIDs;
		this.memberIDs = memberIDs;
		pcs.firePropertyChange("memberIDs", old, memberIDs);
	}

	/** sets the member IDs to the given list of IDs
	 * 
	 * @param memberIDsList list of household IDs as an ArrayList of Value
	 */
	public void setMemberIDs(ArrayList<Value> memberIDsList) {
		Value[] old = this.memberIDs;
		this.memberIDs = memberIDsList.toArray(memberIDs);
		pcs.firePropertyChange("memberIDs", old, memberIDs);
	}

	/** returns the name of this church
	 * 
	 * @return the name
	 */
	public Value getName() {
		return name;
	}

	/** returns the denomination of this church
	 * 
	 * @return the denomination
	 */
	public Value getDenomination() {
		return denomination;
	}

	/** adds a property change listener (for Jess) */	
	public void addPropertyChangeListener(PropertyChangeListener p) {
		pcs.addPropertyChangeListener(p);
	}

	/** removes a property change listener (for Jess) */
	public void removePropertyChangeListener(PropertyChangeListener p) {
		pcs.removePropertyChangeListener(p);
	}

}
