package org.cfpm.caves.za;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;

import jess.JessException;
import jess.Rete;
import jess.Context;
import jess.Fact;
import jess.Value;

/** This class provides functionality to remove facts from the Jess fact base.
 * It keeps track of of which type of fact can be removed after how many ticks. 
 * To be able to be removed with a disposal policy, a fact needs to possess a 
 * slot with a time stamp, denoting its assertion time. The modeller can then 
 * specify when to remove such a fact by calling the <code>addPolicy()<code> method with 
 * the fact header, name of the time stamped slot and the wanted time lag as 
 * parameters. If no time stamped slot is specified, the default <code>timeStamp<code>
 * is assumed. 
 * The disposal policy is executed at the end of every model step after 
 * running the Jess engine, ensuring that a time lag of 0 results in immediate removal.
 * 
 * @author Scott, Ruth
 *
 */
public class DisposalPolicy {
	
	//--- attributes --------------------------------------------

	/** The Jess engine */
	Rete engine;
	/** The context of the Jess engine */
	Context context;
	
	/** the list of policies (key = fact header, value = time lag) */
	HashMap<String, Integer> policy;
	/** the list of slots containing the time stamp information
	 * (key = fact header, value = time stamp slot)
	 */
	HashMap<String, String> timedSlots;  // default is "tickStamp" but other slots may be specified
	
	
	//--- constructors -------------------------------------------

	/** Creates a new empty disposal policy class.
	 * @param engine the Jess engine
	 */
	public DisposalPolicy(Rete engine){
		this.engine = engine;
		context = engine.getGlobalContext();
		policy = new HashMap<String, Integer>();
		timedSlots = new HashMap<String, String>();		
	}
	
	/** Creates a new disposal policy class with the given list of policies.
	 * 
	 * @param policy list of policies (all facts are assumed to have the default "tickStamp" time slot)
	 * @param engine the Jess engine
	 */
	public DisposalPolicy(HashMap<String, Integer> policy, Rete engine){
		this.engine = engine;
		context = engine.getGlobalContext();
		this.policy = policy;
		timedSlots = new HashMap<String, String>();
		Iterator facts = policy.keySet().iterator();
		while (facts.hasNext()){
			String fact = (String)facts.next();
			timedSlots.put(fact, "tickStamp");
		}
	}
	
	
	//--- methods ----------------------------------------------
	
	/** adds a policy for a fact with default tickStamp slot
	 * @param factHeader the fact header
	 * @param lag the time lag (number of ticks until the fact is to be removed)
	 */
	public void addPolicy(String factHeader, int lag){
		policy.put(factHeader, lag);
		timedSlots.put(factHeader, "tickStamp");
	}
	
	/** adds a policy for a fact with a specified timed slot
	 * @param factHeader the fact header
	 * @param slotHeader name of the time stamp slot
	 * @param lag the time lag (number of ticks until the fact is to be removed)
	 */
	public void addPolicy(String factHeader, String slotHeader, int lag){
		policy.put(factHeader, lag);
		timedSlots.put(factHeader, slotHeader);
	}
	
	/** removes the policy for the specified fact type
	 * 
	 * @param factHeader the fact header
	 */
	public void removePolicy(String factHeader){
		policy.remove(factHeader);
		timedSlots.remove(factHeader);
	}
	
	/** Executes the policies at the given time step. To do so, this method
	 * retrieves the list of facts from Jess and runs through it once, determining
	 * every fact that fits a disposal policy and thus has to be removed. 
	 * The determination of fitting facts is realised in the method <code>
	 * determineDisposalList()<code>, which may be overwritten by sub-classes
	 * (Strategy design pattern)
	 * @param tick - the current time step
	 */
	public void executePolicy(int tick){
		ArrayList<Fact> disposalList = new ArrayList<Fact>();
		Iterator factList = engine.listFacts();
		System.out.println("determining disposal list");
		determineDisposalList(factList, disposalList, tick);  // this will be overwritten by sub-class
		for (Fact f : disposalList){
			try {
//				System.out.println("retracting fact " + f.toStringWithParens());
				engine.retract(f);
			} catch (JessException je) {
				System.err.println("An error occurred at line " + je.getLineNumber());
				System.err.println("caused by " + je.getCause());
				System.err.println("Message: " + je.getDetail());
				je.printStackTrace();
			}
		}
		System.out.println("fact removal done");
	}
	
	/** The strategy applied to determine which facts fit the defined policies and should
	 * therefore be removed. Collects all fitting facts in a disposal list.
	 * @param factList the list of all facts to be checked
	 * @param disposalList the list of facts to be removed from the fact base
	 * @param tick the current tick
	 */
	protected void determineDisposalList(Iterator factList, ArrayList<Fact> disposalList, int tick){
		while (factList.hasNext()){
			Fact f = (Fact)factList.next();
			if (policy.containsKey(f.getName())){
				try {
					if (!f.getSlotValue(timedSlots.get(f.getName())).equals("nil")){
						int lag = tick - f.getSlotValue(timedSlots.get(f.getName())).intValue(context);
						if (lag >= policy.get(f.getName())){
							disposalList.add(f);
						}
					}
				} catch (JessException je) {
					System.err.println("An error occurred at line " + je.getLineNumber());
					System.err.println("caused by " + je.getCause());
					System.err.println("Message: " + je.getDetail());
					je.printStackTrace();
				}
			}
		}		
	}
	
	/** replaces the current list of policies with the given list of policies.
	 *  All facts are assumed to have the default <code>tickStamp</code> slot.
	 * @param policy new list of policies
	 */
	public void setPolicy(HashMap<String, Integer> policy) {
		this.policy = policy;
		timedSlots = new HashMap<String, String>();
		Iterator facts = policy.keySet().iterator();
		while (facts.hasNext()){
			String fact = (String)facts.next();
			timedSlots.put(fact, "tickStamp");
		}
	}
	
	/** replaces the current list of policies with the given list of policies.
	 * @param policy new list of policies
	 * @param timedSlots new list of time stamp slots
	 */
	public void setPolicy(HashMap<String, Integer> policy, HashMap<String, String> timedSlots){
		this.policy = policy;
		this.timedSlots = timedSlots;
	}
}
