package org.cfpm.caves.za;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;

import jess.Fact;
import jess.JessException;
import jess.Rete;

/** This class is a specialisation of <code>DisposalPolicy</code>, able
 * to include endorsement facts. This is necessary because the fact header
 * of endorsement facts (<code>endorsed</code>) is the same for all endorsements
 * and the differentiation of removal or not has to be made based on the
 * actual endorsement (slot <code>endorsement</code>). All endorsement facts
 * have the default time stamp slot <code>tickStamp</code>.
 * 
 * @author Ruth
 *
 */
public class EndorsementDisposalPolicy extends DisposalPolicy {
	
	//--- attributes --------------------------------------------

	/** the list of endorsement filters
	 * (key = endorsement token, value = time lag)
	 */
	HashMap<String, Integer> endorsementFilter = new HashMap<String, Integer>();

	
	//--- constructors -------------------------------------------

	/** Creates a new empty endorsement disposal policy
	 * 
	 * @param engine the Jess engine
	 */
	public EndorsementDisposalPolicy(Rete engine) {
		super(engine);
	}
	
	
	//--- methods ----------------------------------------------
	
	/** adds a new endorsement policy
	 * @param endorsementToken the endorsement token
	 * @param lag the time lag
	 */
	public void addEndorsementPolicy(String endorsementToken, int lag){
		endorsementFilter.put(endorsementToken, lag);
	}
	
	/** The strategy applied to determine which facts fit the defined policies and should
	 * therefore be removed. Collects all fitting facts in a disposal list.
	 * @param factList the list of all facts to be checked
	 * @param disposalList the list of facts to be removed from the fact base
	 * @param tick the current tick
	 */
	protected void determineDisposalList(Iterator factList, ArrayList<Fact> disposalList, int tick){
		while (factList.hasNext()){
			try {
				Fact f = (Fact)factList.next();
				if (f.getName().endsWith("endorsed")){
					// it's an endorsement --> see if there's a matching filter defined
					String endorsement = f.getSlotValue("endorsement").stringValue(context);
					if (endorsementFilter.containsKey(endorsement)){
						int lag = tick - f.getSlotValue("tickStamp").intValue(context);
						if (lag >= endorsementFilter.get(endorsement).intValue()){
							disposalList.add(f);
						}
					}
				}
				// not an endorsement --> see if there's a "normal" policy
				else if (policy.containsKey(f.getName())){
					if (!f.getSlotValue(timedSlots.get(f.getName())).equals("nil")){
						int lag = tick - f.getSlotValue(timedSlots.get(f.getName())).intValue(context);
						if (lag >= policy.get(f.getName())){
							disposalList.add(f);
						}
					}
				}
			}
			catch (JessException je) {
				System.err.println("An error occurred at line " + je.getLineNumber());
				System.err.println("caused by " + je.getCause());
				System.err.println("Message: " + je.getDetail());
				je.printStackTrace();
			}
		}
	}
	


}
