package org.cfpm.caves.za;

import java.util.HashMap;
import java.util.ArrayList;

import jess.JessException;
import jess.Value;
import jess.ValueVector;
import jess.RU;

/** This class represents an endorsement scheme of an agent. This is
 * basically a table of endorsement tokens ("labels") and their associated
 * weights ("values"). It allows the agent to evaluate another agent by computing
 * the overall endorsement value of all endorsements he has made 
 * to that agent.
 * 
 * @author Scott Moss, Ruth Meyer
 */
public class EndorsementScheme {

	//--- attributes --------------------------------------------

	/** the name of the endorsement scheme */
	Value name;
	/** the base of the endorsement scheme, used in the calculation of the overall endorsement value */
	double base;
	/** the actual scheme = table of tokens and weights */
	HashMap<Value, Integer> scheme = new HashMap<Value, Integer>();

	//--- constructors -------------------------------------------

	/** Creates a new endorsement scheme with the given name and base
	 * 
	 * @param name the scheme's name as String
	 * @param base the scheme's base
	 */
	public EndorsementScheme(String name, double base) {
		try {
			this.name = new Value(name, RU.SYMBOL);
			}
		catch(JessException je) {
			System.err.println("An error occurred at line "
					+ je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}
		this.base = base;
	}
	
	/** Creates a new endorsement scheme with the given name and base
	 * 
	 * @param name the scheme's name
	 * @param base the scheme's base
	 */
	public EndorsementScheme (Value name, double base){
		this.name = name;
		this.base = base;
	}

	//--- methods ----------------------------------------------
	
	/** adds an endorsement (pair of token and weight value)
	 * @param token the endorsement token
	 * @param value the weight to be associated with the token
	 */
	public void addEndorsement (Value token, int value){
		scheme.put(token, value);
	}
	
	/** calculates the overall endorsement value for the given list of endorsement tokens,
	 * using the formula from Moss (1995).
	 * 
	 * @param endorsements list of endorsement tokens
	 * @return overall endorsement value
	 */
	public double calculateEndorsementValue (ValueVector endorsements){
		double overallEndorsementValue = 0.0;
		for (int i = 0; i < endorsements.size(); i++) {
			try {
				Value end = endorsements.get(i);

				double endValue = scheme.get(end);
				
				//This is the formula to calculate the Overall Endorsement Value (OEV) from
				// Moss (1995).
				overallEndorsementValue += Math.signum(endValue)* Math.pow(base, Math.abs(endValue));
			}catch(JessException je) {
				System.err.println("An error occurred at line "
						+ je.getLineNumber());
				System.err.println("caused by " + je.getCause());
				System.err.println("Message: " + je.getDetail());
				je.printStackTrace();
			}
		}
		return overallEndorsementValue;
	}

	/** returns the name of the endorsement scheme
	 * 
	 * @return the name
	 */
	public Value getName(){
		return name;
	}
	
	/** sets the name of the endorsement scheme
	 * 
	 * @param name the name
	 */
	public void setName(Value name){
		this.name = name;
	}
	
	/** sets the name of the endorsement scheme
	 * 
	 * @param nameString the name as String
	 */
	public void setName (String nameString){
		try {
			name = new Value(nameString, RU.SYMBOL);
		}catch(JessException je) {
			System.err.println("An error occurred at line "
					+ je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}
	}
	
	/** returns the list of endorsement tokens
	 * 
	 * @return list of endorsement tokens
	 */
	public Value[] getEndorsementsList(){
		Value[] result = new Value[scheme.size()];
		scheme.keySet().toArray(result);
		return result;
	}
	
	/** returns a String representation of the endorsement scheme
	 * @return a String consisting of "endorsement-scheme:" and the name
	 */
	public String toString(){
		String str = new String("endorsement-scheme:" + name);
		return str;
	}
	
	/** prints the endorsement scheme (table of tokens and values) to console */
	public void print(){
		for (Value key : scheme.keySet()){
			int endorsementValue = scheme.get(key);
			System.out.print(key + ": " + endorsementValue + ", ");
		}
		System.out.println();
	}
	
	/** returns the weight associated with the given endorsement token
	 * 
	 * @param endorsement endorsement token
	 * @return endorsement value (weight)
	 */
	public int getValue(Value endorsement){
		return this.scheme.get(endorsement).intValue();
	}
}
