package org.cfpm.caves.za;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.HashSet;

import jess.*;

/** Goes through all the facts and updates agents and households accordingly.
 *  Used to outsource procedural / slow stuff from Jess to Java
 *  
 * @author Ruth, Scott
 */
public class FactBaseScourer {
	
	//--- attributes --------------------------------------------

	/** reference to the model */ 
	SouthAfricaModel model;
	/** reference to the Jess engine */
	Rete engine;

	//--- constructors -------------------------------------------

	/** Creates a new fact base scourer for the given model */
	public FactBaseScourer(SouthAfricaModel model) {
		this.model = model;
		this.engine = model.engine;
	}
	
	
	//--- methods ----------------------------------------------
	
	/** Runs through the fact base and sorts facts needed for updates into several 
	 * array lists: friends, endorsements, payments and employments. Then passes
	 * these lists on to the different update procedures: evolving tags, updating
	 * known persons, updating the friendship network, updating the households' cash,
	 * updating the households' expected income.
	 */
	public void update(){
		model.dump("update from Jess facts");
		// runs through list of facts and sorts needed facts into different array lists
		ArrayList<Fact> friendships = new ArrayList<Fact>();
		ArrayList<Fact> endorsements = new ArrayList<Fact>();
		ArrayList<Fact> payments = new ArrayList<Fact>();
		ArrayList<Fact> employments = new ArrayList<Fact>();
		Iterator factList = engine.listFacts();
		while (factList.hasNext()){
			Fact f = (Fact)factList.next();
			if (f.getName().endsWith("friends-list")){
				friendships.add(f);
			}
			if (f.getName().endsWith("endorsed")){
				endorsements.add(f);
			}
			if (f.getName().endsWith("payment")){
				payments.add(f);
			}
			if (f.getName().endsWith("employment")){
				employments.add(f);
			}
		}
		// pass on friendship list to evolve tags of agents
		evolveTags(friendships);
		// pass on endorsement list to update known persons for agents
		updateKnownPersons(endorsements);
		// pass on endorsement list to update friendship network
		updateFriends(endorsements);
		// pass on payment list to update cash of households
		updateHouseholdCash(payments);
		// pass on employment list to update expected income of households
		updateExpectedIncome(employments);
	}
	
	/** evolves the tags of agents so that -- according to their tag evolution propensity --
	 *  they become more similar to their best friends
	 *  
	 * @param friendships list of facts with "is-friend" endorsements
	 */
	void evolveTags(ArrayList<Fact> friendships){
		// evolve tags for agents
		try {
			Context context = engine.getGlobalContext();
			for (Fact f : friendships){
				int tickStamp = f.getSlotValue("tickStamp").intValue(context);
				// only consider facts of last tick
				if (tickStamp == (model.tick - 1)){
					Person agent = model.namePersonMap.get(f.getSlotValue("clauseOwner"));
					ValueVector friendsVV = f.getSlotValue("friends").listValue(context);
					Person[] friends = new Person[friendsVV.size()];
					ValueVector endValuesVV = f.getSlotValue("endorsement-values").listValue(context);
					double[] endValues = new double[endValuesVV.size()];
					for (int i = 0; i < friendsVV.size(); i++){
						String friendString = friendsVV.get(i).stringValue(context);
						friends[i] = model.namePersonMap.get(friendString);
						double endValue = endValuesVV.get(i).floatValue(context);
						endValues[i] = endValue;
					}
					agent.updateTag(friends, endValues);
				}
			}
			
		} catch (JessException je) {
			System.err.println("An error occurred at line " + je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}
	}
	
	/** updates the list of known persons for every agent by turning every endorsed agent
	 * into a known person (if they are not already known).
	 * 
	 * @param endorsements list of endorsement facts
	 */
	void updateKnownPersons(ArrayList<Fact> endorsements){
		// turn endorsed agents into known persons (if they are not already known)
		HashMap<Value, ArrayList<Value>> knownMap = new HashMap<Value, ArrayList<Value>>();
		try {
			for (Fact f : endorsements){
				Value agent = f.getSlotValue("clauseOwner");
				Value another = f.getSlotValue("endorsed-object");
				ArrayList<Value> knownPersons = knownMap.get(agent);
				if (knownPersons == null){
					knownPersons = new ArrayList<Value>();
				}

				if (!knownPersons.contains(another)) {
					knownPersons.add(another);
				}
				knownMap.put(agent, knownPersons);
			}
				
		} catch (JessException je) {
			System.err.println("An error occurred at line " + je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}
		// add new known persons to each person's "memory"
		Iterator<Value> agents = knownMap.keySet().iterator();
		while (agents.hasNext()){
			Value agent = agents.next();
			Person agentP = model.namePersonMap.get(agent.toString());
			if (agent == null || agentP == null){
				System.out.println("updateKnownPersons found a non-existing person: " + agent);
			}
			else {
				ArrayList<Value> known = knownMap.get(agent);
				agentP.addKnownPersons(known);
			}
		}	
	}
	
	/** Maintain the friendship network.
	 * 
	 * @param endorsements list of "is-friend" endorsements
	 */
	void updateFriends(ArrayList<Fact> endorsements){
		// create and maintain friendship network 
		HashMap<Value, ArrayList<Value>> network = new HashMap<Value, ArrayList<Value>>();
		// only consider endorsements from previous tick
		// only consider endorsements from previous tick
		try {
			for (Fact f : endorsements){
				String endorsement = f.getSlotValue("endorsement").toString();
				int tickStamp = f.getSlotValue("tickStamp").intValue(engine.getGlobalContext());
				if (tickStamp == model.getTick()-1 && endorsement.equals("is-friend")){
					Value friendName = f.getSlotValue("endorsed-object");
					Person friend = model.namePersonMap.get(friendName.toString());
					Value pName = f.getSlotValue("clauseOwner");
					Person p = model.namePersonMap.get(pName.toString());
					ArrayList<Value> friendList = network.get(pName);
					if (friend == null){
//						System.out.println("--- friend doesn't exist: " + friendName);
					}
					else if (!friend.isAlive()){
//						System.out.println("--- friend has died: " + friendName);
					}
					else {
						if (friendList == null){
							friendList = new ArrayList<Value>();
						}
						friendList.add(friendName);
						network.put(pName, friendList);
					}
				}
			}
		}catch (JessException je) {
			System.err.println("An error occurred at line " + je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}
		
		// stick friends into Person objects, replacing old list of friends
		// ### aha! When someone suddenly has 0 friends left his old list won't be replaced so he keeps his dead friends!!
		// --> first delete all old lists
		for (Person p : model.personList){
			p.friends.clear();
			if (model.getShowGUI()){
				// drop the old links
				p.getNode().removeFromLinks("friend");
			}
		}
		for (Iterator<Value> it = network.keySet().iterator(); it.hasNext(); ){
			Value name = it.next();
			Person p = model.namePersonMap.get(name.toString());
			p.friends = network.get(name);
			// check for doubles
			checkDoubles(p);
			// check number of friends
			if (p.friends.size() > p.maxNumFriends){
				System.out.println("## " + p.getName() + " has more friends than allowed: " + p.friends.size() + " vs. " + p.maxNumFriends);
			}
			if (model.getShowGUI()){
				// create new links
				for (Value friend : p.friends){
					NetworkLink link = new NetworkLink(p.node, model.namePersonMap.get(friend).getNode());
					link.setType("friend");
				}
			}
		}
	}
	
	private void checkDoubles(Person p){
		ArrayList<Value> singles = new ArrayList<Value>();
		ArrayList<Value> doubles = new ArrayList<Value>();
		for (Value friend : p.friends){
			if (singles.contains(friend)){
				doubles.add(friend);
			}
			else {
				singles.add(friend);
			}
		}
		if (doubles.size() > 0){
			System.out.print("## " + p.getName() + " has double friends:");
			for (Value name : doubles){
				System.out.print(" " + name.toString());
			}
			System.out.println(" ##");
		}
	}
	
	
	/** Updates each household's cash from the payments made and received the previous tick.
	 * Updates the spending commitments too.
	 * 
	 * @param payments list of payment facts
	 */
	void updateHouseholdCash(ArrayList<Fact> payments){
		// collect all payments received and all payments made and update cash accordingly
		// update spending commitments too (by subtracting all payments made)
		HashMap<Value, Double> hhPaymentMap = new HashMap<Value, Double>();
		HashMap<Value, Double> hhCommitmentMap = new HashMap<Value, Double>();
		try {
			for (Fact f : payments){
				int tickStamp = f.getSlotValue("payment-date").intValue(engine.getGlobalContext());
				// only consider payments of last tick
				if (tickStamp == (model.tick - 1)){
					Value recipient = f.getSlotValue("recipient");
					Value payer = f.getSlotValue("payer");
					double amount = f.getSlotValue("amount").floatValue(engine.getGlobalContext());
					String cause = f.getSlotValue("cause").toString();
					Double sum, commitments;
					// add to recipient's sum (if recipient is a household)
					if (recipient.toString().startsWith("household")){
						sum = hhPaymentMap.get(recipient);
						if (sum == null) sum = new Double(amount);
						else sum = new Double(sum.doubleValue() + amount);
						hhPaymentMap.put(recipient, sum);
						// record hh economy
						if (model.chosenHH.contains(recipient.toString()) && amount > 0){
							model.hhEconomyRecorder.recordIn(recipient.toString(), cause, amount);
						}
					}
					// add to recipient's household's sum (if recipient is a person)
					if (recipient.toString().startsWith("person")){
						Value receivingHH = model.namePersonMap.get(recipient.toString()).getHouseholdID();
						sum = hhPaymentMap.get(receivingHH);
						if (sum == null) sum = new Double(amount);
						else sum = new Double(sum.doubleValue() + amount);
						hhPaymentMap.put(receivingHH, sum);
						// record hh economy
						if (model.chosenHH.contains(receivingHH.toString()) && amount > 0){
							model.hhEconomyRecorder.recordIn(receivingHH.toString(), cause, amount);
						}
					}
					// subtract from payer's sum (if payer is a household)
					if (payer.toString().startsWith("household")){
						sum = hhPaymentMap.get(payer);
						if (sum == null) sum = new Double (-amount);
						else sum = new Double(sum.doubleValue() - amount);
						hhPaymentMap.put(payer, sum);
						commitments = hhCommitmentMap.get(payer);
						if (commitments == null) commitments = new Double(amount);
						else commitments = new Double(commitments.doubleValue() + amount);
						hhCommitmentMap.put(payer, commitments);
						// record hh economy
						if (model.chosenHH.contains(payer.toString()) && amount > 0){
							model.hhEconomyRecorder.recordOut(payer.toString(), cause, amount);
						}
					}
					// subtract from payer's household's sum (if payer is a person and it's not a remittance)
					if (payer.toString().startsWith("person") && !cause.equalsIgnoreCase("remittance")){
						Value payingHH = model.namePersonMap.get(payer.toString()).getHouseholdID();
						sum = hhPaymentMap.get(payingHH);
						if (sum == null) sum = new Double (-amount);
						else sum = new Double(sum.doubleValue() - amount);
						hhPaymentMap.put(payingHH, sum);
						commitments = hhCommitmentMap.get(payingHH);
						if (commitments == null) commitments = new Double(amount);
						else commitments = new Double(commitments.doubleValue() + amount);
						hhCommitmentMap.put(payingHH, commitments);
						// record hh economy
						if (model.chosenHH.contains(payingHH.toString()) && amount > 0){
							model.hhEconomyRecorder.recordOut(payingHH.toString(), cause, amount);
						}
					}
					
					// record village economy
//					String fromHH = "", toHH = "";
//					try{
//					if (payer.toString().startsWith("household")){
//						fromHH = payer.toString();
//					}
//					else if (payer.toString().startsWith("person") && !cause.equalsIgnoreCase("remittance")){
//						fromHH = model.namePersonMap.get(payer.toString()).getHouseholdID().toString();
//					}
//					else if (cause.equalsIgnoreCase("burial") || cause.endsWith("-payout")){
//						fromHH = "groups";
//					}
//					else {
//						fromHH = "outside";
//					}
//					if (recipient.toString().startsWith("household")){
//						toHH = recipient.toString();
//					}
//					else if (recipient.toString().startsWith("person")){
//						toHH = model.namePersonMap.get(recipient.toString()).getHouseholdID().toString();
//					}
//					else if (cause.equalsIgnoreCase("group-fee")){
//						toHH = "groups";
//					}
//					else {
//						toHH = "outside";
//					}
//					model.villageEconomyRecorder.record(fromHH, toHH, amount);
//					} catch (NullPointerException e){
//						model.dump("NULLPOINTEREXCEPTION trying to record village economy");
//						model.dump("fromHH = " + fromHH + ", toHH = " + toHH + ", amount = " + amount + ", cause = " + cause + ", payer = " + payer + ", recipient = " + recipient);
//					}
//					
//					// record hh income
//					if (cause.equalsIgnoreCase("wage") || cause.equalsIgnoreCase("remittance")){
//						if (recipient.toString().startsWith("household")){
//							toHH = recipient.toString();
//						}
//						else if (recipient.toString().startsWith("person")){
//							toHH = model.namePersonMap.get(recipient.toString()).getHouseholdID().toString();
//						}
//						model.hhIncomeRecorder.record(toHH, amount);
//					}
//					// record payouts
//					if (cause.endsWith("-payout")){
//						if (recipient.toString().startsWith("household")){
//							toHH = recipient.toString();
//						}
//						else if (recipient.toString().startsWith("person")){
//							toHH = model.namePersonMap.get(recipient.toString()).getHouseholdID().toString();
//						}						
//						model.hhIncomeRecorder.recordPayOut(toHH, amount);						
//					}
//					// record hh group fees
//					if (cause.equalsIgnoreCase("group-fee")){
//						if (payer.toString().startsWith("household")){
//							fromHH = payer.toString();
//						}
//						else if (payer.toString().startsWith("person")){
//							fromHH = model.namePersonMap.get(payer.toString()).getHouseholdID().toString();
//						}
//						model.hhIncomeRecorder.recordFee(fromHH, amount);
//					}
//					// record school fees
//					if (cause.equalsIgnoreCase("school-fee")){
//						if (payer.toString().startsWith("household")){
//							fromHH = payer.toString();
//						}
//						else if (payer.toString().startsWith("person")){
//							fromHH = model.namePersonMap.get(payer.toString()).getHouseholdID().toString();
//						}
//						model.hhIncomeRecorder.recordSchoolFee(fromHH, amount);
//					}
//					// record food costs
//					if (cause.equalsIgnoreCase("food")){
//						if (payer.toString().startsWith("household")){
//							fromHH = payer.toString();
//						}
//						else if (payer.toString().startsWith("person")){
//							fromHH = model.namePersonMap.get(payer.toString()).getHouseholdID().toString();
//						}
//						model.hhIncomeRecorder.recordFoodCost(fromHH, amount);
//					}
//					// record piece jobs
//					if (cause.equalsIgnoreCase("wage")){
//						if (payer.toString().startsWith("household")){
//							fromHH = payer.toString();
//						}
//						else if (payer.toString().startsWith("person")){
//							fromHH = model.namePersonMap.get(payer.toString()).getHouseholdID().toString();
//						}
//						model.hhIncomeRecorder.recordPieceJob(fromHH, amount);
//					}
//					// record donations (incl. lobola) or burial costs
//					if (cause.equalsIgnoreCase("marriage") || cause.endsWith("-request") || cause.equalsIgnoreCase("burial")){
//						if (payer.toString().startsWith("household")){
//							fromHH = payer.toString();
//						}
//						else if (payer.toString().startsWith("person")){
//							fromHH = model.namePersonMap.get(payer.toString()).getHouseholdID().toString();
//						}
//						if (recipient.toString().startsWith("household")){
//							// it's a donation
//							model.hhIncomeRecorder.recordDonation(fromHH, amount);
//						}
//						else {
//							// it's "other" costs
//							model.hhIncomeRecorder.recordOtherExpenditure(fromHH, amount);
//						}
//					}
//					// record other costs: migration, travel, new house
//					if (cause.equalsIgnoreCase("new-house")||cause.equalsIgnoreCase("travel")||cause.equalsIgnoreCase("migration")){
//						if (payer.toString().startsWith("household")){
//							fromHH = payer.toString();
//						}
//						else if (payer.toString().startsWith("person")){
//							fromHH = model.namePersonMap.get(payer.toString()).getHouseholdID().toString();
//						}
//						model.hhIncomeRecorder.recordOtherExpenditure(fromHH, amount);
//					}
				}
			}
				
		} catch (JessException je) {
			System.err.println("An error occurred at line " + je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}
		// add sum to household's cash and update spending commitments
		Iterator<Value> households = hhPaymentMap.keySet().iterator();
		while (households.hasNext()){
			Value hhName = households.next();
			Household hh = model.nameHouseholdMap.get(hhName.toString());
			if (hh == null){
				model.dump(">>>>>>>>>>>>>>>>>> unknown household: " + hhName);
			}
			else {
				hh.addCash(hhPaymentMap.get(hhName).doubleValue());
				Double commitments = hhCommitmentMap.get(hhName);
				if (commitments != null){
					hh.setSpendingCommitments(hh.getSpendingCommitments() - commitments.doubleValue());
				}
			}
		}	
	}
	
	/** Update each household's expected income from the employment facts, once every month.
	 * 
	 * @param employments list of employment facts
	 */
	void updateExpectedIncome(ArrayList<Fact> employments){
		// only done at the beginning of a month
		if (model.getWeek() != 0) return;

		// collect info on employments
		HashMap<Household, Double> employmentMap = new HashMap<Household, Double>();
		try {
			for (Fact f : employments){
				// for now: only consider government jobs (as they are the stable ones)
				Value employer = f.getSlotValue("employer");
				if (employer.toString().equals("government")){
					double wage = f.getSlotValue("wage").floatValue(engine.getGlobalContext());
					Value employee = f.getSlotValue("employee");
					Household hh = model.namePersonMap.get(employee.toString()).household;
					Double sum = employmentMap.get(hh);
					if (sum == null) sum = new Double(wage);
					else sum = new Double(sum.doubleValue() + wage);
					employmentMap.put(hh, sum);
				}
			}
		} catch (JessException je) {
			System.err.println("An error occurred at line " + je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}
		
		// let households calculate their expected incomes
		for (Household hh : model.householdList){
			Double wages = employmentMap.get(hh);
			double income = hh.calculateExpectedIncome((wages == null)? 0 : wages.doubleValue());
			hh.setExpectedIncome(income);
		}
	}
	
	/** internal helper method applied to the friendship network*/
	private void computeReciprocalLinkProportion(HashMap<NetworkNode, ArrayList<NetworkNode>> network){
		int numReciprocal = 0, numOneDirectional = 0;
		for (NetworkNode node : network.keySet()){
			for (NetworkNode friend : network.get(node)){
				if (network.get(friend).contains(node)){
					// reciprocal link found
					numReciprocal++;
				}
				else numOneDirectional++;
			}
		}
		model.dump("There are " + numReciprocal + " reciprocal links and " + numOneDirectional + " one-directional links");
		if ((numReciprocal + numOneDirectional) > 0){
			model.dump("That's " + ((numReciprocal / (double)(numReciprocal + numOneDirectional)) * 100) + "% reciprocal links");
		}
		else {
			model.dump("That's 0% reciprocal links");
		}
	}
	
	/** internal helper method applied to the friendship network*/
	private void computeKinshipProportion(HashMap<NetworkNode, ArrayList<NetworkNode>> network){
		int numKinship = 0, numNonKin = 0;
		for (NetworkNode node : network.keySet()){
			for (NetworkNode friend : network.get(node)){
				Household nodeHH = model.namePersonMap.get(node.id).getHousehold();
				Household friendHH = model.namePersonMap.get(friend.id).getHousehold();
				if (nodeHH.isKinOf(friendHH)) numKinship++;
				else numNonKin++;
			}
		}
		model.dump("There are " + numKinship + " friendship links between kin and " + numNonKin + " friendship links between non-kin");
		if ((numKinship + numNonKin) > 0){
			model.dump("That's " + ((numKinship / (double)(numKinship + numNonKin)) * 100) + "% kin links");
		}
		else {
			model.dump("That's 0% kin links");
		}
	}
	
}
