package org.cfpm.caves.za;

import java.util.ArrayList;
import java.util.HashMap;

import uchicago.src.sim.util.SimUtilities;
//import uchicago.src.sim.network.DefaultEdge;
import java.beans.*;

import jess.JessException;
import jess.RU;
import jess.Rete;
import jess.Value;
import jess.WatchConstants;

/** An abstract group. Provides the basic functionality for every type of group 
 * (like burial societies or stokvels). 
 *    
 * @author Ruth
 *
 */
public abstract class Group {
	
	//--- constants -----------------------------------------------------------
	
	/** counter for running ID */
	private static int lastID = 0;

	//--- group attributes ----------------------------------------------------
	
	/** Name of the group. Build from group type and running id */
	Value name;
	/** member list */
	ArrayList<Person> members;
	/** list of member names, for easier access within Jess */
	Value[] memberIDs;
	protected PropertyChangeSupport pcs = new PropertyChangeSupport(this);	
	SouthAfricaModel model;

	//--- constructors --------------------------------------------------------
	
	/** Creates a new group of the given type with the given members.
	 * @param type the group type
	 * @param members the list of members
	 * @param model the model
	 */
	public Group(String type, ArrayList<Person> members, SouthAfricaModel model){
		setName(type + "-" + (lastID++));
		this.model = model;
		this.members = new ArrayList<Person>();
		this.memberIDs = new Value[0];
		for (Person member : members){
			addMember(member);
		}
		model.dump("New group " + this.name + " created.");		
	}

	/** adds a new member to the group. This will cause all previous members 
	 * and the new member to endorse each other with the same-club endorsement.
	 * 
	 * @param member the new member
	 */
	public void addMember(Person member){
		// have new member endorse every other member with same-club endorsement
		// and vice versa
		endorse(member);
		// have new member add all other members as known persons and vice versa
		addAsKnown(member);
		// add member to group
		members.add(member);
		Value[] newIDs = new Value[members.size()];
		for (int i = 0; i < members.size() - 1; i++) {
			newIDs[i] = members.get(i).name;
		}
		newIDs[members.size() - 1] = member.name;
//		printIDs(newIDs);
		setMemberIDs(newIDs);		
	}

	/** internal helper function for debugging: prints a list of member IDs.
	 */
	private void printIDs(Value[] ids){
		System.out.print("\nnew member ids: ");
		for (Value v : ids){
			System.out.print(v.toString() + " ");
		}
		System.out.println("\n");
	}
	
	/** removes the given member from this group. This will also
	 * cause every other member to retract their same-club endorsement
	 * of this member and vice versa.
	 * 
	 * @param member the member to leave the group
	 */
	public void removeMember(Person member){
		members.remove(member);
		int currIndex = 0;
		Value[] newIDs = new Value[members.size()];
		for (int i = 0; i < members.size(); i++) {
			if (member.name != memberIDs[i]) {
				newIDs[currIndex] = members.get(i).name;
				currIndex++;
			}
		}
		setMemberIDs(newIDs);
		// remove same-club endorsements
		unEndorse(member);
	}
	
	/** internal function: the given member endorses every other member with 
	 * same-club endorsement and vice versa.
	 */
	private void endorse(Person member){
		// the new member endorses every other member with same-club endorsement
		// and vice versa
		for (Person p : this.members){
			if (p != member){
				member.assertSameClubEndorsement(p);
				p.assertSameClubEndorsement(member);
			}
		}
	}
	
	/** internal function: the leaving member removes same-club endorsement 
	 * for every other member and vice versa.
	 */
	private void unEndorse(Person member){
		// the leaving member removes same-club endorsement for every other member
		// and vice versa
		for (Person p : this.members){
			if (p != member){
				member.removeSameClubEndorsement(p);
				p.removeSameClubEndorsement(member);
			}
		}
	}
	
	/** internal function: the given member adds every other member as a known 
	 * person and vice versa.
	 */
	private void addAsKnown(Person member){
		// the new member adds all other member as known persons
		// and vice versa
		member.addKnownPersons(this.memberIDs);
		for (Person p : members){
			p.addKnownPerson(member.name);
		}
	}

	/** returns the member IDs
	 * 
	 * @return list of member names
	 */
	public Value[] getMemberIDs() {
		return memberIDs;
	}

	/** sets the member IDs
	 * 
	 * @param IDs list of member names
	 */
	public void setMemberIDs(Value[] IDs) {
		Value[] old = memberIDs;
		memberIDs = IDs;
		pcs.firePropertyChange("memberIDs", old, memberIDs);
	}
	
	/** adds a property change listener (for Jess) */	
	public void addPropertyChangeListener(PropertyChangeListener p) {
		pcs.addPropertyChangeListener(p);
	}

	/** removes a property change listener (for Jess) */
	public void removePropertyChangeListener(PropertyChangeListener p) {
		pcs.removePropertyChangeListener(p);
	}
	
	/** returns the group's name
	 * 
	 * @return name of the group
	 */
	public Value getName(){
		return name;
	}
	
	/** sets the group's name
	 * 
	 * @param name name of the group
	 */
	public void setName(Value name){
		Value old = this.name;
		this.name = name;
		pcs.firePropertyChange("name", old, this.name);
	}
	
	/** sets the group's name
	 * 
	 * @param nameString name of the group as String
	 */
	public void setName(String nameString){
		try {
			Value name = new Value(nameString, RU.SYMBOL);
			setName(name);
		}
		catch(JessException je) {
			System.err.println("An error occurred at line " + je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}
	}
	
	/** returns a String representation of this group
	 * @return the group's name followed by the number of members.
	 */
	public String toString(){
		StringBuffer buf = new StringBuffer(this.name.toString());
		buf.append(" with ");
		buf.append(members.size());
		buf.append(" members");
		return buf.toString();
	}

}
