package org.cfpm.caves.za;

//import cern.jet.random.Normal;
//import cern.jet.random.engine.MersenneTwister;
import flanagan.math.PsRandom;

/** This class models the HIV infection for a person. It's implemented as a separate class
 * instead of an inner class to Person to be able to use static attributes for the 
 * transmission probabilities and distribution of disease progression.
 * 
 * @author Ruth
 *
 */
public class HIVAIDS {

	// for clinical stages of HIV/AIDS see: http://whqlibdoc.who.int/hq/2005/WHO_HIV_2005.02.pdf
	// for a distribution modelling progression from HIV to AIDS see: http://www.who.int/healthinfo/paper03.pdf

	// for HIV/AIDS we'll use a Normal distribution with mean = 8 years and sd = 3 years
//	static Normal progressionDist = new Normal(8.0, 3.0, new MersenneTwister());	
	// use Weibull distribution with parameters mu = 0, sigma = 9.0, gamma = 5.0 (years)
	static PsRandom progressionDist = new PsRandom();
	static double mu = 0;
	static double sigma = 9.0;
	static double gamma = 5.0;

	// transmission probabilities, according to clinical stage (from Wawer et al.)
	static final double[] transmissionProbs = new double[]{0.0, 0.0082, 0.0015, 0.0007, 0.0028};
	// durations of stages: (1) 3 months, (2) 12 months, (3) n months, (4) after developing AIDS to death
	
	/** the tick of contracting HIV. Default is -1 (no infection) */
	int infectedSince = -1; // tick of contraction of HIV (-1 = no infection)
	/** tick of developing AIDS. Default is -1 (no AIDS yet) */
	int developedAIDS = -1; // tick of developing AIDS symptoms
	/** the current clinical stage of HIV. Default is 0 (no infection)*/
	int currentStage = 0;
	
	/** Creates a new HIV/AIDS handler at the current tick. The clinical stage is set to 1.
	 * 
	 * @param tick the tick when the person contracted HIV
	 */
	HIVAIDS(int tick){
		this(tick, 1);
	}
	
	/** Creates a new HIV/AIDS handler for the current tick and clinical stage.
	 * This constructor is provided for the model initialisation.
	 * @param tick the tick when the person contracted HIV
	 * @param stage the clinical stage of HIV 
	 */
	HIVAIDS(int tick, int stage){
		this.infectedSince = tick;
		this.currentStage = stage;
//		this.developedAIDS = tick + (int)(progressionDist.nextDouble()*12*4);
		int durationToAIDS = (int)(progressionDist.nextWeibull(mu, sigma, gamma)*12*4);
		if (currentStage == 2) durationToAIDS -= 3*4;
		if (currentStage == 3) durationToAIDS -= 15*4;
		if (currentStage == 4) durationToAIDS = 0;
		this.developedAIDS = tick + durationToAIDS;
	}
	
	/** returns the tick of HIV infection 
	 * @return tick of infection
	 */
	int getInfectedSince(){
		return infectedSince;
	}
	
	/** returns the tick of progression to AIDS
	 * @return tick of AIDS progression
	 */
	int getDevelopedAIDS(){
		return developedAIDS;
	}
	
	/** returns if the person is HIV-positive or not
	 * 
	 * @return true if the person is HIV-positive (infectedSince >= 0), false otherwise
	 */
	public boolean isHIVPositive(){
		return (infectedSince >= 0);
	}
	
	/** returns if the person has progressed to AIDS or not
	 * 
	 * @return true if the person has AIDS (developedAIDS >= 0), false otherwise
	 */
	public boolean hasAIDS(int currentTick){
		return (developedAIDS >= 0 && developedAIDS <= currentTick); 
	}
	
	/** updates the clinical stage if necessary
	 * 
	 * @param currentTick the current tick
	 */
	private void updateStage(int currentTick){
		if (infectedSince < 0) return;
		int lag = currentTick - infectedSince;
		if ((currentStage == 1 && lag > 3*4) ||	// change to stage 2
			 (currentStage == 2 && lag > 15*4) ||	// change to stage 3
			 (currentStage == 3 && hasAIDS(currentTick)))	// change to stage 4
			currentStage++;
	}
	
	/** returns the probability of transmitting HIV according to the current clinical stage
	 * 
	 * @param currentTick the current tick
	 * @return transmission probability
	 */
	public double getTransmissionProbability(int currentTick){
		updateStage(currentTick);
		return transmissionProbs[currentStage];
	}
		
	

}
