package org.cfpm.caves.za;

import jess.*;
import uchicago.src.sim.gui.Drawable;
import uchicago.src.sim.gui.SimGraphics;
import uchicago.src.sim.util.Random;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.beans.*;

/** This class represents a household in a village. A household consists of
 * a group of agents, usually but not necessarily related, who pool and share
 * all resources. Decisions on the household level are taken by a designated
 * household head. 
 * Used as a shadow fact in Jess.
 * 
 * @author Shah Alam, Ruth Meyer, Scott Moss
 *
 */
public class Household implements Drawable {
	
	//--- attributes ---------------------------------------------------------
	
	/** the household ID */
	Value id;
	/** the church this household belongs to */
	Value church;
	/** the village this household is situated in */
	Value village;
	/** the household head */
	Person head;
	/** the household members */
	ArrayList<Person> members;
	/** the names of the household members (for easier access in Jess) */
	Value[] memberIDs;
	/** the household's available money */
	double cash;
	/** the household's monthly food cost */
	double monthlyFoodCost;
	/** the household's expected monthly income */
	double expectedIncome;
	/** the household's commitments to pay for stuff */
	double spendingCommitments;
	
	/** the associated network node for the kinship network */
	NetworkNode node = null;
	/** the x coordinate of the household's position on the 2D grid */
	int x;
	/** the y coordinate of the household's position on the 2D grid */
	int y;

	/** reference to the model */
	SouthAfricaModel model;
	private PropertyChangeSupport pcs = new PropertyChangeSupport(this);
	
	//--- contstructors ---------------------------------------------------------
	
	/** Creates a new empty household. Used during model initialisation 
	 *  @param model the model
	 */
	public Household(SouthAfricaModel model){
		this.model = model;
		members = new ArrayList<Person>();
		memberIDs = new Value[0];
	}
	
	/** Create a new household for a newly-wed couple during the simulation
	 * 
	 * @param model the model 
	 * @param id the household ID
	 * @param church the church the household belongs to
	 * @param village the village the household is situated
	 */
	public Household(SouthAfricaModel model, String id, Value church, Village village){
		this(model);
		try {
			this.id = new Value(id, RU.SYMBOL);
		}
		catch(JessException je) {
			System.err.println("An error occurred at line " + je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}
		this.church = church;
		this.village = village.name;
		this.node = new NetworkNode(this.id);
		// locate in village
		int[] extension = village.getExtension();
		int x, y;
		do {
			x = Random.uniform.nextIntFromTo(extension[0], extension[2]);
			y = Random.uniform.nextIntFromTo(extension[1], extension[3]);
		} while (model.space.getObjectAt(x, y) != null);
		setXY(x, y);
		model.space.putObjectAt(x, y, this);
	}
	
	//--- methods ------------------------------------------------------------
	
	/** sets the household ID
	 * @param newId the household ID
	 */
	public void setId(Value newId){
		Value old = id;
		id = newId;
		pcs.firePropertyChange("id", old, id);
	}
	
	/** sets the household ID
	 * @param newId the household ID as String
	 */
	public void setId(String newId){
		Value old = id;
		try {
			id = new Value(newId, RU.SYMBOL);
		}
		catch(JessException je) {
			System.err.println("An error occurred at line " + je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}
		pcs.firePropertyChange("id", old, id);
	}
	
	/** returns the household ID
	 * 
	 * @return household ID
	 */
	public Value getId() {
		return id;
	}
	
	/** returns the household's head
	 * 
	 * @return hh head
	 */
	public Person getHead(){
		return head;
	}
	
	/** sets the household's head
	 * 
	 * @param head the hh head
	 */
	public void setHead(Person head){
		Person old = this.head;
		this.head = head;
		pcs.firePropertyChange("head", old, this.head);
		pcs.firePropertyChange("headID", (old==null)?null:old.getName(), (head==null)?null:this.head.getName());
	}
	
	/** returns the name of the household head. Provided for easier access from Jess.
	 * 
	 * @return name of the household head
	 */
	public Value getHeadID(){  // for easier access from Jess
		return head.getName();
	}
	
	/** sets the household's members
	 * 
	 * @param mems list of members
	 */
	public void setMembers(ArrayList<Person> mems) {
		members = mems;
	}
	
	/** returns the household's members
	 * 
	 * @return list of members
	 */
	public ArrayList<Person> getMembers() {
		return members;
	}
	
	/** Adds the given person as a household member
	 * 
	 * @param member the new member
	 */
	public void addMember(Person member){
		members.add(member);
		Value[] newIDs = new Value[members.size()];
		for (int i = 0; i < members.size() - 1; i++) {
			newIDs[i] = members.get(i).name;
		}
		newIDs[members.size() - 1] = member.name;
		setMemberIDs(newIDs);
	}
	
	/** Removes the given person from the household. If this person
	 * was the last adult member, the household will dissolve unless
	 * any children left can't be accommodated by relatives.
	 * 
	 * @param member the leaving member
	 */
	public void removeMember(Person member){
		members.remove(member);
		Value[] newIDs = new Value[members.size()];
		for (int i = 0; i < members.size(); i++) {
			newIDs[i] = members.get(i).name;
		}
		setMemberIDs(newIDs);
		// if the removed person was the household head, determine new head (if there are any members left)
		if (member.isHead() && members.size() > 0){
			determineNewHead(member);
		}
		// check if household will dissolve due to 
		// (a) no members left or 
		// (b) only children left that can be accommodated by extended family
		if (members.size() == 0) {
			Person relative = member.findLivingRelative(false);
			dissolve(relative);
		}
		else if (head.age <= 16){
			Person relative = head.findLivingRelative(true);
			if (relative == null){
				// no direct relatives found, use kinship links on hh level
				relative = head.findBestAccommodator(getHeadsOfKinHouseholds());
			}
			if (relative != null){
				relative.getHousehold().accommodate(this.members);
				model.dump("Orphan children are accommodated by " + relative.getHousehold().id.toString());
				dissolve(relative);
			}
		}
	}
	
	/** internal helper function: retrieves the heads of households related to this household
	 * via the kinship network
	 * @return list of persons
	 */
	private ArrayList<Person> getHeadsOfKinHouseholds(){
		ArrayList<Person> heads = new ArrayList<Person>();
		ArrayList links = this.node.getOutEdges();
		for (Object o : links){
			NetworkLink link = (NetworkLink)o;
			if (link.getType().equals("kinship")){
				NetworkNode hhNode = link.getTo();
				Household hh = model.nameHouseholdMap.get(hhNode.id.toString());
				heads.add(hh.getHead());
			}
		}
		return heads;
	}
	
	/** internal function: let the given orphans join the household
	 * 
	 * @param orphans list of orphaned children
	 */
	private void accommodate(ArrayList<Person> orphans){
		// add all orphans as members to this household
		for (Person p : orphans){
			addMember(p);
		}
	}
	
	/** internal function: this household dissolves and the given heir inherits all assets
	 * 
	 * @param heir the heir
	 */
	private void dissolve(Person heir){
		model.dump(id.toString() + " dissolves.");
		if (heir != null){
			// household of heir inherits all assets of this household
			// -- at the moment: only the cash
			// TODO: sell the house? 
			model.dump(heir.getName() + " (" + heir.getGender() + ", " + heir.getAge() + ") of " + heir.getHouseholdID() + " inherits all assets.");
			heir.getHousehold().addCash(this.cash);
			this.cash = 0;
		}
		// remove household from model
		model.removeHousehold(this);
	}
	
	/** internal function: finds a new household head among the members
	 * 
	 * @param oldHead the old head who has died
	 */
	private void determineNewHead(Person oldHead){
		Person newHead;
		// pick spouse if old head is married
		if (oldHead.getMaritalStatus().toString().equalsIgnoreCase("married")){
			newHead = oldHead.spouse;
		}
		// else pick eldest of remaining members 
		// idea behind this: it's plausible that the eldest adult (senior?) will take responsibility
		// and if there aren't any seniors/adults left, the eldest child takes over
		// (TODO: this should probably be replaced by a closer look at the family structure)
		else {
			newHead = findEldestMember();
		}
		if (!newHead.isAlive()){ 
			model.dump("ZOMBIE ALERT! " + newHead.getName() + " was picked as new head even though they are dead.");
			model.dump("Properties: " + newHead.getAge() + ", " + newHead.getGender() + ", " + (newHead.spouse != null ? "married to " + newHead.spouse.getName() : newHead.getMaritalStatus()));
		}
		oldHead.setHead(false);
		newHead.setHead(true); // sets it in household too
		model.dump(newHead.getName() + " (" + newHead.getGender() + ", " + newHead.getAge() + ") becomes new head of " + this.id);
		// if new head is migrant and other members are all (school) children, have new head return home
		if (newHead.isMigrant()){
			Person nextEldest = findNextEldestMember(newHead);
			if (nextEldest.isAtSchool() || nextEldest.getAge() < 16){
				newHead.setMigrant(false);
				newHead.setOnVisit(false);
				model.dump(newHead.getName() + " returns home from migration");
			}
		}
	}
	
	/** internal function: finds the eldest person amongst the members to take over as household head
	 * 
	 * @return eldest member
	 */
	private Person findEldestMember(){
		// go through member list and find eldest member
		Person eldest = this.members.get(0);
		for (Person p : this.members){
			if (p.getAge() > eldest.getAge()){
				eldest = p;
			}
		}
		return eldest;
	}
	
	/** internal function: finds the member who is eldest after the given member
	 * 
	 * @param p member
	 * @return next eldest member
	 */
	private Person findNextEldestMember(Person p){
		// go through member list and find eldest after p
		Person eldest = this.members.get(0);
		for (Person person : this.members){
			if (!person.equals(p) && person.getAge() <= p.getAge() && person.getAge() > eldest.getAge()){
				eldest = person;
			}
		}
		return eldest;
	}
	
	/** returns the member IDs
	 * 
	 * @return list of member names
	 */
	public Value[] getMemberIDs() {
		return memberIDs;
	}

	/** sets the member IDs
	 * 
	 * @param IDs list of member names
	 */
	public void setMemberIDs(Value[] IDs) {
		Value[] old = memberIDs;
		memberIDs = IDs;
		pcs.firePropertyChange("memberIDs", old, memberIDs);
	}
	
	/** sets the church this household belongs to
	 * 
	 * @param ch the church
	 */
	public void setChurch(Value ch){
		Value old = church;
		church = ch;
		pcs.firePropertyChange("church", old, church);
	}
	
	/** sets the church this household belongs to
	 * 
	 * @param ch the church as String
	 */
	public void setChurch(String ch){
		try {
			setChurch(new Value(ch, RU.SYMBOL));
		} catch (JessException je) {
			System.err.println("An error occurred at line " + je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}
	}
	
	/** returns the church this household belongs to
	 * 
	 * @return the church
	 */
	public Value getChurch(){
		return church;
	}

	/** returns the associated network node
	 * 
	 * @return the node
	 */
	public NetworkNode getNode(){
		return this.node;
	}
	
	/** sets the associated network node
	 * 
	 * @param node the node
	 */
	public void setNode(NetworkNode node){
		NetworkNode old = this.node;
		this.node = node;
		this.node.setId(id);
		pcs.firePropertyChange("node", old, this.node);
	}

	/** adds a property change listener (for Jess) */
	public void addPropertyChangeListener(PropertyChangeListener p) {
		pcs.addPropertyChangeListener(p);
	}

	/** removes a property change listener (for Jess) */
	public void removePropertyChangeListener(PropertyChangeListener p) {
		pcs.removePropertyChangeListener(p);
	}	
	

	/** returns the household's cash
	 * 
	 * @return amount of cash
	 */
	public double getCash() {
		return cash;
	}

	/** sets the household's cash to the given amount
	 * 
	 * @param cash amount of cash
	 */
	public void setCash(double cash) {
		double old = this.cash;
		this.cash = cash;
		pcs.firePropertyChange("cash", old, this.cash);
	}
	
	/** adds the given cash to the household's cash
	 * 
	 * @param cash amount to add
	 */
	public void addCash(double cash){
		setCash(this.cash + cash);
	}
	
	/** removes the specified amount from the household's cash
	 * 
	 * @param cash amount to remove
	 */
	public void removeCash(double cash){
		setCash(this.cash - cash);
	}

	/** returns the expected monthly income
	 * 
	 * @return expected monthly income
	 */
	public double getExpectedIncome() {
		return expectedIncome;
	}

	/** sets the expected monthly income
	 * 
	 * @param expectedIncome the expected monthly income
	 */
	public void setExpectedIncome(double expectedIncome) {
		double old = this.expectedIncome;
		this.expectedIncome = expectedIncome;
		pcs.firePropertyChange("expectedIncome", old, this.expectedIncome);
	}

	/** returns the monthly food cost
	 * 
	 * @return monthly food cost
	 */
	public double getMonthlyFoodCost() {
		return monthlyFoodCost;
	}

	/** sets the monthly food cost
	 * 
	 * @param monthlyFoodCost monthly food cost
	 */
	public void setMonthlyFoodCost(double monthlyFoodCost) {
		double old = this.monthlyFoodCost;
		this.monthlyFoodCost = monthlyFoodCost;
		pcs.firePropertyChange("monthlyFoodCost", old, this.monthlyFoodCost);
	}
	
	/** calculates the monthly food cost depending on the members' age and gender,
	 * applying the food cost constants defined in the model
	 * 
	 * @return monthly food cost
	 */
	public double calculateMonthlyFoodCost(){
		double cost = 0;
		for (Person p : this.members){
			if (!p.isMigrant()){
				if (p.getAge() <= 7){
					cost += model.foodCostChild;
				}
				else if (p.getGender().toString().equals("female")){
					cost += model.foodCostFemaleAdult;
				}
				else {
					cost += model.foodCostMaleAdult;
				}
			}
		}
		return cost;
	}
	
	/** calculates the expected income, taking state grants and wages
	 * from employment into account
	 * 
	 * @param wages wages from employment
	 * @return expected income
	 */
	public double calculateExpectedIncome(double wages){
		// include pensions, child grants and wages from employment
		// how about remittances?
		double income = wages;
		for (Person p : this.members){
			if (p.getAge() < 7){
				income += model.childGrant;
			}
			else if (p.isRegisteredForPension()){
				income += model.statePension;
			}
		}
		return income;
	}
	
	/** contribute money to the burial costs of a related or neighbouring household
	 * 
	 * @param deceasedPerson the deceased
	 * @param deceasedHousehold the deceased's household
	 * @param isKin kinhsip status (true = related, false = not related)
	 * @return amount of contribution
	 */
	public double contributeToBurial(String deceasedPerson, String deceasedHousehold, boolean isKin){
		// amount contributed depends on (a) own wealth and (b) kinship of deceased or (c) being in same village
		// assumption: relatives will try to give more than neighbours
		// question: would a household borrow money to be able to contribute to a funeral?
		double contribution = 0.0;
		if (cash < 100) return 0.0; // don't contribute if there's not enough cash
		if (this.expectedIncome < 1200){
			// give 10 Rand per adult/senior member attending the funeral
			contribution = 10 * Random.uniform.nextIntFromTo(1, numberOfAdults()+numberOfSeniors());
		}
		else {
			// give between 20 and 50 Rand per adult/senior member attending
			contribution = Random.uniform.nextDoubleFromTo(20, 50) * 
								Random.uniform.nextIntFromTo(1, numberOfAdults()+numberOfSeniors());
		}
		if (isKin){
			// double the amount
			contribution *= 2.0;
		}
//		model.dump(this.id + " contributes to funeral of " + deceasedPerson + ": " + Math.min(contribution, cash) + " Rand (" + (isKin ? "is kin)" : "same village)"));
		return Math.min(contribution, cash);
	}
	
	/** return the number of adults amongst the members
	 * 
	 * @return number of adults (ages 18-59)
	 */
	int numberOfAdults(){
		int num = 0;
		for (Person p : members){
			int age = p.getAge();
			if (age >= 18 & age < 60) num++;
		}
		return num;
	}
	
	/** return the number of children amongst the members
	 * 
	 * @return number of children (ages 0-17)
	 */
	int numberOfChildren(){
		int num = 0;
		for (Person p : members){
			if (p.getAge() < 18) num++;
		}
		return num;
	}
	
	/** return the number of seniors amongst the members
	 * 
	 * @return number of seniors (ages >= 60)
	 */
	int numberOfSeniors(){
		int num = 0;
		for (Person p : members){
			if (p.getAge() >= 60) num++;
		}
		return num;
	}
	
	/** sets the village this household belongs to
	 * 
	 * @param village the village
	 */
	public void setVillage(Value village){
		Value old = this.village;
		this.village = village;
		pcs.firePropertyChange("village", old, village);
	}
	
	/** sets the village this household belongs to
	 * 
	 * @param villageString the village as a String
	 */
	public void setVillage(String villageString){
		try {
			setVillage(new Value(villageString, RU.SYMBOL));
		} catch (JessException je) {
			System.err.println("An error occurred at line " + je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}
	}
	
	/** returns the village this household belongs to
	 * 
	 * @return the village
	 */
	public Value getVillage(){
		return village;
	}
	
	/** returns a random adult from the household members
	 * 
	 * @return random adult member
	 */
	Person pickRandomAdult(){
		ArrayList<Person> adults = new ArrayList<Person>();
		for (Person p : members){
			if (p.getAge() > 16){
				adults.add(p);
			}
		}
		if (adults.size() > 0){
			return adults.get(Random.uniform.nextIntFromTo(0, adults.size()-1));
		}
		return null;
	}
	
	/** checks if the given person is a member of the household 
	 * 
	 * @param p the person in question
	 * @return true if p is a member, false otherwise
	 */
	boolean hasMember(Person p){
		return this.members.contains(p);
	}

	/** returns the spending commitments of this household (sum of payments this 
	 * household has agreed to make, e.g. food costs or wages for piece jobs)
	 * @return the spending commitments
	 */
	public double getSpendingCommitments() {
		return spendingCommitments;
	}

	/** sets the spending commitments of this household 
	 * @param spendingCommitments the spending commitments
	 */
	public void setSpendingCommitments(double spendingCommitments) {
		double old = this.spendingCommitments;
		this.spendingCommitments = spendingCommitments;
		pcs.firePropertyChange("spendingCommitments", old, this.spendingCommitments);
	}
	
	/** checks if the given household is related to this household, i.e. if there is
	 * a kinhip link between the two households.
	 * 
	 * @param hh the household in question
	 * @return true if the households are related, false otherwise
	 */
	boolean isKinOf(Household hh){
		// check if there is a kinship link between this household and hh
		ArrayList links = this.node.getOutEdges();
		NetworkNode hhNode = hh.getNode();
		boolean found = false;
		int i = 0;
		while (!found && i < links.size()){
			NetworkLink link = (NetworkLink)links.get(i);
			found = (link.getType().equals("kinship") && link.getTo().equals(hhNode));
			i++;
		}
		return found;
	}

	/** returns a reference to the model 
	 * 
	 * @return the model
	 */
	public SouthAfricaModel getModel() {
		return model;
	}

	// methods for interface Drawable
	
	/** returns the x coordinate of the household's location
	 * @return x coordinate
	 */
	public int getX(){
		return this.x;
	}
	
	/** returns the y coordinate of the household's location
	 * @return y coordinate
	 */
	public int getY(){
		return this.y;
	}
	
	/** draws a black circle for this household on the 2D grid display */
	public void draw(SimGraphics g){
//		g.drawFastRoundRect(java.awt.Color.green);	
		g.drawFastCircle(java.awt.Color.black);
//		g.drawHollowFastOval(java.awt.Color.black);
	}
	
	/** sets the x coordinate of the household's location
	 * 
	 * @param x the x coordinate
	 */
	public void setX(int x){
		this.x = x;
	}
	
	/** sets the y coordinate of the household's location
	 * 
	 * @param y the y coordinate
	 */
	public void setY(int y){
		this.y = y;
	}
	
	/** sets the household's location on the 2D grid
	 * 
	 * @param x the x coordinate
	 * @param y the y coordinate
	 */
	public void setXY(int x, int y){
		setX(x);
		setY(y);
	}
		
}



