package org.cfpm.caves.za;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.Collections;

import jess.JessException;
import jess.RU;
import jess.Value;

/** This class represents a mine. It needs a certain number of workers, skilled and
 * unskilled. It will bring some workers with it into the area (in-migrants) and
 * advertise jobs for local people. These will be mainly unskilled jobs as the skilled
 * jobs require an engineering degree or the like.
 * Used as a shadow fact in Jess.
 * 
 * @author Ruth
 *
 */
public class Mine {
	
	//--- constants ----------------------
	
	static final double WAGE_SKILLED = 10000.0; 	// monthly wage for a skilled worker
	static final double WAGE_UNSKILLED = 900.0; 	// monthly wage for an unskilled worker
	
	//--- parameters --------------------
	
	int neededSkilled = 20;
	int neededUnskilled = 80;
	
	//--- attributes --------------------
	
	Value name;
	ArrayList<Person> workers = new ArrayList<Person>(); // list of workers, skilled in front, unskilled in back
	ArrayList<Integer> started = new ArrayList<Integer>(); // list of start ticks
	int lastSkilled = -1;
	SouthAfricaModel model;

	private PropertyChangeSupport pcs = new PropertyChangeSupport(this);
	
	/** This inner class is a comparator for Person objects. It is used to compare
	 * applicants according to their mining experience so that the mine can pick 
	 * the most experienced workers. Thus this comparator ensures a descending order.
	 */
	private class ExperienceComparator implements Comparator<Person> {
		
		/** compares two persons' mining experiences. Note that if this comparator
		 * returns 0 the mining experiences of the two persons are equal -- it doesn't
		 * necessarily mean that the two persons are equal (one and the same).
		 * 
		 * @param o1 the first object to be compared (must be of class Person)
		 * @param o2 the second object to be compared (must be of class Person)
		 * @return 1, 0 or -1 if o1's mining experience is lower, equal or greater
		 * than o2's mining experience.
		 */
		public int compare (Person o1, Person o2){
			int exp1 = ((Person)o1).getMineExperience();
			int exp2 = ((Person)o2).getMineExperience();
			if (exp1 > exp2) return -1;
			if (exp2 > exp1) return 1;
			return 0;
		}
	}
	
	private ExperienceComparator expComparator = new ExperienceComparator();

	//--- constructors ------------------
	
	/** constructs a new mine with the given parameters and adds it as a shadow
	 * fact to Jess.
	 * 
	 * @param name this mine's name
	 * @param neededSkilled the number of skilled workers needed in this mine
	 * @param neededUnskilled the number of unskilled workers needed in this mine
	 * @param model a reference to the model
	 */
	public Mine(String name, int neededSkilled, int neededUnskilled, SouthAfricaModel model){
		setName(name);
		this.neededSkilled = neededSkilled;
		this.neededUnskilled = neededUnskilled;
		this.model = model;
		// add mine as shadow fact
		try {
			model.engine.definstance("mine", this, true);
		}
		catch(JessException je) {
			System.err.println("An error occurred at line " + je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}	
	}
	
	//--- methods -----------------------
	
	/** employs the given person as a skilled worker starting from the current tick
	 * @param worker the new skilled worker
	 * @param currentTick the current tick
	 */
	public void employSkilledWorker(Person worker, int currentTick){
		workers.add(0, worker);
		started.add(0, currentTick);
		lastSkilled++;
		pcs.firePropertyChange("numSkilled", lastSkilled, lastSkilled+1);
	}
	
	/** employs the given person as an unskilled worker starting from the current tick. This also
	 * asserts an employment fact in Jess.
	 * 
	 * @param worker the new unskilled worker
	 * @param currentTick the current tick
	 */
	public void employUnskilledWorker(Person worker, int currentTick){
		if (this.workers.contains(worker)) return; // don't employ people twice
		int old = getNumUnskilled();
		workers.add(worker);
		started.add(currentTick);
		pcs.firePropertyChange("numUnskilled", old, old+1);
		// put employment fact into Jess
		// contract lengths of 12 or 18 months seem to be normal, see: http://www.teba.co.za/products_mb_other.asp#cdp
		// also QI: mine workers from Lesotho and Mozambique have part of their wages transferred to their home country
		// (so called compulsory deferred pay, see link above)
		// --> give more experienced (> 4 years) workers contracts of 18 months, others 12 months
		worker.assertEmployment(this.name, WAGE_UNSKILLED, (worker.mineExperience > 4*48 ? 18 : 12)); 
	}
	
	/** sacks the given worker from the mine (either because his contract
	 * ended or he has developed AIDS).
	 * 
	 * @param worker the worker to be sacked
	 * @param currentTick the current tick
	 */
	public void sackWorker(Person worker, int currentTick){
		int index = workers.indexOf(worker);
		if (index < 0) return; // the worker is not employed in this mine
		if (index <= lastSkilled){
			// it's a skilled worker
			pcs.firePropertyChange("numSkilled", lastSkilled+1, lastSkilled);
			lastSkilled--;
		}
		else {
			// it's an unskilled worker
			int old = getNumUnskilled();
			pcs.firePropertyChange("numUnskilled", old, old-1);
		}
		workers.remove(index);
		int startTick = started.remove(index);
		worker.addMineExperience(currentTick-startTick);	// record length of employment as experience
	}
	
	/** sacks the given worker from the mine (either because his contract
	 * ended or he has developed AIDS).
	 * 
	 * @param worker the name of the worker to be sacked
	 * @param currentTick the current tick
	 */
	public void sackWorker(String worker, int currentTick){
		sackWorker(model.namePersonMap.get(worker), currentTick);
	}
	
	/** returns the number of skilled workers needed to operate this mine.
	 * 
	 * @return number of needed skilled workers
	 */
	public int getNeededSkilled(){
		return this.neededSkilled;
	}
	
	/** returns the number of unskilled workers needed to operate this mine.
	 * 
	 * @return number of needed unskilled workers
	 */
	public int getNeededUnskilled(){
		return this.neededUnskilled;
	}
	
	/** returns the number of skilled workers currently employed at the mine.
	 * 
	 * @return number of skilled workers
	 */
	public int getNumSkilled(){
		return (this.lastSkilled + 1);
	}
	
	/** returns the number of unskilled workers currently employed at the mine.
	 * 
	 * @return number of unskilled workers
	 */
	public int getNumUnskilled(){
		return (this.workers.size() - lastSkilled - 1);
	}
	
	/** sets the mine's name
	 * 
	 * @param nameString name of the mine
	 */
	public void setName(String nameString){
		try {
			name = new Value(nameString, RU.SYMBOL);
		}
		catch(JessException je){
			System.err.println("An error occurred at line " + je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}
	}
	
	/** returns the name of the mine
	 * 
	 * @return the mine's name
	 */
	public Value getName(){
		return this.name;
	}

	
	/** picks the best employees from a list of applicants
	 * prefer experienced men aged 23-36, allow for 5% women (not yet implemented)
	 * @param applicants list of persons' names
	 * @param currentTick the current model tick
	 */
	public void pickUnskilledWorkers(String[] list, int currentTick){
		// retrieve the person objects
		ArrayList<Person> applicants = new ArrayList<Person>();
		for (String name : list){
			applicants.add(model.namePersonMap.get(name));
		}
		// filter out males of the desired age group
		ArrayList<Person> rightAge = ageFilter(applicants);
		// filter out experienced workers
		ArrayList<Person> best = experienceFilter(rightAge);
		// check if there are more applicants than jobs
		int demand = neededUnskilled - getNumUnskilled();
		if (best.size() > demand){
			// shorten list of best applicants accordingly
			System.out.println("more good applicants then demand: " + best.size() + " > " + demand);
			while (best.size() > demand){
//				best.remove(model.randomInt(0, best.size()-1)); // remove random applicant
				best.remove(best.size()-1); // remove least best applicant
			}
		}
		int numBest = best.size();
		model.dump("employ the " + numBest + " best applicants");
		
		// if there are fewer experienced applicants than demand employ some unexperienced ones
		demand = demand - numBest;
		int numUnexperienced = Math.min((getNumUnskilled()+numBest)/10, demand);	// allow up to 10% of employed workers
		while (numUnexperienced > 0 && (rightAge.size()-best.size()) > 0){
			// pick an unexperienced worker of the right age
			int i = model.randomInt(0, rightAge.size()-1);
			while (best.contains(rightAge.get(i))){
				i = model.randomInt(0, rightAge.size()-1);
			}
			best.add(rightAge.get(i));
			rightAge.remove(i);
			numUnexperienced--;
		}
		if (best.size() > numBest){
			model.dump("also employ " + (best.size()-numBest) + " unexperienced applicants");
		}
		// employ the chosen applicants
		for (Person applicant : best){
			employUnskilledWorker(applicant, currentTick);
		}
	}
	
	/** internal helper function: filters the experienced applicants from the given list of applicants
	 * 
	 * @param applicants list of applicants
	 * @return list of experienced applicants sorted in descending order of experience level
	 */
	private ArrayList<Person> experienceFilter(ArrayList<Person> applicants){
		ArrayList<Person> chosen = new ArrayList<Person>();
		for (Person applicant : applicants){
			if (applicant.getMineExperience() > 0){
				chosen.add(applicant);
			}
		}
		// sort them in descending order of experience level
		return sort(chosen);
	}
	
	/** internal helper function: sorts the given list of persons in descending order 
	 * of experience level (applying the experience comparator)
	 * 
	 * @param list list of persons
	 * @return sorted list of persons
	 */
	private ArrayList<Person> sort(ArrayList<Person> list){
		Collections.sort(list, expComparator);
		return list;
	}
	
	/** internal helper function: filters the male applicants of the desired age group 
	 * from the given list of applicants
	 * 
	 * @param applicants list of applicants
	 * @return list of male applicants of the right age
	 */
	private ArrayList<Person> ageFilter(ArrayList<Person> applicants){
		ArrayList<Person> chosen = new ArrayList<Person>();
		for (Person applicant : applicants){
			if (/*applicant.getAge() >= 23 && applicant.getAge() <= 36*/ applicant.getAge() <= 50
					&& applicant.getGender().toString().equals("male")) {
				chosen.add(applicant);
			}
		}
		return chosen;
	}
	
	/** adds a property change listener (for Jess) */
	public void addPropertyChangeListener(PropertyChangeListener p) {
		pcs.addPropertyChangeListener(p);
	}

	/** removes a property change listener (for Jess) */
	public void removePropertyChangeListener(PropertyChangeListener p) {
		pcs.removePropertyChangeListener(p);
	}	
	
	/** returns the number of workers who are in-migrants.
	 * @return number of in-migrant workers employed at the mine
	 */
	public int getNumberOfMigrantWorkers(){
		int sum = 0;
		for (Person miner : workers){
			if (miner.getName().toString().startsWith("miner")){
				sum++;
			}
		}
		return sum;
	}

}
