package org.cfpm.caves.za;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;

import uchicago.src.sim.network.DefaultDrawableEdge;

import jess.*;

/** Simple directed link between network nodes. 
 *  Instead of subclasses for different types, the type attribute inherited from DefaultEdge
 *  is used to express different link types --> different network layers. 
 * 
 * @author Ruth Meyer
 *
 */
public class NetworkLink extends DefaultDrawableEdge {
	
	//--- link attributes --------------------------------------------
	
	private PropertyChangeSupport pcs = new PropertyChangeSupport(this);

	
	//--- constructors --------------------------------------------------
	
	/** Parameterless constructor. Creates a new "empty" network link.
	 */
	public NetworkLink() {
		super();
	}
	
	/** Creates a network link between the two given nodes.
	 * 
	 * @param from the from node
	 * @param to the to node
	 */
	public NetworkLink(NetworkNode from, NetworkNode to) {
		super(from, to);
		// and set connections in the nodes!
		from.addOutEdge(this);
		to.addInEdge(this);
	}
	
	/** Creates a network link of the given type between the two given nodes.
	 * 
	 * @param from the from node
	 * @param to the to node
	 * @param type the link type
	 */
	public NetworkLink(NetworkNode from, NetworkNode to, String type){
		this(from, to);
		setType(type);
	}
	
	/** returns the from node.
	 * @return the from node
	 */
	public NetworkNode getFrom(){
		return (NetworkNode)super.getFrom();
	}
	
	/** sets the from node to the given node.
	 * 
	 * @param from the from node
	 */
	public void setFrom(NetworkNode from){
		NetworkNode old = getFrom();
		super.setFrom(from);
		pcs.firePropertyChange("from", old, from);
		pcs.firePropertyChange("fromID", old.getId(), from.getId());
	}
	
	/** returns the to node.
	 * @return the to node
	 */
	public NetworkNode getTo(){
		return (NetworkNode)super.getTo();
	}
	
	/** sets the to node to the given node.
	 * 
	 * @param to the from node
	 */
	public void setTo(NetworkNode to){
		NetworkNode old = getTo();
		super.setTo(to);
		pcs.firePropertyChange("to", old, to);
		pcs.firePropertyChange("toID", old.getId(), to.getId());
	}
	
	/** sets the link type
	 * @param type the link type
	 */
	public void setType(String type) {
		String old = this.type;
		this.type = type;
		pcs.firePropertyChange("type", old, this.type);
	}
	
	/** returns the node id of the from node. For easier access in Jess.
	 * 
	 * @return id of the from node
	 */
	public Value getFromID(){	// for easier access in Jess
		return getFrom().getId();
	}
	
	/** returns the node id of the to node. For easier access in Jess.
	 * 
	 * @return id of the to node
	 */
	public Value getToID(){		// for easier access in Jess
		return getTo().getId();
	}
	
	
	/** returns a String representation of this link
	 * @return String representation 
	 */
	public String toString(){
		return "Link from " + from.toString() + " to " + to.toString();
	}
	
	
	/** adds a property change listener (for Jess) */
	public void addPropertyChangeListener(PropertyChangeListener p) {
		pcs.addPropertyChangeListener(p);
	}

	/** removes a property change listener (for Jess) */
	public void removePropertyChangeListener(PropertyChangeListener p) {
		pcs.removePropertyChangeListener(p);
	}
	

}
