package org.cfpm.caves.za;

import uchicago.src.sim.network.DefaultDrawableNode;
import uchicago.src.sim.gui.OvalNetworkItem;
import jess.*;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.util.ArrayList;


/** A node in a (social) network; refers to a person 
 *  or household.
 *  Was originally planned to be used as shadow fact
 *  in Jess, together with the link class, so that
 *  all networks would be represented as nodes and links
 *  to be simultaneously "visible" in Jess and Java and
 *  displayable in Repast.
 *  Alas, it turned out that this wasn't the most 
 *  efficient way to represent networks so it was
 *  more or less completely dropped. Persons and households
 *  still have nodes and the kinship network and friend network
 *  are maintained as links if the GUI is switched on.
 *  
 * @author Ruth Meyer
 *
 */
public class NetworkNode extends DefaultDrawableNode {
	
	//--- node attributes --------------------------------------------
	
	/** the node id */
	Value id;
	
	private PropertyChangeSupport pcs = new PropertyChangeSupport(this);

	
	//--- constructors --------------------------------------------------
	
	/** Parameterless constructor. Creates a new "empty" network node.
	 */
	public NetworkNode() {
		init();
	}
	
	/** Creates a new network node with the given id. This is either 
	 * the associated person's name or the associated household's id.
	 * 
	 * @param id the node id
	 */
	public NetworkNode(Value id){
		setId(id);
		init();
	}

	//--- methods -------------------------------------------------------
	
	/** internal function: creates the graphical representation of the node
	 */
	private void init(){
		OvalNetworkItem d = new OvalNetworkItem(10,10);
		setDrawable(d);
	}
	
	/** sets the node id
	 * 
	 * @param newId the node's new id
	 */
	public void setId(Value newId){
		Value old = id;
		id = newId;
		pcs.firePropertyChange("id", old, id);
		// pass an abbreviation of it on as the node label
		this.setNodeLabel(abbreviate(id));
	}
	
	/** sets the node id
	 * 
	 * @param newId the node's new id as a String
	 */
	public void setId(String newId){
		try {
			Value newID = new Value(newId, RU.SYMBOL);
			this.setId(newID);
		}
		catch(JessException je) {
			System.err.println("An error occurred at line " + je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}
	}
	
	/** internal helper function: abbreviates the node id to be used as the node label */
	private String abbreviate(Value id){
		// to make the nodes smaller in the display, abbreviate "person" to "p" and "household" to "h"
		String idString = id.toString();
		StringBuffer abbrv = new StringBuffer();
		abbrv.append(idString.substring(0, 1));
		abbrv.append(idString.substring(idString.indexOf('-')));
		return abbrv.toString();
	}
	
	/** returns the node id
	 * @return the node id
	 */
	public Value getId() {
		return id;
	}
	
	/** adds the given link as an incoming edge
	 * 
	 * @param link the network link
	 */
	public void addInEdge(NetworkLink link){
		ArrayList old = getFromNodes();
		super.addInEdge(link);
		pcs.firePropertyChange("fromNodes", old, getFromNodes());
	}
	
	/** adds the given link as an outgoing edge
	 * 
	 * @param link the network link
	 */
	public void addOutEdge(NetworkLink link){
		ArrayList old = getToNodes();
		super.addOutEdge(link);
		pcs.firePropertyChange("toNodes", old, getToNodes());
	}
	
	/** adds a property change listener (for Jess) */
	public void addPropertyChangeListener(PropertyChangeListener p) {
		pcs.addPropertyChangeListener(p);
	}

	/** removes a property change listener (for Jess) */
	public void removePropertyChangeListener(PropertyChangeListener p) {
		pcs.removePropertyChangeListener(p);
	}
	
	/** returns a String representation of this node
	 * @return String made out of the word "Node" and the node's label
	 */
	public String toString(){
		return "Node---" + this.label;
	}
	
	/** Compare the given object to this node. Needed for tests in contains() operation of node lists.
	 * @param o the object to compare with
	 * @return true if the given object is (a) identical with this node or (b) of type NetworkNode
	 * and has the same id; false otherwise
	 */
	public boolean equals(Object o){
		if (this == o) return true;  // identity
		if ((o == null) || !(o instanceof NetworkNode)) return false; // null reference or not a network node
		NetworkNode otherNode = (NetworkNode)o;
		if (this.id != null && this.id == otherNode.getId()) return true; // refers to same agent
		return false;
	}
	
	/** removes the outgoing links of the specified type
	 * @param type the link type
	 */
	public void removeFromLinks(String type){
		// run through list of out edges and remove the ones of the given type
		ArrayList<NetworkLink> toBeDeleted = new ArrayList<NetworkLink>();
		for (Object o : getOutEdges()){
			NetworkLink link = (NetworkLink)o;
			if (link.getType().equals(type)){
				toBeDeleted.add(link);
			}
		}
		for (NetworkLink link : toBeDeleted){
			removeOutEdge(link);
			link.getTo().removeInEdge(link);
		}
	}

}
