
package org.cfpm.caves.za;

import jess.*;
import java.util.ArrayList;
import java.beans.*;

import uchicago.src.sim.util.Random;

/** An individual person. Used as a shadow fact in Jess.
 * @author Ruth Meyer, Scott Moss
 *
 */
public class Person {

	//--- constants ---------------------------------------------------------
	
	public static final int CHILD_AGE_LIMIT = 16;
	public static final int ADULT_AGE_LIMIT = 55; 
	public static final String[] maritalStates = {"single", "married", "separated", "widowed"};
	public static final String[] educationLevels = {"none", "some-primary", "primary", "some-secondary", "secondary", "tertiary"};
	// School in ZA is compulsory from 7-15 (up to grade 9)
	// http://hrw.org/reports/2004/southafrica0504/6.htm
	// Grade 1 - 7 = primary, grade 8 - 12 = secondary
	private static String[] positiveEndorsementTokenStrings = {"reliable", "trustworthy", "capable", "honest",
															   "is-neighbour", "is-kin", "is-friend", 
															   "most-similar", "similar",  "recommended",
															   "same-church", "same-denomination",
															   "same-club", "same-work",	// added after field-trip
															   "has-job"	// for sexual partner choice
															   }; // deleted: "is-acquaintance"
	private static String[] negativeEndorsementTokenStrings = {"unreliable", "untrustworthy", "incapable", "dishonest"};
	
	//--- agent attributes ---------------------------------------------------
	
	Value name;	
	int age;	
	int birthMonth, birthWeek, deathWeek = -1;
	Value gender;
	Value maritalStatus;
	Value achievedEducation;
	boolean head = false;
	boolean alive = true;
	boolean migrant = false; boolean onVisit = false;
	int lastVisit = -1;
	boolean registeredForPension = false;
	boolean registeredForChildGrant = false;
	boolean atSchool = false;
	int lastDelivery = -1; // records tick of last delivery of a baby for females
	public int maxNumFriends, maxNumPartners;
	HIVAIDS hivAIDS = null;
	boolean left = false;
	int lastJob = -1; // records tick of last job end so that miners can decide to leave if they don't get another job
					  // after a certain amount of time
	int mineExperience = 0;	// counts number of weeks spent in mine employment so that the mine can prefer experienced
							// workers when employing
	SouthAfricaModel model = null;
	
	// relations to other agents
	
	Household household;
	Person spouse;
	Value father, mother;
	ArrayList<Value> children = new ArrayList<Value>();
	ArrayList<Value> siblings = new ArrayList<Value>();
	ArrayList<Value> friends = new ArrayList<Value>();
	ArrayList<Value> knownPersons = new ArrayList<Value>();
	ArrayList<Person> currentPartners = new ArrayList<Person>();
	ArrayList<Integer> partnershipStart = new ArrayList<Integer>();
	NetworkNode node = null;
	
	// groups
	
	ArrayList<BurialSociety> burialSocieties = new ArrayList<BurialSociety>();
	ArrayList<Stokvel> stokvels = new ArrayList<Stokvel>();
	
	// endorsement related
	
	private int numEndorsementClasses;
	EndorsementScheme agentEndorsementScheme, friendEndorsementScheme;
	
	// tag related
	
	int[] tag;
	public double tagEvolutionPropensity;
	
	// Jess related	
	
	Rete engine = null;
	private PropertyChangeSupport pcs = new PropertyChangeSupport(this);
	

	//--- Constructors ---------------------------------------------------------------
	
	/** Creates a new person with the given name
	 * @param nameString name of the person
	 * @param hhold household of the person
	 * @param model the model
	 */
	public Person(String nameString, Household hhold, SouthAfricaModel model){
		setName(nameString);
		household = hhold;
		this.model = model;
		engine = model.engine;
		double endSchemeBase = Random.uniform.nextDoubleFromTo(model.minEndorsementBase, model.maxEndorsementBase);
		numEndorsementClasses = Random.uniform.nextIntFromTo(model.minEndorsementClasses, model.maxEndorsementClasses);
		addAgentEndorsementScheme(endSchemeBase);
		createAgentEndorsementTokens();
		// special friend endorsement scheme doesn't have much influence on the number of reciprocal links
//		addFriendEndorsementScheme(endSchemeBase);
//		createFriendEndorsementTokens();
		// max number of friends influences the number of reciprocal links 
		setMaxNumFriends(Random.uniform.nextIntFromTo(model.lowerMaxNumFriends, model.upperMaxNumFriends)); 
		setMaxNumPartners(Random.uniform.nextIntFromTo(model.lowerMaxNumPartners, model.upperMaxNumPartners));
		setAchievedEducation(0);
		setMaritalStatus("single");
	}
	
	
	//--- methods -------------------------------------------------------------------------------------------
	
	/** Creates this person's tag, i.e. characteristic traits modelled as a list of digits
	 */
	public void createTag(){
		int length = model.tagLength;
		int base = model.tagBase;
		int[] initTag = new int[length];
		for (int i = 0; i < length; i++)
			initTag[i] = Random.uniform.nextIntFromTo(0, base - 1);
		setTag(initTag);
	}
	
	/** Creates the contents of the agent's endorsement scheme, 
	 *  i.e. creates particular weights for the endorsement tokens.
	 */
	private void createAgentEndorsementTokens() {
		Value token;
		int endorsementValue;
		try {
			for (String tokenString : positiveEndorsementTokenStrings){
				token = new Value(tokenString, RU.SYMBOL);
				endorsementValue = Random.uniform.nextIntFromTo(1, numEndorsementClasses);
				agentEndorsementScheme.addEndorsement(token, endorsementValue);
			}
			for (String tokenString : negativeEndorsementTokenStrings){
				token = new Value(tokenString, RU.SYMBOL);
				endorsementValue = 
					-1 * Random.uniform.nextIntFromTo(1, numEndorsementClasses);
				agentEndorsementScheme.addEndorsement(token, endorsementValue);
			}
		}catch(JessException je) {
			System.err.println("An error occurred at line "
					+ je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}	
	}

	/** Creates the contents for the friend endorsement scheme that ranks "most-similar" and "similar" highest
	 */
	private void createFriendEndorsementTokens() {
		// make sure that the endorsements "most-similar" and "similar" rank highest
		Value token;
		int endorsementValue;
		try {
			for (String tokenString : positiveEndorsementTokenStrings){
				token = new Value(tokenString, RU.SYMBOL);
				if (tokenString.endsWith("similar")){
					endorsementValue = numEndorsementClasses;
				}
				else {
					endorsementValue = Random.uniform.nextIntFromTo(1, numEndorsementClasses);
				}
				friendEndorsementScheme.addEndorsement(token, endorsementValue);
			}
			for (String tokenString : negativeEndorsementTokenStrings){
				token = new Value(tokenString, RU.SYMBOL);
				endorsementValue = 
					-1 * Random.uniform.nextIntFromTo(1, numEndorsementClasses);
				friendEndorsementScheme.addEndorsement(token, endorsementValue);
			}
		}catch(JessException je) {
			System.err.println("An error occurred at line "
					+ je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}	
	}
	
	/** Creates an empty endorsement scheme for the agent
	 * 
	 * @param base	the base
	 */
	private void addAgentEndorsementScheme(double base){
		try {
			Value agentEndorsementSchemeName = 
				new Value("agentEndorsementScheme", RU.SYMBOL);
			setAgentEndorsementScheme(new EndorsementScheme(agentEndorsementSchemeName, base));
		}catch(JessException je) {
			System.err.println("An error occurred at line "
					+ je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}
	}
	
	/** Creates an empty friend endorsement scheme for the agent
	 * 
	 * @param base	the base
	 */
	private void addFriendEndorsementScheme(double base){
		try {
			Value friendEndorsementSchemeName = 
				new Value("friendEndorsementScheme", RU.SYMBOL);
			setFriendEndorsementScheme(new EndorsementScheme(friendEndorsementSchemeName, base));
		}catch(JessException je) {
			System.err.println("An error occurred at line "
					+ je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}
	}
	
	/** sets the friend endorsement scheme and propagates change to Jess
	 * 
	 * @param es the endorsement scheme
	 */
	public void setFriendEndorsementScheme(EndorsementScheme es){
		EndorsementScheme old = friendEndorsementScheme;
		friendEndorsementScheme = es;
		pcs.firePropertyChange("friendEndorsementScheme", old, friendEndorsementScheme);
	}
	
	/** returns the friend endorsement scheme
	 * 
	 * @return the endorsement scheme
	 */
	public EndorsementScheme getFriendEndorsementScheme(){
		return friendEndorsementScheme;
	}

	/** sets the agent endorsement scheme and propagates change to Jess
	 * 
	 * @param es the endorsement scheme
	 */
	public void setAgentEndorsementScheme(EndorsementScheme es){
		EndorsementScheme old = agentEndorsementScheme;
		agentEndorsementScheme = es;
		pcs.firePropertyChange("agentEndorsementScheme", old, agentEndorsementScheme);
	}
	
	/** returns the agent endorsement scheme
	 * 
	 * @return the endorsement scheme
	 */
	public EndorsementScheme getAgentEndorsementScheme(){
		return agentEndorsementScheme;
	}
	
	/** Updates the agent's tag. Agents tend to become more similar to their best friends.
	 * 
	 * @param friends list of friends
	 * @param endorsementValues list of friends' overall endorsement values
	 */
	public void updateTag(Person[] friends, double[]endorsementValues){
		int[] evolvedTag = new int[tag.length];
		double max = 0.0;
		ArrayList<Integer> maxN = new ArrayList<Integer>();
		for (int i = 0; i < friends.length; i++){
			if (endorsementValues[i] > max){
				max = endorsementValues[i];
				maxN = new ArrayList<Integer>();
				maxN.add(i);
			}
			else {
				if (endorsementValues[i] == max){
					maxN.add(i);
				}
			}
		}
		for (Integer i : maxN){
			Person friend = friends[i];
			int[] targetTag = friend.tag;
			for (int n = 0; n < tag.length; n++){
				if (Random.uniform.nextDoubleFromTo(0, 1.0)< tagEvolutionPropensity){
					if (tag[n] > targetTag[n]) evolvedTag[n] = tag[n] - 1;
					else 
						if (tag[n] < targetTag[n]) evolvedTag[n] = tag[n] + 1;
				}
				else evolvedTag[n] = tag[n];
			}
			
		}
		setTag(evolvedTag);
	}
	
	/** Assert an employment fact in the Jess fact base to show that this agent has
	 * been employed by the given employer for the specified wage and time.
	 * 
	 * @param employer name of the agent's employer
	 * @param wage	monthly wage
	 * @param durationInMonths duration of employment in months
	 */
	public void assertEmployment(Value employer, double wage, int durationInMonths){
		setLastJob(model.tick + durationInMonths*4);
		try {
			Fact f = new Fact("employment", engine);
			f.setSlotValue("employee", this.name);
			f.setSlotValue("employer", employer);
			f.setSlotValue("wage", new Value(wage, RU.FLOAT));
			f.setSlotValue("start-date", new Value(model.tick, RU.INTEGER));
			f.setSlotValue("end-date", new Value(model.tick + durationInMonths*4, RU.INTEGER));
			engine.assertFact(f);
			model.dump(f.toStringWithParens());
		} catch (JessException je) {
			System.err.println("An error occurred at line " + je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}
	}

	/** Assert an employment fact in the Jess fact base to show that this agent has
	 * been employed by the given employer for the specified wage and time.
	 * 
	 * @param employer name of the agent's employer as String
	 * @param wage	monthly wage
	 * @param durationInMonths duration of employment in months
	 */
	public void assertEmployment(String employer, double wage, int durationInMonths){
		try {
			assertEmployment(new Value(employer, RU.SYMBOL), wage, durationInMonths);
		} catch (JessException je) {
			System.err.println("An error occurred at line " + je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}
	}

	/** returns the agent's tag
	 * @return the tag
	 */
	public int[] getTag() {
		return tag;
	}

	/** sets the agent's tag
	 * 
	 * @param tag the new tag
	 */
	public void setTag(int[] tag) {
		int[]old = this.tag;
		this.tag = tag;
		pcs.firePropertyChange("tag", old, this.tag);
	}

	/** returns the agent's age
	 * 
	 * @return the age
	 */
	public int getAge() {
		return age;
	}

	/** sets the agent's age
	 * 
	 * @param age the age
	 */
	public void setAge(int age) {
		int old = this.age;
		this.age = age;
		pcs.firePropertyChange("age", old, this.age);
	}

	/** returns the agent's gender
	 * 
	 * @return the gender
	 */
	public Value getGender() {
		return gender;
	}

	/** sets the agent's gender
	 * 
	 * @param gender the gender as a Value object
	 */
	public void setGender(Value gender) {
		Value old = this.gender;
		this.gender = gender;
		pcs.firePropertyChange("gender", old, this.gender);
	}
	
	/** sets the agent's gender
	 * 
	 * @param genderString the gender as a String
	 */
	public void setGender(String genderString){
		try {
			setGender(new Value(genderString, RU.SYMBOL));
		}
		catch(JessException je) {
			System.err.println("An error occurred at line " + je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}
	}

	/** returns the agent's maximal number of friends
	 * 
	 * @return upper limit on the number of friends
	 */
	public int getMaxNumFriends() {
		return maxNumFriends;
	}

	/** sets the agent's maximal number of friends
	 * 
	 * @param n upper limit on the number of friends
	 */
	public void setMaxNumFriends(int n) {
		int old = maxNumFriends;
		maxNumFriends = n;
		pcs.firePropertyChange("maxNumFriends", old, maxNumFriends);
	}

	/** returns the agent's maximal number of sexual partners
	 * 
	 * @return upper limit on the number of partners
	 */
	public int getMaxNumPartners() {
		return maxNumPartners;
	}

	/** sets the agent's maximal number of sexual partners
	 * 
	 * @param n upper limit on the number of partners
	 */
	public void setMaxNumPartners(int n) {
		int old = maxNumPartners;
		maxNumPartners = n;
		pcs.firePropertyChange("maxNumPartners", old, maxNumPartners);
	}
	
	/** returns the agent's birth month
	 * 
	 * @return month the agent was born as int between 0 and 11
	 */
	public int getBirthMonth() {
		return birthMonth;
	}

	/** sets the agent's birth month
	 * 
	 * @param birthMonth month the agent was born as int between 0 and 11
	 */
	public void setBirthMonth(int birthMonth) {
		int old = this.birthMonth;
		this.birthMonth = birthMonth;
		pcs.firePropertyChange("birthMonth", old, this.birthMonth);
	}
	
	/** returns the agent's birth week
	 * 
	 * @return week the agent was born as int between 0 and 3
	 */
	public int getBirthWeek() {
		return birthWeek;
	}

	/** sets the agent's birth week
	 * 
	 * @param birthWeek week the agent was born as int between 0 and 3
	 */
	public void setBirthWeek(int birthWeek) {
		int old = this.birthWeek;
		this.birthWeek = birthWeek;
		pcs.firePropertyChange("birthWeek", old, this.birthWeek);
	}

	/** returns the agent's death week
	 * 
	 * @return week the agent died as int between 0 and 3
	 */
	public int getDeathWeek() {
		return deathWeek;
	}

	/** sets the agent's death week
	 * 
	 * @param deathWeek week the agent died as int between 0 and 3
	 */
	public void setDeathWeek(int deathWeek) {
		int old = this.deathWeek;
		this.deathWeek = deathWeek;
		pcs.firePropertyChange("deathWeek", old, this.deathWeek);
	}
		
	/** returns the agent's name
	 * 
	 * @return name as a Value object
	 */
	public Value getName() {
		return name;
	}

	/** sets the agent's name
	 * 
	 * @param name name as a Value object
	 */
	public void setName(Value name) {
		Value old = this.name;
		this.name = name;
		pcs.firePropertyChange("name", old, this.name);
	}
	
	/** sets the agent's name
	 * 
	 * @param nameString name as a String
	 */
	public void setName(String nameString){
		try {
			setName(new Value(nameString, RU.SYMBOL));	
		}
		catch(JessException je) {
			System.err.println("An error occurred at line " + je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}
	}
	
	/** returns this agent's spouse 
	 * 
	 * @return spouse or null if the agent isn't married
	 */
	public Person getSpouse() {
		return spouse;
	}

	/** sets the agent's spouse
	 * 
	 * @param spouse the spouse
	 */
	public void setSpouse(Person spouse) {
		Person old = this.spouse;
		this.spouse = spouse;
		pcs.firePropertyChange("spouse", old, this.spouse);
	}
	
	/** sets the agent's marital status
	 * 
	 * @param ms marital status as a Value object
	 */
	public void setMaritalStatus(Value ms){
		checkMaritalStatusValidity(ms);
		Value old = maritalStatus;
		maritalStatus = ms;
		pcs.firePropertyChange("maritalStatus", old, maritalStatus);
	}
	
	/** sets the agent's marital status
	 * 
	 * @param n marital status as index into the list of marital states
	 */
	public void setMaritalStatus(int n){
		Value old = maritalStatus;
		try {maritalStatus = new Value(maritalStates[n], RU.SYMBOL);}
		catch(JessException je) {
			System.err.println("An error occurred at line " + je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}
		pcs.firePropertyChange("maritalStatus", old, maritalStatus);
	}
	
	/** sets the agent's marital status
	 * 
	 * @param msString marital status as String
	 */
	public void setMaritalStatus(String msString){
		try {setMaritalStatus(new Value(msString, RU.SYMBOL));}
		catch(JessException je) {
			System.err.println("An error occurred at line " + je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}
	}
	
	/** returns the agent's marital status
	 * 
	 * @return marital status as a Value
	 */
	public Value getMaritalStatus(){
		return maritalStatus;
	}
	
	/** internal helper function: checks if the given Value is a valid marital status
	 * 
	 * @param ms the marital status as a Value
	 * @return true if ms is a valid marital status, otherwise false
	 */
	private boolean checkMaritalStatusValidity(Value ms){
		boolean flag = false;
		for (String status : maritalStates){
			if (ms.toString()== status) {
				flag = true;
				break;
			}
		}
		if (flag = false) {
			System.err.println(ms + "is not a valid marital status");
			System.exit(1);
		}
		return flag;
	}

	/** returns the agent's life state
	 * 
	 * @return true if the agent is alive, false otherwise
	 */
	public boolean isAlive(){
		return this.alive;
	}
	
	/** sets the agent's life state
	 * 
	 * @param alive true for alive, false for dead
	 */
	public void setAlive(boolean alive){
		boolean old = this.alive;
		this.alive = alive;
		pcs.firePropertyChange("alive", old, this.alive);			
	}

	/** returns if the agent is head of its household
	 * 
	 * @return true if the agent is household head, false otherwise
	 */
	public boolean isHead(){
		return this.head;
	}
	
	/** sets if the agent is head of its household
	 * 
	 * @param hhHead true if the agent is household head, false otherwise
	 */
	public void setHead(boolean hhHead){
		boolean old = this.head;
		this.head = hhHead;
		// change it in household too
		this.household.setHead(this);
		pcs.firePropertyChange("head", old, this.head);	
		// check if child-headed household
		if (this.age <= 17){
			//*//model.dump(this.household.getId() + " is now child-headed");
			if (atSchool){
				// drop out of school
				setAtSchool(false);
			}
		}
	}
	
	/** returns the ID of this agent's household
	 * 
	 * @return household ID as a Value object
	 */
	public Value getHouseholdID() {
		if (household == null) return null;
		return household.getId();
	}

	/** returns this agent's household
	 * 
	 * @return the household
	 */
	public Household getHousehold() {
		return household;
	}

	/** sets this agent's household. 
	 * This also adds the agent as a member to the given household.
	 * 
	 * @param household the household
	 */
	public void setHousehold(Household household) {
		Household old = this.household;
		Value oldID = (old == null)? null : old.getId();
		this.household = household;
		// add this person as a member to the household
		this.household.addMember(this);
		pcs.firePropertyChange("household", old, this.household);
		pcs.firePropertyChange("householdID", oldID, this.household.getId());
	}
	
	/** return this agent's network node
	 * 
	 * @return the node
	 */
	public NetworkNode getNode(){
		return this.node;
	}
	
	/** sets this agent's network node
	 * 
	 * @param node the node
	 */
	public void setNode(NetworkNode node){
		NetworkNode old = this.node;
		this.node = node;
		pcs.firePropertyChange("node", old, this.node);
	}
	
	/** returns a reference to the model
	 * 
	 * @return the model
	 */
	public SouthAfricaModel getModel(){
		return this.model;
	}
	
	/** sets a reference to the model
	 * 
	 * @param model the model
	 */
	public void setModel(SouthAfricaModel model){
		this.model = model;
	}
	
	/** adds a property change listener (for Jess) */
	public void addPropertyChangeListener(PropertyChangeListener p) {
		pcs.addPropertyChangeListener(p);
	}

	/** removes a property change listener (for Jess) */
	public void removePropertyChangeListener(PropertyChangeListener p) {
		pcs.removePropertyChangeListener(p);
	}	
	
	/** updates the agent for the current tick. This includes ageing, checking for death,
	 * determining HIV infection and updating education levels.
	 * 
	 * @param currentTime the current model tick
	 */
	public void update(int currentTime){
		// do we have a birthday this week?
		if ((model.getMonth() == (this.birthMonth)) && (model.getWeek() == (this.birthWeek))){
			setAge(this.age+1);
//			// check for probability to die with current age
//			model.checkDeath(this);
			// check if we register for a pension
			if (this.age > 55 && model.askPensionOracle(this))
				setRegisteredForPension(true);
			// check if we're too old for the child grant
			if (this.age > model.childGrantAgeLimit){
				setRegisteredForChildGrant(false);
			}
		}
		// check if we die this week
		// -- due to "natural" death (according to WHO life tables)
		// -- or due to illness (see health status) or starving (see feed status)
		
		if (this.deathWeek == currentTime){
			die(currentTime);
		}	
		// check education when the new school year begins (January)
		if ((model.getMonth() == 0) && model.getWeek() == 1){
			if (!atSchool && (age >= 6) && (age <= 13)){
				setAtSchool(true);	// start school
				setAchievedEducation(1);
			}
			else if (atSchool){
				switch (age){
				case 13 : setAchievedEducation(2); break; // finished primary school
				case 14 : setAchievedEducation(3); break; // some secondary school
				case 18 : setAchievedEducation(4); setAtSchool(false); break; // finished secondary school
				// case 20 : setAchievedEducation(5); break; // at least some tertiary education --> done in Jess
				}
				// let them finish uni between 22 and 25
				if (age >= Random.uniform.nextIntFromTo(22,25)){
					setAtSchool(false);
					setMigrant(false);
					setOnVisit(false);
					model.dump(name + " finished college/uni at age " + age);
				}
			}
		}
		
		// check if we contract HIV 
		if (!isHIVPositive()){
			// at least one of the sexual partners has to be HIV positive
			// higher risk when on migration
			// higher risk when there is only one partner who is HIV positive
			// needs to be influenced by frequency of sexual intercourse --> modelled through tags?? ### needs to be split between partners, really
			if (currentPartners.size() > 0 || spouse != null){			
				boolean infected = false;
				for (Person p : currentPartners){
					if (p.isHIVPositive()){
						infected |= (model.probabilityp(p.hivAIDS.getTransmissionProbability(currentTime)*tag[2]));		
//						infected |= (model.probabilityp(0.003*tag[2]));		// #### which numbers to choose??
								// "The rate of HIV transmission with sexual intercourse is ... approximately 0.3% per sexual contact with an HIV-infected person"
								// http://library.med.utah.edu/WebPath/TUTORIAL/AIDS/HIV.html
					}
				}
				if (spouse != null && spouse.isHIVPositive()){
					infected |= (model.probabilityp(spouse.hivAIDS.getTransmissionProbability(currentTime)*tag[2]));			// #### which numbers to choose??
				}	
				if (infected) setInfectedSince(currentTime);
			}
			else if (this.migrant && this.age >= 14 || this.name.toString().startsWith("miner")){
				// we don't yet model partners of migrants outside the village, so just a probability here
				// higher risk when on migration or an in-migrant miner
				if (model.probabilityp(0.003*tag[2]*3)){			// #### which numbers to choose??
					setInfectedSince(currentTime);
				}
			}
		}
		else {
			// check if we have developed AIDS
			if (this.hivAIDS.getDevelopedAIDS() == currentTime){
				pcs.firePropertyChange("hasAIDS", false, true);
			}
		}
		
		// a miner leaves the area if he's without job for x months or has developed AIDS

		// source for the latter: "[there is a] current trend of repatriating incapacitated mineworkers to their homes"
		// with incapacitated meaning "injured ex-mineworkers that have been rehabilitated and mineworkers with terminal illnesses"
		// see http://www.teba.co.za/products_mineworkerbenefits.asp
		if (name.toString().startsWith("miner")) {
			if (getHasAIDS() || 
				(lastJob > 0 && (currentTime - lastJob)/4 > model.randomInt(4,6)))
			{
				left = true; // mark miner for deletion from the model in model.updateAgents()
				model.dump(name + " leaves the area because he " + (getHasAIDS() ? "developed AIDS" : "can't find a job"));
				// record leave in Jess
				try {
					Fact leftFact = new Fact("leave", model.engine);
					leftFact.setSlotValue("miner", this.name);
					leftFact.setSlotValue("tickStamp", new Value(currentTime, RU.INTEGER));
					model.engine.assertFact(leftFact);
				}
				catch(JessException je) {
					System.err.println("An error occurred at line "
							+ je.getLineNumber());
					System.err.println("caused by " + je.getCause());
					System.err.println("Message: " + je.getDetail());
					je.printStackTrace();
				}
				// record departure of miner
				model.miningRecorder.recordDeparture(this);
			}
		}
	}
		
	/** lets the agent die in the current tick
	 * 
	 * @param tick the current model tick
	 */
	void die(int tick){
		setAlive(false);
		model.dump(name.toString() + " (" + gender.toString() + ") DIED AT AGE " + age + (isHIVPositive() ? " OF AIDS" : ""));
		if (name.toString().startsWith("miner")){
			// let him just leave the model
			left = true; 
			alive = true;
			return;
		}
		model.dump("was " + (head ? "head of " : "member of ") + household.getId().toString() + " with following members:");
		for (Person p : household.getMembers()){
			if (p != this){
				model.dump(" * " + p.getName() + " (" + p.getGender() + ", " + p.getAge() + ")");
			}
		}
		model.dump(name + " had " + getKin().size() + " relatives in " + getNumKinHouseholds() + " households (excluding own)");
		ArrayList<Person> friends = getFriends();
		model.dump(name.toString() + " had " + friends.size() + " friends");
		for (Person p : friends){
			model.dump("* " + p.getName() + " (" + p.getGender() + ", " + p.getAge() + ") of " + p.getHouseholdID()
					+ ", now has " + (p.getFriends().size()-1) + " friends left");
		}
		// record death in Jess
		try {
			Fact deathFact = new Fact("death", model.engine);
			deathFact.setSlotValue("name", this.name);
			deathFact.setSlotValue("household", this.household.getId());
			deathFact.setSlotValue("tickStamp", new Value(tick, RU.INTEGER));
			model.engine.assertFact(deathFact);
		}
		catch(JessException je) {
			System.err.println("An error occurred at line "
					+ je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}
		// person has to be taken off agent list in model and removed from the gui
		// --> can't be done here as it would change the list while it's being iterated in model.updateAgents()
		// --> will be done in updateAgents() after iterating through the whole list 
		
		// person has to be removed from household and any societies he/she belongs to
		// includes deleting the respective network links
		
		// if married, set spouse's status to widowed
		if (spouse != null){
			spouse.setMaritalStatus("widowed");
			spouse.setSpouse(null);
			model.dump("Spouse " + spouse.getName()+ " is now widowed");
		}
		// remove person from all current partners
		model.dump(name + " had " + currentPartners.size() + " sexual partners" + ((spouse != null) ? " outside of the marriage" : ""));
		for (Person p : currentPartners){
			p.removeCurrentPartner(this);
			model.dump("* " + p.getName() + " (" + p.getAge() + ") of " + p.getHouseholdID());
		}
		currentPartners.clear();
	}
	
	/** returns if the agent is a migrant
	 * 
	 * @return true if the agent is a migrant, false otherwise
	 */
	public boolean isMigrant() {
		return migrant;
	}

	/** sets if the agent is a migrant
	 * 
	 * @param migrant true if the agent is a migrant, false otherwise
	 */
	public void setMigrant(boolean migrant) {
		boolean old = this.migrant;
		this.migrant = migrant;
		pcs.firePropertyChange("migrant", old, this.migrant);
	}

	/** returns if the agent is a migrant currently home on a visit
	 * 
	 * @return true if home on a visit, false otherwise
	 */
	public boolean isOnVisit(){
		return onVisit;
	}
	
	/** sets if the agent is a migrant currently home on a visit
	 * 
	 * @param onVisit true if home on a visit, false otherwise
	 */
	public void setOnVisit(boolean onVisit){
		boolean old = this.onVisit;
		this.onVisit = onVisit;
		if (old != this.onVisit)
			pcs.firePropertyChange("onVisit", old, this.onVisit);		
	}
	
	/** returns the tick when this agent last visited home (from being a migrant)
	 * 
	 * @return tick of last visit home or -1 if the agent has never left yet
	 */
	public int getLastVisit(){
		return lastVisit;
	}
	
	/** sets the tick when this agent last visited home (from being a migrant)
	 * 
	 * @param tick tick of last visit home
	 */
	public void setLastVisit(int tick){
		int old = this.lastVisit;
		this.lastVisit = tick;
		pcs.firePropertyChange("lastVisit", old, this.lastVisit);	
	}

	/** returns the tick when this agent last gave birth
	 * 
	 * @return tick of last delivery or -1 if no births yet
	 */
	public int getLastDelivery() {
		return lastDelivery;
	}

	/** sets the tick when this agent last gave birth
	 * 
	 * @param lastDelivery tick of last delivery
	 */
	public void setLastDelivery(int lastDelivery) {
		this.lastDelivery = lastDelivery;
	}

	/** returns this agent's tag evolution propensity
	 * 
	 * @return the tag evoluation propensity
	 */
	public double getTagEvolutionPropensity() {
		return tagEvolutionPropensity;
	}

	/** sets this agent's tag evolution propensity
	 * 
	 * @param tep the new tag evoluation propensity
	 */
	public void setTagEvolutionPropensity(double tep) {
		double old = tagEvolutionPropensity;
		tagEvolutionPropensity = tep;
		pcs.firePropertyChange("tagEvolutionPropensity", old, tagEvolutionPropensity);
	}

	/** returns if this agent is registered to receive a state pension
	 * 
	 * @return true if agent is registered, false otherwise
	 */
	public boolean isRegisteredForPension() {
		return registeredForPension;
	}

	/** sets if this agent is registered to receive a state pension
	 * 
	 * @param registeredForPension true if agent is registered, false otherwise
	 */
	public void setRegisteredForPension(boolean registeredForPension) {
		boolean old = this.registeredForPension;
		this.registeredForPension = registeredForPension;
		pcs.firePropertyChange("registeredForPension", old, this.registeredForPension);
	}
	
	/** returns if this agent is registered to receive a child grant.
	 * 
	 * @return true if agent is registered, false otherwise
	 */
	public boolean isRegisteredForChildGrant() {
		return registeredForChildGrant;
	}

	/** sets if this agent is registered to receive a child grant.
	 * 
	 * @param registeredForChildGrant true if agent is registered, false otherwise
	 */
	public void setRegisteredForChildGrant(boolean registeredForChildGrant) {
		boolean old = this.registeredForChildGrant;
		this.registeredForChildGrant = registeredForChildGrant;
		pcs.firePropertyChange("registeredForChildGrant", old, this.registeredForChildGrant);
	}
	
	
	/** returns the persons this agent knows
	 * 
	 * @return list of person names
	 */
	public Value[] getKnownPersons(){
		Value[] list = new Value[knownPersons.size()];
		return knownPersons.toArray(list);
	}
	
	/** adds the given person to the list of known persons
	 * 
	 * @param person name of person 
	 */
	public void addKnownPerson(Value person){
		if (!knownPersons.contains(person)){
			Value[] old = getKnownPersons();
			knownPersons.add(person);
			pcs.firePropertyChange("knownPersons", old, getKnownPersons());
		}
	}
	
	/** adds the given person to the list of known persons
	 * 
	 * @param personString name of person as String
	 */
	public void addKnownPerson(String personString){
		try {
			Value person = new Value(personString, RU.SYMBOL);
			addKnownPerson(person);
		}
		catch (JessException je){
			System.err.println("An error occurred at line " + je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();			
		}
	}

	/** adds the given persons to the list of known persons
	 * 
	 * @param persons list of persons
	 */
	public void addKnownPersons(ArrayList<Value> persons){
		Value[] old = getKnownPersons();
		for (Value person : persons){
			if (!knownPersons.contains(person)){
				knownPersons.add(person);
			}
		}
		pcs.firePropertyChange("knownPersons", old, getKnownPersons());
	}
	
	/** adds the given persons to the list of known persons
	 * 
	 * @param persons list of person names
	 */
	public void addKnownPersons(Value[] persons){
		ArrayList<Value> temp = new ArrayList<Value>(persons.length);
		for (int i = 0; i < persons.length; i++){
			temp.add(persons[i]);
		}
		addKnownPersons(temp);
	}
	
	/** removes the given person from the list of known persons
	 * 
	 * @param person name of the person to be removed
	 */
	public void removeKnownPerson(Value person){
		if (knownPersons.contains(person)){
			Value[] old = getKnownPersons();
			knownPersons.remove(person);
			pcs.firePropertyChange("knownPersons", old, getKnownPersons());
		}
	}
	
	/** prints the agent endorsement scheme to console */
	public void printEndorsementScheme(){
		System.out.print(this.name + " has following endorsement scheme: ");
		this.agentEndorsementScheme.print();
	}
	
	/** returns this agent's father
	 * 
	 * @return name of father
	 */
	public Value getFather(){
		return this.father;
	}
	
	/** returns this agent's mother
	 * 
	 * @return name of mother
	 */
	public Value getMother(){
		return this.mother;
	}
	
	/** returns the parents of this agent
	 * 
	 * @return list of parent names
	 */
	protected Value[] getParents(){
		ArrayList<Value> parents = new ArrayList<Value>(2);
		if (mother != null) parents.add(mother);
		if (father != null) parents.add(father);
		Value[] pArray = new Value[parents.size()];
		return parents.toArray(pArray);
	}
	
	/** returns the children of this agent
	 * 
	 * @return list of children names
	 */
	public Value[] getChildren(){
		Value[] cArray = new Value[children.size()];
		return children.toArray(cArray);
	}
	
	/** returns the siblings of this agent
	 * 
	 * @return list of sibling names
	 */
	public Value[] getSiblings(){
		Value[] sArray = new Value[siblings.size()];
		return siblings.toArray(sArray);
	}
	
	/** add a child for this agent
	 * 
	 * @param child this agent's new child
	 */
	public void addChild(Person child){
		Value[] old = getChildren();
		children.add(child.getName());
		pcs.firePropertyChange("children", old, getChildren());
	}
	
	/** add a sibling for this agent
	 * 
	 * @param sibling this agent's new sibling
	 */
	public void addSibling(Person sibling){
		Value[] old = getSiblings();
		siblings.add(sibling.getName());
		pcs.firePropertyChange("siblings", old, getChildren());
	}
	
	/** searches for the closest living relative of this agent
	 * 
	 * @param forAccommodation true if the relative is to accomodate the agent, false if the relative
	 *  is to inherit
	 * @return closest living relative
	 */
	Person findLivingRelative(boolean forAccommodation){
		// depending on purpose of search (accommodation vs. inheritance)
		// -- look for grandparents/aunts/uncles 
		// -- or for parents/children/siblings/grandchildren
		// TODO: find out how inheritance really works in cases where the last remaining member of a hh dies
		ArrayList<Person> relatives = new ArrayList<Person>();

		if (forAccommodation){
			// look on father's side since the child belongs to the father's family
			// exception: if there is no known father, look at mother's side
			Person parent = null;
			if (father != null){
				model.namePersonMap.get(father.toString()); 
			}
			else if (mother != null){
				model.namePersonMap.get(mother.toString());
			} 
			if (parent == null) return null; // ### PROBLEM: no mother, if this child was initialised as other
			relatives.addAll(model.retrieveObjects(parent.getParents()));  // grandparents
			relatives.addAll(model.retrieveObjects(parent.getSiblings())); // aunts/uncles

			// remove dead relatives
			relatives = removeDead(relatives);
			// find best relative to accommodate
			return findBestAccommodator(relatives);
		}
		
		// find closest living relative to inherit 
		// (1) children -- if there are any
		relatives = removeDead(model.retrieveObjects(getChildren()));
		if (relatives.size() > 0){
			// youngest son inherits the house
			Person youngestSon = findYoungestSon(relatives);
			if (youngestSon != null){
				return youngestSon;
			}
			else {
				// if there is no son give it to a randomly chosen daughter, preferably from the same village
				return (pickRandomlyFromSameVillage(relatives));
			}
		}
		// (2) parents
		relatives = removeDead(model.retrieveObjects(getParents()));
		if (relatives.size() > 0){
			// randomly pick mother or father (only makes a difference if they are living in separate households)
			return relatives.get(0);
		}
		// (3) siblings
		relatives = removeDead(model.retrieveObjects(getSiblings()));
		if (relatives.size() > 0){
			// randomly pick one from the same village (if possible)
			return pickRandomlyFromSameVillage(relatives);
		}
		// (4) pick someone from the kinship network ###
		
		// no living relative found
		return null;
	}
	
	/** internal helper function: pick a random relative from the same village (if possible)
	 */
	private Person pickRandomlyFromSameVillage(ArrayList<Person> relatives){
		ArrayList<Person> sameVillage = new ArrayList<Person>();
		for (Person p : relatives){
			if (p.getHousehold().village.equals(this.household.village)){
				sameVillage.add(p);
			}
		}
		if (sameVillage.size() > 0) {
			return sameVillage.get(Random.uniform.nextIntFromTo(0, sameVillage.size()-1));
		}
		else {
			return relatives.get(Random.uniform.nextIntFromTo(0, relatives.size()-1));
		}		
	}
	
	/** internal helper function: determine youngest son of this agent to inherit the house
	 * of this agent (who has died) 
	 */
	private Person findYoungestSon(ArrayList<Person> children){
		int minAge = Integer.MAX_VALUE, minIndex = -1, i = -1;
		for (Person p : children){
			i++;
			if (p.gender.toString().equalsIgnoreCase("male") && p.getAge() < minAge){
				minAge = p.getAge();
				minIndex = i;
			}
		}
		if (minIndex >= 0) return children.get(minIndex);
		return null;
	}
	
	/** internal helper function: find the relative whose household is best suited to 
	 * accomodate this agent (who is an orphan) 
	 */
	Person findBestAccommodator(ArrayList<Person> relatives){
		// for now: find richest household
		double maxIncome = 0.0;
		int i = -1, incomeIndex = -1;
		for (Person p : relatives){
			i++;
			Household hh = p.getHousehold();
			if (hh.expectedIncome > maxIncome){
				maxIncome = hh.expectedIncome;
				incomeIndex = i;
			}
		}
		if (incomeIndex >= 0 && maxIncome > 1000) return relatives.get(incomeIndex);
		return null;
	}

	/** internal helper function: removes the dead persons from a list of persons */
	private ArrayList<Person> removeDead(ArrayList<Person> list){
		ArrayList<Person> living = new ArrayList<Person>();
		for (Person p : list){
			if (p.isAlive()) living.add(p);
		}
		return living;
	}
	
	/** lets this agent join the specified burial society
	 * 
	 * @param bs burial society
	 */
	public void joinBurialSociety(BurialSociety bs){
		Value[] old = getBurialSocieties();
		this.burialSocieties.add(bs);
		pcs.firePropertyChange("burialSocieties", old, getBurialSocieties());
	}
	
	/** lets this agent leave the specified burial society
	 * 
	 * @param bs burial society
	 */
	public void leaveBurialSociety(BurialSociety bs){
		Value[] old = getBurialSocieties();
		this.burialSocieties.remove(bs);
		pcs.firePropertyChange("burialSocieties", old, getBurialSocieties());
	}
	
	/** returns the burial societies this agent is a member of
	 * 
	 * @return list of burial society names
	 */
	public Value[] getBurialSocieties(){
		Value[] bsList = new Value[this.burialSocieties.size()];
		for (int i = 0; i < bsList.length; i++){
			bsList[i] = this.burialSocieties.get(i).getName();
		}
		return bsList;
	}

	/** lets the agent join the specified stokvel
	 * 
	 * @param s the stokvel
	 */
	public void joinStokvel(Stokvel s){
		Value[] old = getStokvels();
		this.stokvels.add(s);
		pcs.firePropertyChange("stokvels", old, getStokvels());
	}
	
	/** lets the agent leave the specified stokvel
	 * 
	 * @param s the stokvel
	 */
	public void leaveStokvel(Stokvel s){
		Value[] old = getStokvels();
		this.stokvels.remove(s);
		pcs.firePropertyChange("stokvels", old, getStokvels());
	}
	
	/** returns thestokvels this agent is a member of
	 * 
	 * @return list of stokvel names
	 */
	public Value[] getStokvels(){
		Value[] sList = new Value[this.stokvels.size()];
		for (int i = 0; i < sList.length; i++){
			sList[i] = this.stokvels.get(i).getName();
		}
		return sList;
	}
	
	/** lets the agent rate the performance of another agent with regard to the given endorsement token.
	 * This is to decide if the agent should endorse the other agent with this endorsement.
	 * 
	 * @param endorsement	the endorsement in question (e.g. reliable) 
	 * @param performanceValue	the given performance, measured as proportion of 1.0
	 * @return true if the performance is good enough, false otherwise
	 */
	public boolean rate(Value endorsement, double performanceValue){
		// retrieve personal weight of endorsement from endorsement scheme
		double weight = Math.abs(this.agentEndorsementScheme.getValue(endorsement));
		double relativeWeight = weight / this.numEndorsementClasses;
		boolean result = (performanceValue >= relativeWeight);
		//*//model.dump(name + " rates " + performanceValue + " of " + endorsement.toString() + " as " + result);
		return result;
	}
	// ### needs to be rewritten
	// at the moment allows for people to be endorsed as reliable and unreliable at the same time!
	// example from a trace:
	/* person-120 rates 0.42857142857142855 of reliable as true
	 	head of household-21 endorses person-59 as reliable
	 	person-120 rates 0.5714285714285714 of unreliable as true
	 	head of household-21 endorses person-59 as unreliable
	*/
	
	/** lets the agent rate the performance of another agent with regard to the given endorsement token.
	 * This is to decide if the agent should endorse the other agent with this endorsement.
	 * 
	 * @param endorsementString	the endorsement in question (e.g. reliable) as String
	 * @param performanceValue	the given performance, measured as proportion of 1.0
	 * @return true if the performance is good enough, false otherwise
	 */
	public boolean rate(String endorsementString, double performanceValue){
		try{
			boolean result = rate(new Value(endorsementString, RU.SYMBOL), performanceValue);
			return result;
		}
		catch (JessException je){
			System.err.println("An error occurred at line "
					+ je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}
		return false;
	}

	/** returns if the agent is attending school or not
	 * 
	 * @return true if the agent is still at school, false otherwise
	 */
	public boolean isAtSchool() {
		return atSchool;
	}

	/** sets if this agent is attending school or not
	 * 
	 * @param atSchool true if the agent is still at school, false otherwise
	 */
	public void setAtSchool(boolean atSchool) {
		boolean old = this.atSchool;
		this.atSchool = atSchool;
		pcs.firePropertyChange("atSchool", old, this.atSchool);
	}
	
	/** returns the achieved education level
	 * 
	 * @return education level
	 */
	public Value getAchievedEducation(){
		return this.achievedEducation;
	}
	
	/** sets the achieved education level to the given level
	 * 
	 * @param level education level as a Value object
	 */
	protected void setAchievedEducation(Value level){
		Value old = this.achievedEducation;
		this.achievedEducation = level;
		pcs.firePropertyChange("achievedEducation", old, this.achievedEducation);
	}
	
	/** sets the achieved education level to the given level
	 * 
	 * @param levelIndex education level as an index into the list of education levels
	 */
	public void setAchievedEducation(int levelIndex){
		try {
			setAchievedEducation(new Value(educationLevels[levelIndex], RU.SYMBOL));
		}
		catch(JessException je) {
			System.err.println("An error occurred at line " + je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}
	}
	
	/** sets the achieved education level to the given level
	 * 
	 * @param elString education level as String
	 */
	public void setAchievedEducation(String elString){
		try {
			setAchievedEducation(new Value(elString, RU.SYMBOL));
		}
		catch(JessException je) {
			System.err.println("An error occurred at line " + je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}
	}
	
	/** let this agent marry the given person
	 * 
	 * @param spouse this agent's new spouse
	 */
	public void marry(Person spouse){
		setSpouse(spouse);
		setMaritalStatus("married");
		// reduce max. number of concurrent partners
		setMaxNumPartners(Math.max(0, getMaxNumPartners() - Random.uniform.nextIntFromTo(1,2)));
		// drop all current partners
		for (Person p : currentPartners){
			p.removeCurrentPartner(this);
			model.dump(this.name + " breaks up with " + p.getName() + " due to marriage");
		}
		removeCurrentPartners();
	}
	
	/** returns the tick this agent contracted HIV
	 * 
	 * @return tick of HIV infection or -1 if not infected
	 */
	public int getInfectedSince(){
		if (hivAIDS != null){
			return hivAIDS.infectedSince;
		}
		return -1;
	}
	
	/** sets the tick this agent contracted HIV
	 * 
	 * @param tick tick of HIV infection
	 */
	public void setInfectedSince(int tick){
		hivAIDS = new HIVAIDS(tick);
	}
	
	/** sets the tick this agent contracted HIV and the stage of the disease
	 * 
	 * @param tick tick of HIV infection
	 * @param stage stage of HIV progression
	 */
	public void setInfectedSince(int tick, int stage){
		hivAIDS = new HIVAIDS(tick, stage);
	}
	
	/** returns the current sexual partners
	 * 
	 * @return list of the names of the sexual partners 
	 */
	public Value[] getCurrentPartners(){
		Value[] partners = new Value[currentPartners.size()];
		for (int i = 0; i < partners.length; i++){
			partners[i] = this.currentPartners.get(i).getName();
		}
		return partners;
	}
	
	/** adds the given sexual partner
	 * 
	 * @param p the new sexual partner
	 */
	public void addCurrentPartner(Person p){
		Value[] old = getCurrentPartners();
		currentPartners.add(p);
		pcs.firePropertyChange("currentPartners", old, getCurrentPartners());
		partnershipStart.add(model.getTick());
//		System.out.println(model.getTick() + " " + this.name + " adds partner " + p.getName());
	}
	
	/** removes the given sexual partner
	 * 
	 * @param p sexual partner to be removed
	 */
	public void removeCurrentPartner(Person p){
//		System.out.println(model.getTick() + " " + this.name + " removes partner " + p.getName());
		updatePartnerDuration(p);
		Value[] old = getCurrentPartners();
		currentPartners.remove(p);
		pcs.firePropertyChange("currentPartners", old, getCurrentPartners());
	}
	
	/** internal helper function: updates the partner duration for the given partner */
	private void updatePartnerDuration(Person p){
		int index = currentPartners.indexOf(p);
		if (index < 0 || index >= currentPartners.size()){
			System.out.println("!!! index = " + index + " of " + currentPartners.size());
			System.out.println("while trying to remove partner " + p.getName() + " from " + this.name);
			return;
		}
		int startTick = partnershipStart.get(index);
		partnershipStart.remove(index);
		// update data 
		updatePartnerDuration(p, (model.getTick()-startTick));
	}
	
	/** internal helper function: update recorded data about partnership duration */
	private void updatePartnerDuration(Person p, int duration){
		// update data (only once per partnership!)
		if (this.gender.toString().equals("female")){
			model.partnershipRecorder.record(new Person[]{this, p}, duration);
		}
	}
	
	/** returns the duration of the sexual partnership with the given person up to the current model tick
	 * 
	 * @param p the sexual partner
	 * @return duration of sexual partnership
	 */
	int currentPartnerDuration(Person p){
		int index = currentPartners.indexOf(p);
		if (index < 0 || index >= currentPartners.size()){
			return -1;
		}
		return (model.getTick() - partnershipStart.get(index));
	}
	
	/** removes all current sexual partners */
	public void removeCurrentPartners(){
		// to be more efficient when removing all partners: iterate backwards through both lists 
		Value[] old = getCurrentPartners();
		while (currentPartners.size() > 0){
			int i = currentPartners.size()-1;
			Person p = currentPartners.get(i);
//			System.out.println(model.getTick() + " " + this.name + " removes partner " + p.getName());
			updatePartnerDuration(p, (model.getTick()-partnershipStart.get(i)));
			currentPartners.remove(i);
			partnershipStart.remove(i);
		}
		pcs.firePropertyChange("currentPartners", old, getCurrentPartners());
	}
		
	/** returns if this agent has contracted HIV
	 * 
	 * @return true if the agent is HIV-positive, false otherwise
	 */
	public boolean isHIVPositive(){
		if (hivAIDS != null) {
			return hivAIDS.isHIVPositive();
		}
		return false;
	}
	
	/** returns if this agent has developed AIDS
	 * 
	 * @return true if the agent has AIDS, false otherwise
	 */
	public boolean getHasAIDS(){	// method name changed from hasAIDS() so that the slot in Jess is called hasAIDS
		if (hivAIDS != null){
			return hivAIDS.hasAIDS(model.getTick());
		}
		return false;
	}
	
	/** returns this agent's friends
	 * 
	 * @return list of friends
	 */
	protected ArrayList<Person> getFriends(){
		Value[] fArray = new Value[friends.size()];
		return model.retrieveObjects(friends.toArray(fArray));
	}
	
	/** removes the given friend from this agent's list of friends
	 * @param friend the friend to be removed
	 */
	protected void removeFriend(Person friend){
		this.friends.remove(friend.getName());
	}
	
	/** returns if the given person is a friend of this agent
	 * @return true if the person is a friend, false otherwise
	 */
	protected boolean hasFriend(Person p){
		return this.friends.contains(p.getName());
	}
	
	/** retrieves relatives in other households (from kinship network)
	 * 
	 * @return list of relatives in households other than own
	 */
	protected ArrayList<Person> getKin(){
		ArrayList<Person> kin = new ArrayList<Person>();
		// retrieve kin from kinship network between households
		for (Object o : household.getNode().getOutEdges()){
			NetworkLink link = (NetworkLink)o;
			if (link.getType().equals("kinship")){
				NetworkNode node = link.getTo();
				kin.addAll(model.nameHouseholdMap.get(node.getId()).getMembers());
			}
		}
		return kin;
	}
	
	/** returns the number of kin households, i.e. the number of households
	 * this agent's household maintains a kinship link with
	 * 
	 * @return number of kin households
	 */
	protected int getNumKinHouseholds(){
		int n = 0; 
		for (Object o : household.getNode().getOutEdges()){
			NetworkLink link = (NetworkLink)o;
			if (link.getType().equals("kinship")){
				n++;
			}
		}
		return n;
	}
	
	/** returns a String representation of this person. 
	 * @return this agent's name
	 */
	public String toString(){
		return name.toString();
	}

	//--- outsourced from Jess for efficiency --------------------------------------
	
	/** endorse the given neighbour, i.e. assert a new endorsed fact with endorsement "is-neighbour"
	 * for this agent
	 * 
	 * @param neighbour the neighbour 
	 */
	public void assertNeighbourEndorsement(Person neighbour){
		//MUCH more efficient than calculating distances and asserting endorsement in Jess!!!
		try{
			Fact f = new Fact("endorsed", engine);
			f.setSlotValue("endorsement", new Value("is-neighbour", RU.SYMBOL));
			f.setSlotValue("endorsed-object", neighbour.name);
			f.setSlotValue("clauseOwner", this.name);
			f.setSlotValue("tickStamp", new Value(model.tick, RU.INTEGER));
			engine.assertFact(f);
		}
		catch(JessException je) {
			System.err.println("An error occurred at line " + je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}
	}
	
	/** endorse the given fellow club member, i.e assert a new endorsed fact with endorsement "same-club"
	 * 
	 * @param fellowMember the fellow club member
	 */
	public void assertSameClubEndorsement(Person fellowMember){
		try{
			Fact f = new Fact("endorsed", engine);
			f.setSlotValue("endorsement", new Value("same-club", RU.SYMBOL));
			f.setSlotValue("endorsed-object", fellowMember.name);
			f.setSlotValue("clauseOwner", this.name);
			f.setSlotValue("tickStamp", new Value(model.tick, RU.INTEGER));
			engine.assertFact(f);
		}
		catch(JessException je) {
			System.err.println("An error occurred at line " + je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}
	}
	
	/** un-endorse the given fellow club member, i.e retract the endorsed fact with endorsement "same-club"
	 * for this agent
	 * 
	 * @param fellowMember the fellow club member
	 */
	public void removeSameClubEndorsement(Person fellowMember){
		try {
			Fact f = new Fact("endorsed", engine);
			f.setSlotValue("endorsement", new Value("same-club", RU.SYMBOL));
			f.setSlotValue("endorsed-object", fellowMember.name);
			f.setSlotValue("clauseOwner", this.name);
			engine.watch(WatchConstants.FACTS);
			engine.retract(f);
			engine.unwatch(WatchConstants.FACTS);
		}
		catch(JessException je) {
			System.err.println("An error occurred at line " + je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();		
		}
	}

	/** returns the tick of the end of the last job this person had
	 *  or -1 if they did not have a job yet
	 * @return tick of the end of the last job
	 */
	public int getLastJob() {
		return lastJob;
	}

	/** sets the tick of the end of the last job this person had
	 * 
	 * @param lastJob tick of the end of the last job
	 */
	public void setLastJob(int lastJob) {
		int old = this.lastJob;
		this.lastJob = lastJob;
		pcs.firePropertyChange("lastJob", old, lastJob);
	}
	
	/** returns the persons this agent is linked to, regarding the given
	 * link type (network layer)
	 * 
	 * @param linkType the type of link (friend, partner, stokvel or burial)
	 * @return list of persons
	 */
	public ArrayList<Person> retrieveLinkedPersons(String linkType){
		if (linkType.equalsIgnoreCase("friend")){
			return getFriends();
		}
		if (linkType.equalsIgnoreCase("partner")){
			if (this.spouse != null){
				ArrayList<Person> partners = new ArrayList<Person>(currentPartners.size()+1);
				partners.addAll(currentPartners);
				partners.add(spouse);
				return partners;
			}
			else {
				return this.currentPartners;
			}
		}
		if (linkType.equalsIgnoreCase("stokvel")){
			ArrayList<Person> otherMembers = new ArrayList<Person>();
			for (Stokvel stokvel : stokvels){
				otherMembers.addAll(stokvel.members);
				otherMembers.remove(this);
			}
			return otherMembers;
		}
		if (linkType.equalsIgnoreCase("burial")){
			ArrayList<Person> otherMembers = new ArrayList<Person>();
			for (BurialSociety bs : burialSocieties){
				otherMembers.addAll(bs.members);
				otherMembers.remove(this);
			}
			return otherMembers;
		}
		return new ArrayList<Person>(0);
	}
	
	/** returns the mine experience this person has had so far
	 * @return number of ticks of mine employment
	 */
	protected int getMineExperience(){
		return this.mineExperience;
	}
	
	/** adds the given number of ticks to the mine experience
	 * @param mineExperience number of ticks of new mine experience
	 */
	protected void addMineExperience(int mineExperience){
		this.mineExperience += mineExperience;
	}

}
