package org.cfpm.caves.za;

import java.util.ArrayList;
import java.util.HashMap;

import jess.Value;

/** An abstract savings club. Used as super-class for Stokvel and BurialSociety.
 * 
 * @author Ruth
 *
 */
public abstract class SavingsClub extends Group {
	
	//--- savings club attributes --------------------------------

	/** list of members on a grace period */
	HashMap<Person,Integer> onGrace = new HashMap<Person,Integer>();
	/** current holdings of the club */
	double holdings = 0;
	/** monthly membership fee. Default is 30 Rand */
	double monthlyFee = 30; 	// default: 30 Rand
	
	//--- constructors ----------------------------------------------
	
	/** Creates a new savings club of the given type with the given holdings and members
	 * @param type the club's type
	 * @param holdings the club's holdings
	 * @param members the list of members
	 * @param model the model
	 */
	public SavingsClub(String type, double holdings, ArrayList<Person> members, SouthAfricaModel model){
		super(type, members, model);
		this.holdings = holdings;
	}

	//--- methods -------------------------------------------------------
	
	/** returns the current holdings
	 * 
	 * @return amount of money
	 */
	public double getHoldings(){
		return holdings;
	}

	/** sets the holdings to the given amount
	 * 
	 * @param amount new holdings
	 */
	public void setHoldings(double amount){
		double old = this.holdings;
		this.holdings = amount;
		pcs.firePropertyChange("holdings", old, this.holdings);
	}

	/** returns the monthly membership fee
	 * 
	 * @return monthly fee
	 */
	public double getMonthlyFee(){
		return monthlyFee;
	}
	
	/** sets the monthly membership fee
	 * 
	 * @param amount monthly fee
	 */
	public void setMonthlyFee(double amount){
		double old = this.monthlyFee;
		this.monthlyFee = amount;
		pcs.firePropertyChange("monthlyFee", old, this.monthlyFee);
	}

	/** lets a member pay in the specified amount
	 * 
	 * @param p the member paying in
	 * @param amount the amount the member is paying in
	 */
	public void payIn(Person p, double amount){
		this.holdings += amount;
		// check if member was on grace period
		if (onGrace.containsKey(p)){
			// assume member paid all outstanding debts
			// TODO: keep track of outstanding amounts so that members can pay off debt over several months
			Value[] oldMembers = getMembersOnGrace();
			int[] oldPeriods = getGracePeriods();
			onGrace.remove(p);
			pcs.firePropertyChange("membersOnGrace", oldMembers, getMembersOnGrace());
			pcs.firePropertyChange("gracePeriods", oldPeriods, getGracePeriods());
		}
	}
	
	/** removes the given member from the club
	 * @param member the member to be removed
	 */
	public void removeMember(Person member){
		super.removeMember(member);
		// check if member was on grace period
		if (onGrace.containsKey(member)){
			Value[] oldMembers = getMembersOnGrace();
			int[] oldPeriods = getGracePeriods();
			onGrace.remove(member);
			pcs.firePropertyChange("membersOnGrace", oldMembers, getMembersOnGrace());
			pcs.firePropertyChange("gracePeriods", oldPeriods, getGracePeriods());
		}
	}
	
	/** lets the given member request a grace period from the club. This happens
	 * when a member can't pay the monthly fee but wants to retain club membership.
	 * 
	 * @param member the requesting member
	 */
	public void requestGracePeriod(Person member){
		Value[] old = getMembersOnGrace();
		int[] oldPeriods = getGracePeriods();
		//	 record current tick as start of grace period
		onGrace.put(member, new Integer(member.getModel().getTick())); 
		pcs.firePropertyChange("membersOnGrace", old, getMembersOnGrace());
		pcs.firePropertyChange("gracePeriods", oldPeriods, getGracePeriods());
	}

	/** returns the members currently on a grace period
	 * 
	 * @return list of member names
	 */
	public Value[] getMembersOnGrace(){
		Value[] membersOnGrace = new Value[onGrace.size()];
		int i = 0;
		for (Person p : onGrace.keySet()){
			membersOnGrace[i++] = p.getName();
		}
		return membersOnGrace;
	}
	
	/** returns the starts of the currently granted grace periods
	 * 
	 * @return list of grace period starts (model ticks)
	 */
	public int[] getGracePeriods(){
		int[] gracePeriods = new int[onGrace.size()];
		int i = 0;
		for (Person p : onGrace.keySet()){
			gracePeriods[i++] = onGrace.get(p).intValue();
		}
		return gracePeriods;
	}
	
	
}
