package org.cfpm.caves.za;

import java.util.IdentityHashMap;

import jess.*;

/** This class implements a user function in Jess.
 *  The function "sort-values" is used to sort a list of endorsement values
 *  in either descending or ascending order to allow the agent calling
 *  this function to pick the best / worst out of a group of endorsed agents.
 * 
 * @author Ruth
 *
 */
public class SortValues implements Userfunction {
	
	/** returns the name of the function in Jess */	
	public String getName(){
		return "sort-values";
	}
	
	/**
	 * The first argument of the call is a list of numerical values, e.g.
	 * endorsement values, to be sorted, the second determines if the list
	 * should be sorted in descending order (TRUE) or ascending order (FALSE).
	 * The function returns the sorted list in the form of the original one-based INDICES. 
	 * These indices can then be used to access the original list or any other 
	 * associated list in the sorted order via Jess' nth$ function.
	 * 
	 * Example: Let (5 4 5 6 3 0 7) be the list of (endorsement) values to be sorted
	 * in descending order. Calling sort-values will return the following list:
	 * (7 4 1 3 2 5 6), meaning that element #7 (7) in the endorsement value list is
	 * the largest, element #4 (6) is the second-largest, etc. 
	 *
	 * 
	 * @see jess.Userfunction#call(jess.ValueVector, jess.Context)
	 * @param vv -- the ValueVector containing the list of values (first parameter) and 
	 * sort order (second parameter)
	 * @param context -- the context of the Rete engine
	 */
	public Value call(ValueVector vv, Context context) throws JessException{
		ValueVector endorsementValues = vv.get(1).listValue(context);
		boolean descending = Boolean.valueOf(vv.get(2).stringValue(context)).booleanValue();
		Double[] values = new Double[endorsementValues.size()];
		for (int i = 0; i < values.length; i++){
			values[i] = new Double(endorsementValues.get(i).floatValue(context));
		}
		int[] result = sort(values);
		ValueVector vvResult = new ValueVector();
		if (descending){
			for (int i = result.length-1; i >= 0; i--){
				vvResult.add(result[i]);
			}
		}
		else {
			for (int i = 0; i < result.length; i++){
				vvResult.add(result[i]);
			}
		}
		return new Value(vvResult, RU.LIST);
	}
	
	// use Mergesort to sort value array
	
	/** internal function: sorts the list of given values and returns the
	 *  sorted list of indices.
	 */
	private int[] sort(Double[] values){
		int[] results = new int[values.length];
		IdentityHashMap<Double, Integer> map = new IdentityHashMap<Double,Integer>();
		for (int i= 0; i < values.length; i++){
			map.put(values[i], new Integer(i+1));
		}
		mergesort(values, 0, values.length-1);
		for (int i = 0; i < values.length; i++){
			results[i] = map.get(values[i]).intValue();
		}
		return results;
	}
	
	/** internal function: recursive mergesort */
	public static void mergesort(Double[] values, int lo, int hi){
		if (lo < hi){
			int m = (lo+hi)/2;
			mergesort(values, lo, m);
			mergesort(values, m+1, hi);
			merge(values, lo, m, hi);
		}
	}
	
	/** internal function: the merge step in Mergesort */
	static void merge(Double[] values, int lo, int m, int hi){
		int i, j, k;
		// copy values (lo, m) into auxiliary array temp
		Double[] temp = new Double[m+1];
		i = 0; j = lo;
		while (j <= m){
			temp[i++] = values[j++];
		}
		// put next-greatest element (from either temp or second half of values) into correct position
		i = 0; k = lo;
		while (k < j && j <= hi) {
			if (temp[i].doubleValue() <= values[j].doubleValue())
				values[k++] = temp[i++];
			else
				values[k++] = values[j++];	
		}
		// copy back remaining elements of temp (if there are any)
		while (k < j){
			values[k++] = temp[i++];
		}
	}
	
	/** internal helper function for debugging: prints the given list of doubles */
	private void print(Double[] list){
		System.out.print("( ");
		for (int i = 0; i < list.length; i++){
			System.out.print(list[i].doubleValue() + " ");
		}
		System.out.println(")");
	}

	/** internal helper function for debugging: prints the given list of integers */
	private void print(int[] list){
		System.out.print("( ");
		for (int i = 0; i < list.length; i++){
			System.out.print(list[i] + " ");
		}
		System.out.println(")");
	}

	/** internal helper function for debugging: prints the given list of Values */
	private void print(ValueVector list) throws JessException {
		System.out.print("( ");
		for (int i = 0; i < list.size(); i++){
			System.out.print(list.get(i).toString() + " ");
		}
		System.out.println(")");
	}	
}
