package org.cfpm.caves.za;

import uchicago.src.sim.engine.Schedule;
import uchicago.src.sim.engine.SimModelImpl;
import uchicago.src.sim.util.Random;
import uchicago.src.sim.engine.BasicAction;
import uchicago.src.sim.gui.RandomGraphLayout;
import uchicago.src.sim.network.NetworkFactory;
import cern.jet.random.*;
import cern.jet.random.engine.MersenneTwister;
import cern.jet.random.engine.RandomEngine;
import uchicago.src.sim.gui.Object2DDisplay;
import uchicago.src.sim.space.Object2DGrid;
import uchicago.src.sim.gui.DisplaySurface;
import uchicago.src.sim.analysis.DataFileWriter;
import uchicago.src.sim.analysis.DataFileHeader;
import uchicago.src.sim.analysis.LocalDataRecorder;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import java.awt.Color;
import java.io.PrintWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;

import jess.*;

import org.cfpm.caves.za.util.NetworkUtilities;

/** The model class.
 * 
 * @author Ruth, Scott
 *
 */
public class SouthAfricaModel extends SimModelImpl {
	
	//---- model-global constants ------------------------------
	
	private static int CHILD = 0, GRANDCHILD = 1, OTHER = 2;	// household member types (with regard to hh head)
	
	//---- model parameters ------------------------------------
	
	/** model parameter: number of initial households. Default is 24. */
	int numHouseholds = 24;
	/** model parameter: number of villages. Default is 1. */
	int numVillages = 1;
	/** model parameter: number of church denominations. Default is 4. */
	int numDenominations = 4;
	/** model parameter: overall proportion of households being member of a church. Default is 0.8. */
	double churchParticipationRate = 0.8;
	/** model parameter: overall proportion of adults with a job. Default is 0.4. */
	double adultEmploymentRate = 0.4;
	/** model parameter: overall proportion of adults with a government job. Default is 0.1. */
	double governmentEmploymentRate = 0.1;
	/** model parameter: monthly wage within village. Default is 200 Rand. */
	double standardWage = 200;  
	/** model parameter: monthly child grant for children under the age limit. Default is 200 Rand. */
	double childGrant = 200.0;  
	/** model parameter: proportion of eligible children receiving the child grant. Default is 1.0 */
	double childGrantProportion = 1.0;
	/** model parameter: upper age limit for child grant. Default is 7 years. */
	int childGrantAgeLimit = 7;
	/** model parameter: monthly state pension for seniors. Default is 870 Rand. */
	double statePension = 870.0; 
	/** model parameter: proportion of female seniors >= 60 receiving a pension. Default is 0.58. */
	double pensionProportionFemale = 0.58;
	/** model parameter: proportion of male seniors >= 65 receiving a pension. Default is 0.5. */
	double pensionProportionMale = 0.5;
	/** model parameter: lower age limit for female seniors to receive a state pension. Default is 60. */
	int pensionAgeLimitFemale = 60;
	/** model parameter: lower age limit for male seniors to receive a state pension. Default is 65. */
	int pensionAgeLimitMale = 65;
	/** model parameter: monthly food cost for a child up to 7 years. Default is 25 Rand. */
	double foodCostChild = 25.0; 
	/** model parameter: monthly food cost for a female. Default is 100 Rand. */
	double foodCostFemaleAdult = 100; 
	/** model parameter: monthly food cost for a male. Default is 120 Rand. */
	double foodCostMaleAdult = 120;  
	/** model parameter: yearly school fee for non-compulsory secondary school (pupils > 15 years). Default is 200 Rand. */
	double schoolFee = 200;
	/** model parameter: yearly university fees. Default is 10000 Rand. */
	double collegeFee = 10000;
	/** model parameter: minimum lobola (bride price). Default is 8000 Rand */
	int minLobola = 8000;
	/** model parameter: maximum lobola (bride price). Default is 12000 Rand */
	int maxLobola = 12000;
	/** model parameter: minimum remittance from a migrant agent. Default is 200 Rand. */
	int minRemittance = 200;
	/** model parameter: maximum remittance from a migrant agent. Default is 600 Rand. */
	int maxRemittance = 600;
	/** model parameter: probability for sending remittance home for any migrant in any month. Default is 0.1 */
	double remittanceProbability = 0.1;
	/** model parameter: average cost of a burial. Default is 5000 Rand. */
	double burialCost = 5000;
		// funerals can sometimes cost the equivalent of seven months' income. Source: http://knowledge.allianz.com/en/globalissues/microfinance/microcredit/financial_diaries_microfinance.html
   	// 10000-15000 Rand. Source: http://www.financialdiaries.com/files/Funerals%20and%20Finance%20-%20Events%20in%20the%20lives%20of%20Financial%20Diaries%20households.pdf 
 		// from last field trip to Ga-Selala: 4000-10000 Rand
	/** model parameter: proportion of households in the village with a shack to rent out if necessary.
	 * Default is 0.4 */
	double shackProportion = 0.4;
	/** model parameter: lower bound of the monthly rent asked for a shack in the village. Default is 200 Rand */
	int minShackRent = 200;
	/** model parameter: upper bound of the monthly rent asked for a shack in the village. Default is 500 Rand */
	int maxShackRent = 500;
	/** model parameter: average cost for a new house in the village. Default is 2000 Rand. */
	int houseCost = 2000;	
	/** model parameter: mother-to-child transmission rate of HIV. Default is 0.33 */
	double motherChildHIVTransmissionRate = 0.33;
		// "one-third [is] a reasonable value for the mother-to-child HIV-transmission rate" (Heuveline et al. 2004, p5)	
	/** model parameter: birth rate as number of births per 1000 per year. Default is 19.0. */
	double birthRate = 19.0; // 19 per 1000 population for Sekhukhune (2002 survey) according to http://www.statssa.gov.za/Publications/RuralDevelopment/RuralDevelopment.pdf
							 // 18.2 / 1000 pop. for South Africa (2006 estimate) according to http://www.indexmundi.com/south_africa/birth_rate.html

	/** model parameter: size (number of cells) of the 2D grid used as spatial model in x direction. 
	 *  Default is 30.*/
	int gridSizeX = 30;
	/** model parameter: size (number of cells) of the 2D grid used as spatial model in y direction. 
	 *  Default is 30.*/
	int gridSizeY = 30;
	/** model parameter: radius (number of cells) used to determine neighbourhood of a household.
	 *  Default is 4.24. */
	double neighbourhoodRadius = 4.24;
	/** model parameter: density factor of villages. Determines how "packed" villages are with houses.
	 *  Default is 1.3. */
	double densityFactor = 1.3;

	/** model parameter: tick to start mining in the area. Set to -1 to not have any mines.
	 *  Default is 240 (5 years). */
	int miningTick = 240;
	/** model parameter: number of skilled workers needed for the mine. Default is 20. */
	int neededSkilled = 20;
	/** model parameter: number of unskilled workers needed for the mine. Default is 80. */
	int neededUnskilled = 80;
	/** model parameter: HIV prevalence amongst in-migrating mine workers. Default is 0.3. */
	double mineHIVPrevalence = 0.3;
		// a study found 29.4% prevalence: http://gateway.nlm.nih.gov/MeetingAbstracts/ma?f=102279783.html
	/** model parameter: mean inter-arrival time for in-migrating mine workers. Default is 10. */
	double interArrivalTime = 10;
	/** model parameter: mean number of in-migrants arriving in the area at one time. 
	 *  Default is 5. */
	int meanNumArrived = 5;
	/** model parameter: standard deviation of the number of in-migrants. Default is 2. */
	int sdNumArrived = 2;
	/** model parameter: mean age of in-migrants. Default is 29. */
	int meanAgeArrived = 29;
	/** model parameter: standard deviation of the age of in-migrants. Default is 7. */
	int sdAgeArrived = 7;
	
	/** model parameter: number of tags (= length of the tag array) used to model agents' 
	 * characteristic traits. Default is 7. */
	int tagLength = 7;
	/** model parameter: number of different values a tag can take (integers between 0 and tagBase - 1).
	 *  Default is 5. */
	int tagBase = 5;
	/** model parameter: maximal tag evolution propensity. Default is 0.08. 
	 *  Every agent is assigned a random tag evolution propensity between max and min. */
	double maxTagEvolutionPropensity = 0.08;
	/** model parameter: minimal tag evolution propensity.
	 *  Default is 0.01. */
	double minTagEvolutionPropensity = 0.01; 
	/** model parameter: minimum value for the base used to compute the overall endorsement value.
	 *  Default is 1.0. Every agent is assigned a random base between max and min. */
	double minEndorsementBase = 1.0;
	/** model parameter: maximum value for the base used to compute the overall endorsement value.
	 *  Default is 3.0. Every agent is assigned a random base between max and min. */
	double maxEndorsementBase = 3.0;
	/** model parameter: minimum number of endorsement classes. Default is 2. 
	 * Every agent is assigned a random number of classes between max and min. */
	int minEndorsementClasses = 2;
	/** model parameter: maximum number of endorsement classes. Default is 5. 
	 * Every agent is assigned a random number of classes between max and min. */
	int maxEndorsementClasses = 5;
	/** model parameter: upper limit for the maximal number of friends. Default is 12. 
	 *  Every agent is assigned a random maximal number of friends between upper and lower limit. */
	int upperMaxNumFriends = 12;
	/** model parameter: lower limit for the maximal number of friends. Default is 6. 
	 *  Every agent is assigned a random maximal number of friends between upper and lower limit. */
	int lowerMaxNumFriends = 6; 
	/** model parameter: upper limit for the maximal number of concurrent sexual partners. Default is 5. 
	 *  Every agent is assigned a random maximal number of partners between upper and lower limit. */
	int upperMaxNumPartners = 5;
	/** model parameter: lower limit for the maximal number of concurrent sexual partners. Default is 1. 
	 *  Every agent is assigned a random maximal number of partners between upper and lower limit. */
	int lowerMaxNumPartners = 1; 
	/** model parameter: age range for similarity assessment (basically, age range for friends) given
	 * as proportion of age of the agent. Default is 0.25 (age range is +- 25% of own age). */
	double similarityAgeRange = 0.25;
	/** model parameter: probability to encounter a random sexual partner. Default is 0.1. */
	double randomPartnerProbability = 0.1;
	
	/** model parameter: flag to turn the GUI on (true) or off (false). Default is false. */
	boolean showGUI = false;
	/** model parameter: flag to turn the output to file on (true) or off (false). Default is true. */
	boolean printToFile = true;
	/** model parameter: flag to turn the output to console on (true) or off (false). Default is true. */
	boolean showOutput = true;
	/** model parameter: Path for the output directory. */
	String outputPath = "/data/experiments/southAfricaModel/";
	/** model parameter: seed for the random number generators. Default is 123456. */
	int seed = 123456;
	/** model parameter: tick to stop the simulation run. Default is 2400 (= 50 years) */
	int stopTime = 2400;

	
	//---- model fields ----------------------------------------
	
	/** list of agents (persons) */
	ArrayList<Person> personList;	// the list of agents (persons)
	/** list of households */
	ArrayList<Household> householdList;	// the list of households
	/** the 2D grid used as a representation of space */
	Object2DGrid space;
	/** the mine */
	Mine mine;
	/** the source of miners from outside the area */
	MinerSource minerSource;

	/** data structure to store the dates (week/month) of upcoming births at the beginning of a simulated year */ 
	int[][] nextBirths = new int[12][4]; // to store how many births will occur in which week of which month 
										 // at the beginning of a simulated year 
	
	public int year = 0, month = 0, week = 0, tick;
	/** counter for person IDs */
	public int personIndex = 0;

	/** data structure to map names to Person objects, helpful for the interface with Jess */
	HashMap<String, Person> namePersonMap;
	/** data structure to map names to Household objects, helpful for the interface with Jess */
	HashMap<String, Household> nameHouseholdMap;
	/** data structure to map names to Village objects, helpful for the interface with Jess */
	HashMap<String, Village> nameVillageMap;

	private Schedule schedule;
    
	/** the Jess engine */
	Rete engine;
	/** the class containing procedural functionality outsourced from Jess. Goes through the fact base every tick
	 * before the engine is run to update agents and households. */
	FactBaseScourer jessLink;
	/** Retracts facts when they become obsolete according to the specified filters ("disposal policies"). 
	 *  Run every tick after Jess has finished. */
	EndorsementDisposalPolicy disposalPolicy;
	/** Necessary for the interface with Jess to make changes in the slots (fields) of shadow facts (Java objects)
	 *  immediately visible in Jess. */
	private PropertyChangeSupport pcs = new PropertyChangeSupport(this);
	
	/** This inner class is a source of miners, i.e. it creates miners and passes them into the system 
	 * over the course of the simulation.
	 * Needs a distribution to determine arrival of miners (schedules itself for the next arrival time)
	 * and a distribution to determine how many miners arrive at any given arrival time.
	 */
	public class MinerSource extends BasicAction {
		
		int lastID = 0;
		Exponential arrivalDist;
		Normal numberDist;
		Normal ageDist;
		
		/** creates a new MinerSource */
		public MinerSource(){
			this.arrivalDist = new Exponential(1/interArrivalTime, new MersenneTwister(seed));
			this.numberDist = new Normal(meanNumArrived, sdNumArrived, new MersenneTwister(seed));
			this.ageDist = new Normal(meanAgeArrived, sdAgeArrived, new MersenneTwister(seed));
		}
		
		/** creates the specified number of miners and puts them into the model
		 * @param numMiners number of miners to create
		 * @return the list of miners
		 */
		public ArrayList<Person> createMiners(int numMiners){
			ArrayList<Person> miners = new ArrayList<Person>(numMiners);
			dump("creating " + numMiners + " miners");
			SouthAfricaModel.this.numMiners += numMiners;
			for (int i = 0; i < numMiners; i++){
				Person miner = new Person("miner-" + lastID++, null, SouthAfricaModel.this);
				// make miners at least 18 years old
				miner.setAge(Math.max(18, ageDist.nextInt()));	// was: randomInt(23,36)
				miner.setBirthMonth(randomInt(0,11));
				miner.setBirthWeek(randomInt(0,3));
				miner.setGender("male");
				miner.setAchievedEducation(randomInt(0,4)); // no tertiary education amongst the unskilled
				miner.setTagEvolutionPropensity (Random.uniform.nextDoubleFromTo(minTagEvolutionPropensity, maxTagEvolutionPropensity));
				miner.createTag();
				miner.setNode(new NetworkNode(miner.getName()));
				miner.addMineExperience(Math.min(randomInt(2,6), miner.age-16)*12*4);	// between 2 and 6 years of mine experience
				// assume HIV prevalence of 30% 
				if (probabilityp(mineHIVPrevalence)){
					// this miner is HIV positive
					// assign stage of disease but make sure they haven't proceeded to stage 4 (AIDS) yet
					miner.setInfectedSince(tick, Math.min(stageDist.nextInt(),3));
				}
				personList.add(miner);
				namePersonMap.put(miner.getName().toString(), miner);
				// put miner into Jess
				try {
					engine.definstance("person", miner, true);
				}
				catch(JessException je) {
					System.err.println("An error occurred at line " + je.getLineNumber());
					System.err.println("caused by " + je.getCause());
					System.err.println("Message: " + je.getDetail());
					je.printStackTrace();
				}
				miners.add(miner);
				// record arrival of miner
				miningRecorder.recordArrival(miner);
			}
			// re-schedule the miner source at the next arrival time
			int nextArrival = (int)getTickCount() + arrivalDist.nextInt() + 1;
			dump("next arrival of miners at tick " + nextArrival);
			schedule.scheduleActionAt(nextArrival, this, Schedule.CONCURRENT);
			return miners;
		}
		
		public void execute(){
			// create new miners 
			createMiners(numberDist.nextInt());
		}
		
	}
	
	//---- data collection stuff -----------------------------------

	PrintWriter printOut = null;
	File outputFile = null;
	String pathName;

	int numHHAskingForFood = 0, numHHWithoutEnoughFood = 0;
	int numMiners = 0;
	LocalDataRecorder monthlyDataRecorder, weeklyDataRecorder;
	PartnershipRecorder partnershipRecorder;
	SexualNetworkRecorder sexualNetworkRecorder;
	FriendshipNetworkRecorder friendNetworkRecorder;
	HIVDeathRecorder deathRecorder;
	ArrayList<String> chosenHH;
	HouseholdEconomyRecorder hhEconomyRecorder;
	VillageEconomyRecorder villageEconomyRecorder;
	HHWealthDistRecorder hhWealthDistRecorder;
	HHIncomeRecorder hhIncomeRecorder;
	StokvelRecorder stokvelRecorder;
	MiningRecorder miningRecorder;
	AgeDistributionRecorder ageDistRecorder;
	
	public class PartnershipRecorder {
		ArrayList<Integer> duration;
		ArrayList<Integer>[] age, numPartners;
		ArrayList<String>[] name, maritalStatus;
		DataFileWriter writer;
		String fileName;
		
		public PartnershipRecorder(String fileName, SouthAfricaModel model){
			duration = new ArrayList<Integer>();
			age = new ArrayList[]{new ArrayList<Integer>(), new ArrayList<Integer>()};
			numPartners = new ArrayList[]{new ArrayList<Integer>(), new ArrayList<Integer>()};
			name = new ArrayList[] {new ArrayList<String>(), new ArrayList<String>()};
			maritalStatus = new ArrayList[] {new ArrayList<String>(), new ArrayList<String>()};
			this.fileName = fileName;
			writer = new DataFileWriter(this.fileName, new DataFileHeader(model));
			writer.writeToFile("duration\tfemale\tage\tnumPartners\tmaritalStatus\tmale\tage\tnumPartners\tmaritalStatus");
		}
		
		public void record(Person[] partners, int ticks){
			duration.add(ticks);
			for (int i = 0; i < 2; i++){
				name[i].add(partners[i].getName().toString());
				age[i].add(partners[i].getAge());
				numPartners[i].add(partners[i].getMaxNumPartners());
				maritalStatus[i].add(partners[i].getMaritalStatus().toString());
			}
		}
		
		public void write(){
			for (int i = 0; i < duration.size(); i++){
				StringBuffer buf = new StringBuffer("" + duration.get(i));
				for (int j = 0; j < name.length; j++){
					buf.append("\t" + name[j].get(i));
					buf.append("\t" + age[j].get(i));
					buf.append("\t" + numPartners[j].get(i));
					buf.append("\t" + maritalStatus[j].get(i));
				}
				writer.writeToFile(buf.toString());
			}
			// delete recorded data
			duration.clear();
			for (int i = 0; i < name.length; i++){
				name[i].clear();
				age[i].clear();
				numPartners[i].clear();
				maritalStatus[i].clear();
			}
		}
	}
	
	public class HHIncomeRecorder {
		HashMap<String, Double> incomes;
		HashMap<String, Double> payOuts;
		HashMap<String, Double> donations;
		HashMap<String, Double> groupFees;
		HashMap<String, Double> schoolFees;
		HashMap<String, Double> foodCosts;
		HashMap<String, Double> pieceJobs;
		HashMap<String, Double> otherExp;
		
		public HHIncomeRecorder(){
			incomes = new HashMap<String, Double>();
			payOuts = new HashMap<String, Double>();
			groupFees = new HashMap<String, Double>();
			schoolFees = new HashMap<String, Double>();
			donations = new HashMap<String, Double>();
			foodCosts = new HashMap<String, Double>();
			pieceJobs = new HashMap<String, Double>();
			otherExp = new HashMap<String, Double>();
		}
		
		public void record(String hh, double income){
			double sum = 0;
			if (incomes.containsKey(hh)){
				sum = incomes.get(hh);
			}
			sum += income;
			incomes.put(hh, sum);
		}
		
		public void recordPayOut(String hh, double amount){
			double sum = 0;
			if (payOuts.containsKey(hh)){
				sum = payOuts.get(hh);
			}
			sum += amount;
			payOuts.put(hh, sum);
		}

		public void recordFee(String hh, double fee){
			double sum = 0;
			if (groupFees.containsKey(hh)){
				sum = groupFees.get(hh);
			}
			sum += fee;
			groupFees.put(hh, sum);
		}
		
		public void recordSchoolFee(String hh, double fee){
			double sum = 0;
			if (schoolFees.containsKey(hh)){
				sum = schoolFees.get(hh);
			}
			sum += fee;
			schoolFees.put(hh, sum);
		}

		public void recordFoodCost(String hh, double cost){
			double sum = 0;
			if (foodCosts.containsKey(hh)){
				sum = foodCosts.get(hh);
			}
			sum += cost;
			foodCosts.put(hh, sum);
		}

		public void recordDonation(String hh, double amount){
			double sum = 0;
			if (donations.containsKey(hh)){
				sum = donations.get(hh);
			}
			sum += amount;
			donations.put(hh, sum);
		}

		public void recordPieceJob(String hh, double amount){
			double sum = 0;
			if (pieceJobs.containsKey(hh)){
				sum = pieceJobs.get(hh);
			}
			sum += amount;
			pieceJobs.put(hh, sum);
		}
		
		public void recordOtherExpenditure(String hh, double amount){
			double sum = 0;
			if (otherExp.containsKey(hh)){
				sum = otherExp.get(hh);
			}
			sum += amount;
			otherExp.put(hh, sum);			
		}

		public double getMedianIncome(){
			// sort income values in ascending order
			Double[] values = new Double[incomes.size()];
			int i = 0;
			for (Iterator<String> it = incomes.keySet().iterator(); it.hasNext(); ){
				values[i++] = incomes.get(it.next());
			}
			if (values.length == 0) return 0.0;
			SortValues.mergesort(values, 0, values.length-1);
			// pick the middle one
			return values[(values.length-1)/2];
		}
		
		public double getMedianFees(){
			// sort fee values in ascending order
			Double[] values = new Double[groupFees.size()];
			int i = 0;
			for (Iterator<String> it = groupFees.keySet().iterator(); it.hasNext(); ){
				values[i++] = groupFees.get(it.next());
			}
			if (values.length == 0) return 0.0;
			SortValues.mergesort(values, 0, values.length-1);
			// pick the middle one
			return values[(values.length-1)/2];
		}
		
		public double getAccumulatedIncome(){
			double sum = 0.0;
			for (Iterator<String> it = incomes.keySet().iterator(); it.hasNext(); ){
				sum += incomes.get(it.next());
			}
			return sum;
		}

		public double getAccumulatedPayOuts(){
			double sum = 0.0;
			for (Iterator<String> it = payOuts.keySet().iterator(); it.hasNext(); ){
				sum += payOuts.get(it.next());
			}
			return sum;
		}
		
		public double getAccumulatedFoodCosts(){
			double sum = 0.0;
			for (Iterator<String> it = foodCosts.keySet().iterator(); it.hasNext(); ){
				sum += foodCosts.get(it.next());
			}
			return sum;
		}
		
		public double getAccumulatedGroupFees(){
			double sum = 0.0;
			for (Iterator<String> it = groupFees.keySet().iterator(); it.hasNext(); ){
				sum += groupFees.get(it.next());
			}
			return sum;
		}

		public double getAccumulatedSchoolFees(){
			double sum = 0.0;
			for (Iterator<String> it = schoolFees.keySet().iterator(); it.hasNext(); ){
				sum += schoolFees.get(it.next());
			}
			return sum;
		}

		public double getAccumulatedDonations(){
			double sum = 0.0;
			for (Iterator<String> it = donations.keySet().iterator(); it.hasNext(); ){
				sum += donations.get(it.next());
			}
			return sum;
		}

		public double getAccumulatedPieceJobs(){
			double sum = 0.0;
			for (Iterator<String> it = pieceJobs.keySet().iterator(); it.hasNext(); ){
				sum += pieceJobs.get(it.next());
			}
			return sum;
		}

		public double getAccumulatedOtherExp(){
			double sum = 0.0;
			for (Iterator<String> it = otherExp.keySet().iterator(); it.hasNext(); ){
				sum += otherExp.get(it.next());
			}
			return sum;
		}

		public void reset(){
			incomes.clear();
			groupFees.clear();
			payOuts.clear();
			schoolFees.clear();
			donations.clear();
			foodCosts.clear();
			pieceJobs.clear();
			otherExp.clear();
		}
	}
	
	public class HHWealthDistRecorder {
		DataFileWriter writer;
		String fileName;
		SouthAfricaModel model;
		int[] classes;
		ArrayList<String> data;
		
		public HHWealthDistRecorder(String fileName, SouthAfricaModel model){
			this(new int[]{0, 500, 1000, 1500, 2000, 2500, 3000}, fileName, model);
		}
		
		public HHWealthDistRecorder(int[] classes, String fileName, SouthAfricaModel model){
			this.classes = classes;
			this.fileName = fileName;
			this.model = model;
			this.data = new ArrayList<String>();
			writer = new DataFileWriter(this.fileName, new DataFileHeader(model));			
			StringBuffer header = new StringBuffer("tick");
			for (int i = 0; i < classes.length; i++){
				header.append("\t<=");
				header.append(classes[i]);
			}
			header.append("\t>");
			header.append(classes[classes.length-1]);
			writer.setBlockHeader(header.toString());			
		}
		
		public void record(){
			// compute distribution
			int[] dist = computeDistribution();
			// add it to data with current tick
			StringBuffer temp = new StringBuffer();
			temp.append(model.getTick());
			for (int i = 0; i < dist.length; i++){
				temp.append("\t");
				temp.append(dist[i]);
			}
			data.add(temp.toString());
		}
		
		private int[] computeDistribution(){
			int[] dist = new int[classes.length+1];
			for (int i = 0; i < dist.length; i++){
				dist[i] = 0;
			}
			for (Household hh : model.householdList){
				int index = determineClassOf(hh.getCash());
				dist[index]++;
			}
			return dist;
		}
		
		private int determineClassOf(double cash){
			int i = 0;
			if (cash < 0.0) cash = 0.0;
			while (i < classes.length && cash > classes[i]){
				i++;
			}
			return i;
		}
		
		public void write(){
			for (String line : data){
				writer.writeToFile(line);
			}
			// delete recorded data
			data.clear();
		}
	}
	
	public class SexualNetworkRecorder {
		ArrayList<HashMap<String, String[]>> data;
		ArrayList<Integer> recordedAt;
		DataFileWriter writer;
		String fileName;
		
		public SexualNetworkRecorder(String fileName, SouthAfricaModel model){
			data = new ArrayList<HashMap<String, String[]>>();
			recordedAt = new ArrayList<Integer>();
			this.fileName = fileName;
			writer = new DataFileWriter(this.fileName, new DataFileHeader(model));
		}
		
		public void record(){
			recordedAt.add(getTick());
			data.add(compileSexualNetwork());
		}
		
		public void write(){
			for (int i = 0; i < data.size(); i++){
				writer.writeToFile(recordedAt.get(i));
				HashMap<String, String[]> network = data.get(i);
				for (Iterator<String> it = network.keySet().iterator(); it.hasNext(); ){
					String person = it.next();
					String[] partners = network.get(person);
					for (String partner : partners){
						writer.writeToFile(person + "\t" + partner);
					}
				}
			}
			// delete recorded data
			recordedAt.clear();
			data.clear();
		}
		
		public void writeToPajek(){
			for (int i = 0; i < data.size(); i++){
				int recordedTick = recordedAt.get(i);
				writer.writeToFile(recordedTick);
				HashMap<String, String[]> network = data.get(i);
				// collect info on vertices
				HashMap<String, Boolean> verticesMale = new HashMap<String, Boolean>();
				HashMap<String, Boolean> verticesFemale = new HashMap<String, Boolean>();
				for (Iterator<String> it = network.keySet().iterator(); it.hasNext(); ){
					String personName = it.next();
					Person person = namePersonMap.get(personName);
					if (person.getGender().toString().equalsIgnoreCase("male")){
						if (person.getInfectedSince() >= 0 && person.getInfectedSince() <= recordedTick){
							verticesMale.put(personName, true);
						}
						else {
							verticesMale.put(personName, false);
						}
					}
					else {
						if (person.getInfectedSince() >= 0 && person.getInfectedSince() <= recordedTick){
							verticesFemale.put(personName, true);
						}
						else {
							verticesFemale.put(personName, false);
						}
					}
				}
				// write out vertices
				writer.writeToFile("*Vertices " + (verticesMale.size()+verticesFemale.size()) + " " + verticesMale.size());
				int id = 1;
				HashMap<String, Integer> vertices = new HashMap<String, Integer>();
				for (Iterator<String> it = verticesMale.keySet().iterator(); it.hasNext(); ){
					String name = it.next();
					boolean infected = verticesMale.get(name);
					writer.writeToFile(id + " " + name + " ic " + ((infected)? "Purple" : "RoyalBlue"));
					vertices.put(name, id++);
				}
				for (Iterator<String> it = verticesFemale.keySet().iterator(); it.hasNext(); ){
					String name = it.next();
					boolean infected = verticesFemale.get(name);
					writer.writeToFile(id + " " + name + " ic " + ((infected)? "BrickRed" : "Dandelion"));
					vertices.put(name, id++);
				}
				// write out edges
				writer.writeToFile("*Edges");
				for (Iterator<String> it = network.keySet().iterator(); it.hasNext(); ){
					String person = it.next();
					id = vertices.get(person);
					String[] partners = network.get(person);
					for (int j = 0; j < partners.length; j++){
						writer.writeToFile(id + " " + vertices.get(partners[j]));
					}
				}
			}
			// delete recorded data
			recordedAt.clear();
			data.clear();
		}
	}
	
	public class FriendshipNetworkRecorder {
		public String[] pajekMaleColours = {"Gray05","Gray10","Gray15","Gray20","Gray25","Gray30",
														"Gray35","Gray40","Gray45","Gray55","Gray60","Gray65",
														"Gray70","Gray75","Gray80","Gray85","Gray90","Gray95"};
		public String[] pajekFemaleColours = {"LightYellow","Canary","Yellow","Goldenrod","Dandelion","YellowOrange",
														  "BurntOrange","Peach","Orange","RedOrange","Red","Bittersweet",
														  "BrickRed","Maroon","Mahogany","RawSienna","Brown","Sepia"};
		
		HashMap<String, ArrayList<String>> network;
		HashMap<String, Integer> vertices;
		ArrayList<String> males;
		
		String fileName;
		SouthAfricaModel model;
		
		public FriendshipNetworkRecorder(String fileName, SouthAfricaModel model){
			this.model = model;
			this.fileName = fileName;
			males = new ArrayList<String>();
		}
		
		public void record(){
			// record network
			network = new HashMap<String, ArrayList<String>>();
			vertices = new HashMap<String, Integer>();
			int id = 1;
			for (Person p : model.personList){
				String pName = p.getName().toString();
				vertices.put(pName, id++);
				if (p.gender.toString().equalsIgnoreCase("male") && !males.contains(pName)){
					males.add(pName);
				}
				ArrayList<String> friends = new ArrayList<String>();
				for (Value friendName : p.friends){
					friends.add(friendName.toString());
				}
				network.put(pName, friends);
			}
		}
		
		public void writeToPajek(){
			DataFileWriter writer = new DataFileWriter(fileName + model.getTick() + ".net", new DataFileHeader(""));
			// write vertices with gender/age encoded in colour/hue
			writer.writeToFile("*Vertices " + vertices.size());
			for (Iterator<String> it = vertices.keySet().iterator(); it.hasNext(); ){
				String vertex = it.next();
				String[] colourTable = (males.contains(vertex) ? pajekMaleColours : pajekFemaleColours);
				int colourIndex = model.namePersonMap.get(vertex).getAge() / 5;
				if (colourIndex > colourTable.length - 1)
					colourIndex = colourTable.length -1 ;
				writer.writeToFile(vertices.get(vertex) + " " + vertex + " ic " + colourTable[colourIndex]);
			}
			// sort arcs and edges, storing edges and writing arcs directly 
			ArrayList<int[]> edges = new ArrayList<int[]>();
			writer.writeToFile("*Arcs");
			for (Iterator<String> it = network.keySet().iterator(); it.hasNext(); ){
				String from = it.next();
				int fromID = vertices.get(from);
				for (String to : network.get(from)){
					if (vertices.get(to) == null){
						// dead friend encountered!
						dump("\tfriendshipNetworkRecorder found a dead friend (" + to + ") for " + from);
					}
					else {
						int toID = vertices.get(to);
						// mutual link?
						if (network.get(to).contains(from)){
							// it's an edge; make sure to only record it once
							if (fromID < toID){
								edges.add(new int[]{fromID, toID});
							}
						}
						else {
							// it's an arc
							writer.writeToFile(fromID + " " + toID + " 1");
						}
					}
				}
			}
			// write edges
			writer.writeToFile("*Edges");
			for (int[] edge : edges){
				writer.writeToFile(edge[0] + " " + edge[1] + " 1");
			}
		}
	}
	
	public class HouseholdEconomyRecorder {
		HashMap<String, ArrayList<String>> in, out;
		HashMap<String, DataFileWriter> writers;
		ArrayList<String> hhNames;
		SouthAfricaModel model;
		
		public HouseholdEconomyRecorder(ArrayList<String> hhNames, String fileName, SouthAfricaModel model){
			in = new HashMap<String, ArrayList<String>>();
			out = new HashMap<String, ArrayList<String>>();
			writers = new HashMap<String, DataFileWriter>();
			this.model = model;
			for (String hh : hhNames){
				in.put(hh, new ArrayList<String>());
				out.put(hh, new ArrayList<String>());
				writers.put(hh, new DataFileWriter(fileName + hh + ".txt", new DataFileHeader(model)));
			}
			this.hhNames = hhNames;
		}
		
		public void recordIn(String hh, String cause, double amount){
			ArrayList<String> inList = in.get(hh);
			record(inList, cause, amount);
		}
		
		public void recordOut(String hh, String cause, double amount){
			ArrayList<String> outList = out.get(hh);
			record(outList, cause, amount);
		}
		
		private void record(ArrayList<String> list, String cause, double amount){
			list.add(model.tick + "\t" + cause + "\t" + amount);
		}
		
		public void writeHHInfo(){
			for (String hh : hhNames){
				writeHHInfo(hh);
			}
		}
		private void writeHHInfo(String hhName){
			Household hh = model.nameHouseholdMap.get(hhName);
			DataFileWriter writer = writers.get(hhName);
			writer.writeToFile("\n" + hh.getId().toString());
			writer.writeToFile("Members: ");
			for (Person p : hh.getMembers()){
				writer.writeToFile(p.getName() + " (" + p.getAge() + ")" + 
						(p.getGender().toString().equalsIgnoreCase("female")? ", female" : ", male") +
						(p.isMigrant()? ", migrant" : "") + 
						(p.isAtSchool()? ", at school" : ""));
			}
			
			writer.writeToFile("in\t\t\tout\t\t");
			writer.writeToFile("tick\tcause\tamount\ttick\tcause\tamount");
		}
		
		public void write(){
			for (String hh : hhNames){
				write(writers.get(hh), in.get(hh), out.get(hh));
			}
		}
		
		private void write(DataFileWriter writer, ArrayList<String> in, ArrayList<String> out){
			int i = 0;
			while (i < in.size() && i < out.size()){
				// write a line of in and out
				writer.writeToFile(in.get(i) + "\t" + out.get(i));
				i++;
			}
			while (i < in.size() && i >= out.size()){
				// write a line of in
				writer.writeToFile(in.get(i));
				i++;
			}
			while (i >= in.size() && i < out.size()){
				// write a line of out
				writer.writeToFile("\t\t\t" + out.get(i));
				i++;
			}
			// delete recorded data
			in.clear();
			out.clear();
		}
	}
	
	public class VillageEconomyRecorder {
		HashMap<String, HashMap<String, Double>> matrix;
		ArrayList<String> hhNames;
		DataFileWriter writer;
		SouthAfricaModel model;
//		String fileName;
		
		public VillageEconomyRecorder(String fileName, SouthAfricaModel model){
			this.model = model;
			writer = new DataFileWriter(fileName, new DataFileHeader(model));
			// store household names for easy access
			hhNames = new ArrayList<String>();
			for (Household hh : model.householdList){
				hhNames.add(hh.getId().toString());
			}
			hhNames.add("groups");
			hhNames.add("outside");
			// initialise matrix
			matrix = new HashMap<String, HashMap<String, Double>>();
			for (String hh : hhNames){
				HashMap<String, Double> row = new HashMap<String, Double>();
				for (String hh2 : hhNames){
					row.put(hh2, 0.0);
				}
				matrix.put(hh, row);
			}
		}
		
		public void record(String from, String to, double amount){
			matrix.get(from).put(to, matrix.get(from).get(to) + amount);
		}
		
		public void write(){
			writer.writeToFile("\n" + model.tick);
			HashMap<String, Double> colSums = new HashMap<String, Double>();
			StringBuffer temp = new StringBuffer();
			for (String hh : hhNames){
				temp.append("\t" + hh);
				colSums.put(hh, 0.0);
			}
			temp.append("\ttotal outgoings");
			writer.writeToFile(temp.toString());
			for (String fromHH : hhNames){
				double rowSum = 0;
				temp = new StringBuffer(fromHH);
				for (String toHH : hhNames){
					temp.append("\t");
					double amount = matrix.get(fromHH).get(toHH);
					temp.append(amount);
					rowSum += amount;
					colSums.put(toHH, colSums.get(toHH) + amount);
				}
				temp.append("\t" + rowSum);
				writer.writeToFile(temp.toString());
			}
			temp = new StringBuffer("total income");
			for (String hh : hhNames){
				temp.append("\t");
				temp.append(colSums.get(hh));
			}
			writer.writeToFile(temp.toString());
			resetMatrix();
		}
		
		private void resetMatrix(){
			for (String fromHH : hhNames){
				for (String toHH : hhNames){
					matrix.get(fromHH).put(toHH, 0.0);
				}
			}
		}
		
		public void addHousehold(String name){
			// add column to every row
			for (String hh : hhNames){
				HashMap<String, Double> row = matrix.get(hh);
				row.put(name, 0.0);
			}
			// add new row
			hhNames.add(0, name);
			HashMap<String, Double> row = new HashMap<String, Double>();
			for (String hh : hhNames){
				row.put(hh, 0.0);
			}
			matrix.put(name, row);
		}
	}
	
	public class HIVDeathRecorder {
		ArrayList<Integer> infectedTick, aidsTick, deathTick, age;
		ArrayList<String> gender;
		DataFileWriter writer;
		String fileName;
		
		public HIVDeathRecorder(String fileName, SouthAfricaModel model){
			infectedTick = new ArrayList<Integer>();
			aidsTick = new ArrayList<Integer>();
			deathTick = new ArrayList<Integer>();
			age = new ArrayList<Integer>();
			gender = new ArrayList<String>();
			this.fileName = fileName;
			writer = new DataFileWriter(this.fileName, new DataFileHeader(model));
			writer.writeToFile("deathTick\tinfectedTick\taidsTick\tage\tgender");
		}
		
		public void record(Person p){
			age.add(p.age);
			if (p.isHIVPositive()){
				infectedTick.add(p.hivAIDS.getInfectedSince());
			}
			else {
				infectedTick.add(-1);
			}
			if (p.getHasAIDS()){
				aidsTick.add(p.hivAIDS.getDevelopedAIDS());
			}
			else {
				aidsTick.add(-1);
			}
			deathTick.add(getTick());
			gender.add(p.gender.toString());
		}
		
		public void write(){
			for (int i = 0; i < age.size(); i++){
				writer.writeToFile(deathTick.get(i) + "\t" + infectedTick.get(i) 
									+ "\t" + aidsTick.get(i) + "\t" + age.get(i) + "\t" + gender.get(i));
			}
			// delete recorded data
			age.clear();
			gender.clear();
			deathTick.clear();
			infectedTick.clear();
		}
	}
	
	public class StokvelRecorder {
		HashMap<String, DataFileWriter> writers;
		HashMap<String, Stokvel> stokvels;
		String fileName;
		SouthAfricaModel model;
		
		public StokvelRecorder(String fileName, SouthAfricaModel model){
			this.fileName = fileName;
			this.model = model;
			this.stokvels = new HashMap<String, Stokvel>();
			this.writers = new HashMap<String, DataFileWriter>();
		}
		
		public void add(Stokvel stokvel){
			// create writer
			String name = stokvel.getName().toString();
			DataFileWriter writer = new DataFileWriter(fileName + stokvels.size() + ".txt", new DataFileHeader(model));
			writers.put(name, writer);
			stokvels.put(name, stokvel);
			// write out initial stokvel information
			writer.writeToFile(name + " created at tick " + model.getTick() + " by: ");
			for (Person p : stokvel.members){
				writer.writeToFile(p.getName() + ", " + p.getGender() + ", " + p.getAge() + " of " + p.getHouseholdID());
				recordFriends(writer, stokvel, p);
			}
			writer.writeToFile("monthly fee is " + stokvel.getMonthlyFee());
		}
		
		public void recordEvent(String stokvel, String event){
			writers.get(stokvel).writeToFile(event);
		}
		
		public void recordEvent(String stokvel, String event, Person member){
			recordEvent(stokvel, event);
			// record friendships too
			recordFriends(writers.get(stokvel), stokvels.get(stokvel), member);
		}
		
		private void recordFriends(DataFileWriter writer, Stokvel stokvel, Person p){
			ArrayList<Person> friends = p.getFriends();
			StringBuffer friendsList = new StringBuffer(p.getName() + "\tis friends with ");
			for (Person member : stokvel.members){
				if (member != p && friends.contains(member)){
					friendsList.append(member.getName());
					friendsList.append(", ");
				}
			}
			writer.writeToFile(friendsList.toString());
		}
		
	}
	
	public class MiningRecorder {
		DataFileWriter writer;
		String fileName;
		SouthAfricaModel model;
		HashMap<String, Integer> miners; // to store arrival tick of miner
		
		public MiningRecorder(String fileName, SouthAfricaModel model){
			this.fileName = fileName;
			this.model = model;
			this.miners = new HashMap<String, Integer>();
			this.writer = new DataFileWriter(fileName, new DataFileHeader(model));
			this.writer.writeToFile("duration\tname\tage\texperience\tAIDS");
		}
		
		public void recordArrival(Person miner){
			miners.put(miner.getName().toString(), tick);
		}
		
		public void recordDeparture(Person miner){
			StringBuffer buf = new StringBuffer();
			String name = miner.getName().toString();
			buf.append((tick - miners.get(name)) + "\t" + name + "\t");
			buf.append(miner.getAge() + "\t" + miner.getMineExperience() + "\t");
			buf.append((miner.getHasAIDS() ? 1 : 0));
			this.writer.writeToFile(buf.toString());
		}
	}
	
	public class AgeDistributionRecorder {
		DataFileWriter writer;
		SouthAfricaModel model;
		int[] lowerAgeBounds = new int[]{0,5,10,15,20,25,30,35,40,45,50,55,60,65,70,75,80,85};
		
		public AgeDistributionRecorder(String fileName, SouthAfricaModel model){
			this.model = model;
			this.writer = new DataFileWriter(fileName, new DataFileHeader(model));
			writer.writeToFile("tick\tgender\t0-4\t'5-9\t'10-14\t15-19\t20-24\t25-29\t30-34\t35-39\t40-44\t45-49\t" +
					"50-54\t55-59\t60-64\t65-69\t70-74\t75-79\t80-84\t85+");
		}
		
		public void recordAndWrite(){
			int[] males = new int[lowerAgeBounds.length];
			int[] females = new int[lowerAgeBounds.length];
			// record all persons except miners
			for (Person p : personList){
				if (p.getName().toString().startsWith("person")){
					int index = indexOf(p.getAge());
					if (p.getGender().toString().equalsIgnoreCase("male")){
						males[index]++;
					}
					else {
						females[index]++;
					}
				}
			}
			// write age distributions to file
			writeDistribution(males, "male");
			writeDistribution(females, "female");
		}
		
		private int indexOf(int age){
			int index = 0;
			while (index < lowerAgeBounds.length && age >= lowerAgeBounds[index]){
				index++;
			}
			return index-1;
		}
		
		private void writeDistribution(int[] dist, String gender){
			StringBuffer buf = new StringBuffer(getTick() + "\t" + gender);
			for (int i = 0; i < dist.length; i++){
				buf.append("\t");
				buf.append(dist[i]);
			}
			writer.writeToFile(buf.toString());
		}
	}

	
	
	//---- gui stuff --------------------------------------------
	
	DisplaySurface friendNetworkDisplay, kinshipNetworkDisplay, stokvelNetworkDisplay, spaceDisplay;
	MultiLayerNetwork2DDisplay friendNetwork, kinshipNetwork, stokvelNetwork;
	
	//---- statistical stuff -----------------------------------
	
	/** This inner class defines a discrete distribution of empirical values, e.g. the marital status of females.
	 *  The distribution does not store the actual values (e.g. single, married, divorced, widowed) but replaces them
	 *  with integer numbers from 0 to n-1, to be used as indices into the array of values. The user has to
	 *  specify the probabilities for the values and a random number generator.
	 *  @author Ruth
	 */
	class EmpiricalDiscreteDist{
		RandomEngine randomGenerator;
		double[] probabilities;
		
		public EmpiricalDiscreteDist(double[] probabilities, RandomEngine randomGenerator){
			this.randomGenerator = randomGenerator;
			this.probabilities = new double[probabilities.length]; 
			// turn probs into cumulative probs
			this.probabilities[0] = probabilities[0];
			for (int i = 1; i < probabilities.length; i++){
				this.probabilities[i] = this.probabilities[i-1] + probabilities[i]; 
			}			
			// needs to check that all the probs add up to 1
			// for now, let's just assume it's correct
		}
		
		/** returns a new sample, i.e. index value between 0 and n-1 */
		public int nextInt(){
			int index = 0;
			double randomNumber = randomGenerator.nextDouble();
			while (index < this.probabilities.length && randomNumber > this.probabilities[index]){
				index++;
			}
			return index;
		}
	}
	
	/** An array of discrete distributions of empirical values. Each distribution is associated with a particular
	 * interval. This class is used in the model for the distribution of education levels according to age groups. 
	 * The user has to specify a set of intervals (age groups), the probabilities for each distribution and a
	 * random number generator.
	 *  @author Ruth
	 */
	class EmpiricalDiscreteDistArray {
		EmpiricalDiscreteDist[] distArray;
		double[] intervals;
		
		public EmpiricalDiscreteDistArray(double[] intervals, double[][] probabilities, RandomEngine randomGenerator){
			this.intervals = intervals;
			this.distArray = new EmpiricalDiscreteDist[intervals.length];
			for (int i = 0; i < intervals.length; i++){
				distArray[i] = new EmpiricalDiscreteDist(probabilities[i], randomGenerator);
			}
		}
		
		/** returns a new sample from the distribution associated with the interval containing x */
		public int nextInt(double x){
			// determine in which interval x falls
			int index = 0;
			while (index < intervals.length && x > intervals[index]){
				index++;
			}
			// return random number from x's distribution
			return distArray[index].nextInt();
		}
	}
	
	/** This class represents a life table as produced by the WHO. See:
	 * http://www.who.int/whosis/database/life_tables/life_tables.cfm
	 * Only the column nqx (probability to die within the current age range) is needed. 
	 * The class is used to determine deaths of agents in the model.
	 * @author Ruth
	 *
	 */
	class WHOLifeTable{
		int[] ageIntervals;  // the age range classes, given as (inclusive) lower bounds 
		double[] deathProbabilities;  // the probability to die within the current age range (nqx)
		
		public WHOLifeTable(int[] ageIntervals, double[] deathProbabilities){
			this.ageIntervals = ageIntervals;
			this.deathProbabilities = deathProbabilities;
		}
		
		/** returns the death probability for the given age */
		public double getDeathProbability(int age){
			// lookup age in ageIntervals, then use that index to find corresponding death probability
			int index = findAgeInterval(age);
			if (index >= 0 ) return this.deathProbabilities[index];
			return 1.0;
		}
		
		/** returns the size of the age interval into which the given age falls */
		public int getAgeIntervalSize(int age){
			int i = findAgeInterval(age);
			if (i == this.ageIntervals.length-1) return 3; // assume last interval has size 3
			else if (i >= 0) return (this.ageIntervals[i+1] - this.ageIntervals[i]);
			return -1;
		}
		
		/** returns true if the given age is the lower bound of an age range, false otherwise. */
		public boolean isLowerBound(int age){
			int i = findAgeInterval(age);
			if (i > 0 && this.ageIntervals[i] == age) return true;
			return false;
		}
		
		/** returns the age interval into which the given age falls. */
		public int findAgeInterval(int age){
			if (age < 0) return -1;
			// linear search
			int index = 0;
			while (index < this.ageIntervals.length 
					&& this.ageIntervals[index] <= age){
				index++;
			}
			return index-1;
		}
	}
	
	// life tables for South Africa are taken from:
	// http://www3.who.int/whosis/life/life_tables/life_tables_process.cfm?country=zaf&language=en#

	WHOLifeTable malesTable = new WHOLifeTable(
			new int[]{0, 1, 5, 10, 15, 20, 25, 30, 35, 40, 45, 50, 55, 60, 65, 
					70, 75, 80, 85, 90, 95, 100},
			new double[]{
			0.05769,
			0.01474,
			0.01903,
			0.00694,
			0.01451,
			0.037,
			0.05493,
			0.11001,
			0.1725,
			0.14059,
			0.15916,
			0.167,
			0.16182,
			0.19377,
			0.23759,
			0.31431,
			0.4283,
			0.57377,
			0.73448,
			0.82419,
			0.87064,
			1});
	
	WHOLifeTable femalesTable = new WHOLifeTable(
			new int[]{0, 1, 5, 10, 15, 20, 25, 30, 35, 40, 45, 50, 55, 60, 65, 
					70, 75, 80, 85, 90, 95, 100},
			new double[]{
			0.0502,
			0.01273,
			0.01813,
			0.00632,
			0.01167,
			0.04169,
			0.08878,
			0.13015,
			0.14072,
			0.13819,
			0.10376,
			0.09637,
			0.10613,
			0.1262,
			0.16889,
			0.24338,
			0.35132,
			0.49935,
			0.66916,
			0.78066,
			0.84565,
			1});
	
	// assign clinical stage of HIV infection according to following distribution (adapted from Wawer et al.)
	// durations are 1: 3 months, 2: 12 months, 3: 96 months, 4: 14 months 
	// => probabilities 0.024, 0.096, 0.768, 0.112
	EmpiricalDiscreteDist stageDist = new EmpiricalDiscreteDist(new double[]{0.0, 0.024, 0.096, 0.768, 0.112}, new MersenneTwister(seed));
	
	
	//---- methods ----------------------------------------------------------------------------------------
	
	//---- REPAST STUFF ---------------------------
	
	/** required by Repast: resets the model */
	public void setup() {
		if (printOut != null){
			printOut.flush();
			printOut.close();
		}
		if (engine != null){
			try {
				engine.clear();
			}
			catch (JessException je){
				System.err.println("Jess engine.clear() didn't work!");
				System.err.println("An error occurred at line " + je.getLineNumber());
				System.err.println("caused by " + je.getCause());
				System.err.println("Message: " + je.getDetail());
				je.printStackTrace();
			}
		}
		engine = null;
		personList = null;
		householdList = null;
		schedule = null;
		friendNetworkDisplay = null;
		kinshipNetworkDisplay = null;
		stokvelNetworkDisplay = null;

		System.gc();

		engine = new Rete();
		
		Random.createUniform();
		schedule = new Schedule(1);
		personList = new ArrayList<Person>();
		householdList = new ArrayList<Household>();
	}
		
	/** required by Repast: returns a list of model parameters, which can be manipulated via the Repast GUI */
	public String[] getInitParam() {
		String[] params = {"numHouseholds", "numVillages", "numDenominations", "churchParticipationRate",
				"birthRate", "adultEmploymentRate", "governmentEmploymentRate", "houseCost",
				"standardWage", "childGrant", "statePension", "burialCost", "shackProportion",
				"minShackRent", "maxShackRent",
				"childGrantAgeLimit", "pensionAgeLimitFemale", "pensionAgeLimitMale",
				"childGrantProportion", "pensionProportionFemale", "pensionProportionMale",
				"foodCostChild", "foodCostFemaleAdult", "foodCostMaleAdult",
				"schoolFee", "collegeFee", "minLobola", "maxLobola",
				"minRemittance", "maxRemittance", "remittanceProbability",
				"gridSizeX", "gridSizeY", "densityFactor", "neighbourhoodRadius", 
				"tagLength", "tagBase", "maxTagEvolutionPropensity", "minTagEvolutionPropensity", 
				"minEndorsementBase", "maxEndorsementBase", "minEndorsementClasses", "maxEndorsementClasses",
				"lowerMaxNumFriends", "upperNumFriends", "lowerNumPartners", "upperNumPartners",
				"similarityAgeRange", "randomPartnerProbability", "motherChildHIVTransmissionRate",
				"miningTick", "neededSkilled", "neededUnskilled", "mineHIVPrevalence", 
				"interArrivalTime", "meanNumArrived", "sdNumArrived", "meanAgeArrived", "sdAgeArrived",
				"showGUI", "printToFile", "showOutput", "outputPath", "seed", "stopTime"};
		return params;
	}
	
	
	/** required by Repast: creates all model components and the schedule */
	public void begin() {
		pathName = outputPath + "run-" + System.currentTimeMillis() + "/";
		File temp = new File(pathName);
		if (!temp.exists()){
			if (temp.mkdir()){
				System.out.println("OUTPUT directory created");
			}
			else {
				System.out.println("FAILED to create OUTPUT directory");
			}
		}
		else {
			System.out.println("OUTPUT directory exists already");
		}
		if (printToFile) openOutput();
		buildModel();
		if (showGUI) buildDisplay();
		buildSchedule();
		
		// add model as shadow fact
		try {
			engine.definstance("model", this, true);
		}
		catch(JessException je) {
			System.err.println("An error occurred at line " + je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}	
	}

	/** helper method: opens trace file and directs Jess ouput to it */
	private void openOutput(){
		outputFile = new File(pathName + "trace.txt");
		try {
			printOut = new PrintWriter(new FileWriter(outputFile));
			engine.addOutputRouter("t", printOut);
			engine.addOutputRouter("WSTDOUT", printOut);
		}
		catch (IOException e){
			System.out.println(e.getMessage());
			System.out.println(e.getStackTrace());
			System.exit(-1);
		}
	}
	
	/** closes trace file on finalisation of model object */
	public void finalize(){
		if (printOut != null){
			printOut.flush();
			printOut.close();
		}
		printOut = null;
		outputFile = null;
	}

	/** recommended by Repast: initialises the model */
	private void buildModel(){
		// init model components
		space = new Object2DGrid(gridSizeX, gridSizeY);
		personList = new ArrayList<Person>();
		namePersonMap = new HashMap<String, Person>();
		householdList = new ArrayList<Household>();
		nameHouseholdMap = new HashMap<String, Household>();
		nameVillageMap = new HashMap<String, Village>();
		jessLink = new FactBaseScourer(this);
		RulebaseCreator rc = new RulebaseCreator(this);
		rc.initialise();
		setDisposalPolicy();
		
		// data collection 
		hhIncomeRecorder = new HHIncomeRecorder();
		monthlyDataRecorder = new LocalDataRecorder(pathName + "OutputMonthly.txt", this);
		monthlyDataRecorder.setDelimeter("\t");
		monthlyDataRecorder.createNumericDataSource("HIVIncidence", this, "getHIVIncidence");
		monthlyDataRecorder.createNumericDataSource("numHouseholds", this, "getCurrentNumHouseholds");
		monthlyDataRecorder.createNumericDataSource("numHHAskingForFood", this, "getNumHHAskingForFood");
		monthlyDataRecorder.createNumericDataSource("numHHWithoutEnoughFood", this, "getNumHHWithoutEnoughFood");
		monthlyDataRecorder.createNumericDataSource("maxHHCash", this, "getMaxHHCash"/*, 7, 2*/);
		monthlyDataRecorder.createNumericDataSource("minHHCash", this, "getMinHHCash"/*, 7, 2*/);
		monthlyDataRecorder.createNumericDataSource("avgHHCash", this, "getAvgHHCash"/*, 7, 2*/);
		monthlyDataRecorder.createNumericDataSource("medianHHIncome", hhIncomeRecorder, "getMedianIncome"/*, 7, 2*/);
		monthlyDataRecorder.createNumericDataSource("medianGroupFees", hhIncomeRecorder, "getMedianFees"/*, 7, 2*/);
		monthlyDataRecorder.createNumericDataSource("accHHIncome", hhIncomeRecorder, "getAccumulatedIncome"/*, 7, 2*/);
		monthlyDataRecorder.createNumericDataSource("accPayouts", hhIncomeRecorder, "getAccumulatedPayOuts"/*, 7, 2*/);
		monthlyDataRecorder.createNumericDataSource("accFoodCosts", hhIncomeRecorder, "getAccumulatedFoodCosts"/*, 7, 2*/);
		monthlyDataRecorder.createNumericDataSource("accGroupFees", hhIncomeRecorder, "getAccumulatedGroupFees"/*, 7, 2*/);
		monthlyDataRecorder.createNumericDataSource("accSchoolFees", hhIncomeRecorder, "getAccumulatedSchoolFees"/*, 7, 2*/);
		monthlyDataRecorder.createNumericDataSource("accDonations", hhIncomeRecorder, "getAccumulatedDonations"/*, 7, 2*/);
		monthlyDataRecorder.createNumericDataSource("accPieceJobs", hhIncomeRecorder, "getAccumulatedPieceJobs"/*, 7, 2*/);
		monthlyDataRecorder.createNumericDataSource("accOtherExp", hhIncomeRecorder, "getAccumulatedOtherExp"/*, 7, 2*/);
		
		weeklyDataRecorder = new LocalDataRecorder(pathName + "OutputWeekly.txt", this);
		weeklyDataRecorder.setDelimeter("\t");
		weeklyDataRecorder.createNumericDataSource("numHIVPositives", this, "getNumHIVPositives");
		weeklyDataRecorder.createNumericDataSource("numFemaleHIVPositives", this, "getNumFemaleHIVPositives");
		if (this.miningTick >= 0){
			weeklyDataRecorder.createNumericDataSource("numHIVPositiveMiners", this, "getNumMinerHIVPositives");
			weeklyDataRecorder.createNumericDataSource("numMiners", this, "getNumMiners");
			weeklyDataRecorder.createNumericDataSource("numUnskilledWorkers", this, "getNumUnskilled");
			weeklyDataRecorder.createNumericDataSource("numEmployedMiners", this, "getNumEmployedMiners");
		}
		weeklyDataRecorder.createNumericDataSource("numPersons", this, "getNumPersons");
		weeklyDataRecorder.createNumericDataSource("numMigrants", this, "getNumMigrants");
		weeklyDataRecorder.createNumericDataSource("numVisitingMigrants", this, "getNumMigrantsOnVisit");
		weeklyDataRecorder.createNumericDataSource("numStokvelMembers", this, "getNumStokvelMembers");
		weeklyDataRecorder.createNumericDataSource("numBSMembers", this, "getNumBSMembers");
		weeklyDataRecorder.createNumericDataSource("friendNetClusterCoefficient", this, "calculateClusterCoefficient");
		
		sexualNetworkRecorder = new SexualNetworkRecorder(pathName + "SexualNetwork.txt", this);
		friendNetworkRecorder = new FriendshipNetworkRecorder(pathName + "FriendNetwork-", this);
		partnershipRecorder = new PartnershipRecorder(pathName + "Partnerships.txt", this);
		deathRecorder = new HIVDeathRecorder(pathName + "HIVDeaths.txt", this);
		hhWealthDistRecorder = new HHWealthDistRecorder(pathName + "HHWealthDist.txt", this);
		stokvelRecorder = new StokvelRecorder(pathName + "StokvelHistory", this);
		ageDistRecorder = new AgeDistributionRecorder(pathName + "AgeDistributions.txt", this);
		if (miningTick >= 0) miningRecorder = new MiningRecorder(pathName + "Mining.txt", this);
		
		// create agents and their environment
		System.out.println("------- model initialisation -------");
		System.out.print("creating households... ");
		createHouseholds();
		System.out.print("done.\ncreating kinship network and clusters... ");
		ArrayList<ArrayList<NetworkNode>> clusters = createKinshipNetwork();
		System.out.print("done.\ncreating villages, locating households and creating churches... ");
		locateHouseholds(clusters);
		System.out.print("done.\ncreating neighbour network... ");
		createNeighbourNetwork();
		System.out.print("done.\npicking migrants... ");
		pickMigrants();
		System.out.print("done.\nassigning education levels...");
		assignEducationLevels();
		System.out.print("done.\nassigning employments...");
		assignEmployments();
		System.out.print("done.\ncreating burial societies...");
		createBurialSocieties();
		System.out.print("done.\npicking initial HIV positives...");
		pickHIVPositives();
		System.out.println("done.\n------- initialisation completed -------\n");
		
		// pick some households to be monitored for their income/expenses
		ArrayList<Household> chosen = new ArrayList<Household>(3);
		chosen.add(householdList.get(6));
		chosen.add(householdList.get(13));
		chosen.add(householdList.get(21));
		chosenHH = new ArrayList<String>(3);
		for (Household hh : chosen){
			chosenHH.add(hh.getId().toString());
		}
		hhEconomyRecorder = new HouseholdEconomyRecorder(chosenHH, pathName + "HHEconomy", this);
		hhEconomyRecorder.writeHHInfo();
		for (int i = 0; i < chosen.size(); i++) {
			this.hhEconomyRecorder.recordIn(chosenHH.get(i), "initial cash", chosen.get(i).getCash());
		}
		
		villageEconomyRecorder = new VillageEconomyRecorder(pathName + "VillageEconomy.txt", this);
		// record initial cash for all households
		for (Household hh : householdList){
			villageEconomyRecorder.record("outside", hh.getId().toString(), hh.getCash());
		}
		villageEconomyRecorder.write();
	}
	
	/** recommended by Repast: initialises the model GUI */
    private void buildDisplay() {
    	int xSize = 800, ySize = 600;

		friendNetworkDisplay = new DisplaySurface(this, "Friend Network Display");
		registerDisplaySurface("Friend Network Display", friendNetworkDisplay);
		kinshipNetworkDisplay = new DisplaySurface(this, "Kinship Network Display");
		registerDisplaySurface("Kinship Network Display", kinshipNetworkDisplay);
		stokvelNetworkDisplay = new DisplaySurface(this, "Stokvel Network Display");
		registerDisplaySurface("Stokvel Network Display", stokvelNetworkDisplay);
		spaceDisplay = new DisplaySurface(this, "Space Display");
		registerDisplaySurface("Space Display", spaceDisplay);
		
		Object2DDisplay hhInSpace = new Object2DDisplay(this.space);
		hhInSpace.setObjectList(this.householdList);
		spaceDisplay.addDisplayableProbeable(hhInSpace, "Space");
		addSimEventListener(spaceDisplay);
		spaceDisplay.setBackground(Color.LIGHT_GRAY);

		RandomGraphLayout friendLayout = new RandomGraphLayout(getNodes(personList), xSize, ySize);
		friendNetwork = new MultiLayerNetwork2DDisplay(friendLayout, "friend");
		friendNetworkDisplay.addDisplayableProbeable(friendNetwork, "Friends");
		addSimEventListener(friendNetworkDisplay);
		friendNetworkDisplay.setBackground(Color.white);

		kinshipNetwork = new MultiLayerNetwork2DDisplay(getNodes(householdList), xSize, ySize, "kinship");
		kinshipNetworkDisplay.addDisplayableProbeable(kinshipNetwork, "Kinship");
		addSimEventListener(kinshipNetworkDisplay);	
		kinshipNetworkDisplay.setBackground(Color.white);
		
		RandomGraphLayout stokvelLayout = new RandomGraphLayout(getNodes(personList), xSize, ySize);
		stokvelNetwork = new MultiLayerNetwork2DDisplay(stokvelLayout, "stokvel");
		stokvelNetworkDisplay.addDisplayableProbeable(stokvelNetwork, "Stokvels");
		addSimEventListener(stokvelNetworkDisplay);
		stokvelNetworkDisplay.setBackground(Color.white);
		

		friendLayout.updateLayout();
		stokvelLayout.updateLayout();
		
		friendNetworkDisplay.display();
		kinshipNetworkDisplay.display();
		stokvelNetworkDisplay.display();
		spaceDisplay.display();
    }  
	
	/** helper method for buildDisplay(): returns a list of nodes for the given agents */
	private ArrayList getNodes(ArrayList agents){
		ArrayList nodes = new ArrayList();
		for (Object agent : agents){
			if (agent instanceof Person) {nodes.add(((Person)agent).getNode());}
			else if (agent instanceof Household) {
				// pass x,y coordinates of household on to node
				NetworkNode node = (NetworkNode)((Household)agent).getNode();
				node.setX(((Household)agent).getX()*25);
				node.setY(((Household)agent).getY()*20);
				nodes.add(node);
			}
		}
		return nodes;
	}

	/** recommended by Repast: creates the schedule */
	private void buildSchedule(){

		// define inner class for model actions to be performed at every tick
		class SAModelStep extends BasicAction{
			public void execute(){
				long start = System.currentTimeMillis();
				// set time
				setTick((int)getTickCount() - 1);
				setDateValues();
				long end = System.currentTimeMillis();		
				if (!showOutput){
					System.out.println("\n*** TICK " + tick + " ***\n");  //###

				}
				dump("\n*** TICK " + tick + " ***\n");
				dump("Time to update tick: " + (end - start)/1000.0 + " seconds");
				start = System.currentTimeMillis();
				// run update procedures
				update();
				end = System.currentTimeMillis();
				dump("Time to run Java: " + (end - start)/1000.0 + " seconds");
				
				// run Jess
				try {
					//put modules on stack: they become current in reverse order
					engine.setFocus("CLEANUP");
					engine.setFocus("FINAL");
					engine.setFocus("CONTENT");
					engine.setFocus("INIT");
					start = System.currentTimeMillis();
					engine.run();
					end = System.currentTimeMillis();
					dump("Time to run Jess: " + (end - start)/1000.0 + " seconds");
				}
				catch(JessException je){
					System.err.println("An error occurred at line "
							+ je.getLineNumber());
					System.err.println("caused by " + je.getCause());
					System.err.println("Message: " + je.getDetail());
					je.printStackTrace();
				}
				
				// retract unneeded facts 
				start = System.currentTimeMillis();
				disposalPolicy.executePolicy(tick);
				end = System.currentTimeMillis();
				dump("Time to retract facts: " + (end - start)/1000.0 + " seconds");
					
				if (showGUI){
					// update the displays
					friendNetworkDisplay.updateDisplay();
					kinshipNetworkDisplay.updateDisplay();
					stokvelNetworkDisplay.updateDisplay();
				}
				
				// do some data collection
				weeklyDataRecorder.record();
				sexualNetworkRecorder.record();
				sexualNetworkRecorder.writeToPajek();
				if (week == 3){
					monthlyDataRecorder.record();
					resetOutputVariables();
				}
				// every 10th tick record friendship network
				if (tick % 10 == 0){
					friendNetworkRecorder.record();
					friendNetworkRecorder.writeToPajek();
				}
				// every year (48th tick) record age distribution
				if (tick % 48 == 0){
					ageDistRecorder.recordAndWrite();
				}
			}
		}
		
		// build the schedule
		schedule.scheduleActionBeginning(0, new SAModelStep());
		schedule.scheduleActionAt(miningTick, this, "createMine");
		schedule.scheduleActionAtInterval(4, villageEconomyRecorder, "write", Schedule.LAST);
		schedule.scheduleActionAtInterval(1, hhEconomyRecorder, "write", Schedule.LAST);
		schedule.scheduleActionAtInterval(4, weeklyDataRecorder, "writeToFile", Schedule.LAST);
		schedule.scheduleActionAtInterval(16, monthlyDataRecorder, "writeToFile", Schedule.LAST);
		schedule.scheduleActionAtInterval(24, hhWealthDistRecorder, "record", Schedule.LAST);
		schedule.scheduleActionAtInterval(48, hhWealthDistRecorder, "write", Schedule.LAST);
		schedule.scheduleActionAtInterval(4, partnershipRecorder, "write", Schedule.LAST);
		schedule.scheduleActionAtInterval(100, deathRecorder, "write", Schedule.LAST);
		schedule.scheduleActionAtEnd(monthlyDataRecorder, "writeToFile");
		schedule.scheduleActionAtEnd(weeklyDataRecorder, "writeToFile");
		schedule.scheduleActionAtEnd(sexualNetworkRecorder, "writeToPajek");
		schedule.scheduleActionAtEnd(deathRecorder, "write");
		schedule.scheduleActionAtEnd(hhEconomyRecorder, "write");
		schedule.scheduleActionAtEnd(villageEconomyRecorder, "write");
		schedule.scheduleActionAt(stopTime, this, "stop");
	}

	/** Required by Repast: returns the schedule */
	public Schedule getSchedule() {
		return schedule;
	}

	/** Required by Repast: returns the model name */
	public String getName() {
		return "South Africa CAVES Model";
	}
	

	// ---- Functionality outsourced from Jess ------------------------------
	
	/** Jess related: define disposal policies for unneeded facts */
	private void setDisposalPolicy(){
		disposalPolicy = new EndorsementDisposalPolicy(engine);
		disposalPolicy.addPolicy("MAIN::help-needed", 3);
		disposalPolicy.addPolicy("MAIN::asked-for-help", 3);
		disposalPolicy.addPolicy("MAIN::payment", "payment-date", 1);
		disposalPolicy.addPolicy("MAIN::employment", "end-date", 1);
		disposalPolicy.addPolicy("MAIN::job-application", 1);
		disposalPolicy.addPolicy("MAIN::employees-demanded", 1);
		disposalPolicy.addPolicy("MAIN::object-endorsement-value", 0);
		disposalPolicy.addPolicy("MAIN::similarity-index", 0);
		disposalPolicy.addPolicy("MAIN::friends-list", 0);
		disposalPolicy.addPolicy("MAIN::group-wanted", 1);
		disposalPolicy.addPolicy("MAIN::group-proposal", 1);
		disposalPolicy.addPolicy("MAIN::death", 0);
		disposalPolicy.addPolicy("MAIN::death-in-hh", 1);
		disposalPolicy.addPolicy("MAIN::membership-reminder", 4);
		disposalPolicy.addPolicy("MAIN::marriage", 0);
		disposalPolicy.addPolicy("MAIN::rent-a-shack", "end-date", 0);
		disposalPolicy.addPolicy("MAIN::leave", 0);
		disposalPolicy.addEndorsementPolicy("similar", 1);
		disposalPolicy.addEndorsementPolicy("most-similar", 1);
		disposalPolicy.addEndorsementPolicy("same-work", 4);
		disposalPolicy.addEndorsementPolicy("is-friend", 1);
		disposalPolicy.addEndorsementPolicy("has-job", 6);
	}
	
	/** equivalent to (watch facts) in Jess */
	public void watchFacts(){
		try {
			engine.watch(WatchConstants.FACTS);
		} catch (JessException je) {
			System.err.println("An error occurred at line " + je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}
	}
	
	/** equivalent to (unwatch facts) in Jess */
	public void unwatchFacts(){
		try {
			engine.unwatch(WatchConstants.FACTS);
		} catch (JessException je) {
			System.err.println("An error occurred at line " + je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}
	}
	
	

	// ---- Model initialisation related: creation of agents (persons), households etc. -----------------
	
	/** Initialisation related: choose some agents to be HIV positive */
	private void pickHIVPositives(){
		// assign some people to be HIV positive to get things rolling
		// -- only adults between 15-64
		// -- use AIDS prevalence distribution from http://www.avert.org/safricastats.htm
		int[] ageClasses = new int[]{19,24,29,34,39,44,49,54,59,64};
		double[] malePrevalence = new double[]{3.2,6.0,12.1,23.3,23.3,17.5,10.3,14.2,6.4,4.0};
		double[] femalePrevalence = new double[]{9.4,23.9,33.3,26.0,19.3,12.4,8.7,7.5,3.0,3.7};

		// sort adults into age classes
		ArrayList<ArrayList<Person>> males = new ArrayList<ArrayList<Person>>();
		ArrayList<ArrayList<Person>> females = new ArrayList<ArrayList<Person>>();
		for (int i = 0; i < ageClasses.length; i++){
			males.add(new ArrayList<Person>());
			females.add(new ArrayList<Person>());
		}
		for (Person p : personList){
			if (p.age >= 15 && p.age <= 64){
				int index = ageClassIndexOf(p.age, ageClasses);
				if (p.gender.toString().equals("female")){
					females.get(index).add(p);
				}
				else {
					males.get(index).add(p);
				}
			}
		}
		int infected = 0;
		for (int i = 0; i < ageClasses.length; i++){
			for (Person p : males.get(i)){
				if (probabilityp(malePrevalence[i]*0.01)){
					p.setInfectedSince(0, stageDist.nextInt());
					infected++;
				}
			}
			for (Person p : females.get(i)){
				if (probabilityp(femalePrevalence[i]*0.01)){
					p.setInfectedSince(0, stageDist.nextInt());
					infected++;
				}
			}
		}
		System.out.print(" (" +  infected + " people are infected) ... ");
	}
	
	/** helper function for pickHIVPositives() */
	private int ageClassIndexOf(int age, int[] classes){
		int index = 0;
		while (index < classes.length && age > classes[index]){
			index++;
		}
		return index;
	}
	
	
	/** Initialisation related: assign initial education levels to agents according to RADAR data */	
	private void assignEducationLevels(){
		// use distributions extracted from RADAR data
		EmpiricalDiscreteDistArray femaleLevels = new EmpiricalDiscreteDistArray(
				new double[]{5,6,13,18,28,37,64,120}, 
				new double[][]{{1,0,0,0,0,0},{0.2,0.8,0,0,0,0},{0.02,0.93,0.02,0.04,0.00,0.00},
									{0.01,0.17,0.12,0.69,0.01,0.00},{0.01,0.06,0.04,0.70,0.12,0.07},
									{0.06,0.19,0.06,0.48,0.16,0.05},{0.45,0.37,0.05,0.10,0.02,0.01},
									{0.93,0.06,0.00,0.01,0.00,0.00}},
				new MersenneTwister(seed));
		EmpiricalDiscreteDistArray maleLevels = new EmpiricalDiscreteDistArray(
				new double[]{5,6,13,18,33,41,51,120}, 
				new double[][]{{1,0,0,0,0,0},{0.2,0.8,0,0,0,0},{0.02,0.98,0,0,0,0},{0,0.2,0.1,0.7,0,0},
									{0.01,0.10,0.05,0.58,0.18,0.08},{0.03,0.33,0.05,0.33,0.18,0.08},
									{0.20,0.42,0.10,0.20,0.05,0.03},{0.46,0.37,0.03,0.13,0.01,0.00},
									{0.83,0.10,0.01,0.06,0.00,0.00}},
				new MersenneTwister(seed));
		EmpiricalDiscreteDistArray femaleAttendance = new EmpiricalDiscreteDistArray(
				new double[]{5,6,13,16,18,20,24,28,120}, 
				new double[][]{{1.0, 0.0},{0.2, 0.8},{0.02, 0.98},{0.06, 0.94},{0.22, 0.78},{0.5, 0.5},{0.85, 0.15},{0.95, 0.05},{1.0, 0.0}},
				new MersenneTwister(seed));
		EmpiricalDiscreteDistArray maleAttendance = new EmpiricalDiscreteDistArray(
				new double[]{5,6,16,18,21,23,28,120}, 
				new double[][]{{1.0, 0.0},{0.2, 0.8},{0.02, 0.98},{0.17, 0.83},{0.5, 0.5},{0.75, 0.25},{0.95,0.05},{1.0, 0.0}},
				new MersenneTwister(seed));
		
		// have to make sure that
		// (a) persons with government jobs have high education level (5 or at least 4)
		// --> this is now taken care of in assignEmployments()!
		// (b) young migrants are still at school and have correct education level (3, 4, 5)
		// (c) all kids of school-age are at school
		// (d) adults > 19 who are still at school have to be migrant
		for (Person p : personList){
			int level = 0;
			boolean attendance = false;
			if (p.getGender().toString().equalsIgnoreCase("female")){
				level = femaleLevels.nextInt(p.getAge());
				attendance = (femaleAttendance.nextInt(p.getAge()) > 0);
			}
			else {
				level = maleLevels.nextInt(p.getAge());
				attendance = (maleAttendance.nextInt(p.getAge()) > 0);
			}
			p.setAchievedEducation(level);
			p.setAtSchool(attendance);
			if (p.getAge() > 19 && p.isAtSchool()){
				p.setMigrant(true);
			}
			if (p.isMigrant() && p.getAge() <= 23){ // see pickMigrants()
				p.setAtSchool(true);
				// adjust level if necessary
				adjustEducationLevel(p);
			}
		}
	}
	
	/** helper function for assignEducationLevels() */
	private void adjustEducationLevel(Person p){
		// p still goes to school so education level has to be matched to age group
		int level = 0;
		if (p.age >= 18) level = 5; // in tertiary education
		else if (p.age >= 14) level = 3; // in secondary education
		else level = 1; // in primary education
		if (!p.getAchievedEducation().toString().equals(Person.educationLevels[level])){
			p.setAchievedEducation(level);
		}
	}
		
	/** Initialisation related: create some burial societies */
	private void createBurialSocieties(){
		// start model with some burial societies already running
		// assumptions: 
		// -- burial societies do not operate across villages
		// -- household heads are most likely to be a member, spouses of hh heads may be too
		// -- size is between 20 and 40
		// -- monthly fee is between 20 and 60 Rand
		// -- members are from related households or friends from same church/neighbours/.. 
		// -- only hh with an expected income > 500 Rand will be members <-- not implemented yet ##
		
		// sort hh list into villages
		ArrayList<ArrayList<Household>> villages = new ArrayList<ArrayList<Household>>(numVillages);
		for (int i = 0; i < numVillages ; i++){
			villages.add(new ArrayList<Household>());
		}
		for (Household hh : householdList){
			String villageName = hh.getVillage().toString();
			int villageIndex = Integer.valueOf(hh.getVillage().toString().substring(8)).intValue();
			villages.get(villageIndex).add(hh);
		}
		// for each village create some burial societies
		for (int i = 0; i < numVillages; i++){
			ArrayList<Household> households = villages.get(i);
			int numBS = households.size() / Random.uniform.nextIntFromTo(20,40);	// may result in 0 ==> some villages might not have any bs from the start
			for (int j = 0; j < numBS; j++){
				int numMembers = Random.uniform.nextIntFromTo(20,40);
				ArrayList<Person> members = new ArrayList<Person>(numMembers);
				// pick random first household
				Household hh = households.get(Random.uniform.nextIntFromTo(0, households.size()-1));
				members.add(hh.getHead());
				// pick some of their relatives from the same village
				pickRelatives(findRelatedHouseholds(hh, households), (int)(numMembers*Random.uniform.nextIntFromTo(30,70)/100), members);
				// pick the rest from neighbours and same church
				pickFriends(members, numMembers, households);
				// create burial society
				BurialSociety bs = new BurialSociety(Random.uniform.nextIntFromTo(10000, 30000), members, this);
				bs.setMonthlyFee(Random.uniform.nextIntFromTo(20,60));
				System.out.println(bs.toString() + " created for village " + i);
			}
		}
	}
	
	/** helper function for createBurialSocieties() */
	private void pickRelatives(ArrayList<Household> related, int n, ArrayList<Person> members){
		// pick n from the related households of the same village
		ArrayList<Household> chosen = new ArrayList<Household>();
		int numChosen = Math.min(n, related.size());
		for (int i = 0; i < numChosen; i++){
			Household h = related.get(Random.uniform.nextIntFromTo(0, related.size()-1));
			chosen.add(h);
			related.remove(h);
			// add head of chosen household to member list
			members.add(h.getHead());
		}
		// if there are less than n members, add spouses of heads or any other adult household member
		while (members.size() < n && chosen.size() > 0){
			Household h = chosen.get(Random.uniform.nextIntFromTo(0, chosen.size()-1));
			chosen.remove(h);
			Person head = h.getHead();
			if (head.spouse != null){
				members.add(head.spouse);
			}
			else {
				Person p = h.pickRandomAdult();
				if (p != null && p != head){
					members.add(p);
				}
			}
		}
	}
	
	/** helper function for createBurialSocieties() */
	private ArrayList<Household> findRelatedHouseholds(Household hh, ArrayList<Household> hhList){
		ArrayList<Household> related = new ArrayList<Household>();
		for (Household h : hhList){
			if (hh.isKinOf(h)){
				related.add(h);
			}
		}
		return related;
	}
	
	/** helper function for createBurialSocieties() */
	private void pickFriends(ArrayList<Person> members, int maxNum, ArrayList<Household> village){
		// since we don't have any friends yet pick from same church and neighbours
		
		// find out which churches the members attend
		ArrayList<Value> churches = findChurches(members);
		// pick people of same church but only if their household doesn't belong the the burial society yet
		int i = 0;
		while (members.size() < maxNum && i < village.size()){
			Household hh = village.get(i++);
			// does hh belong to the same church?
			if (churches.contains(hh.getChurch())){
				// are any members of the burial society members of this household?
				boolean alreadyMember = false; 
				int j = 0;
				while (!alreadyMember && j < members.size()){
					alreadyMember |= hh.hasMember(members.get(j++));
				}
				if (!alreadyMember){
					// pick the household head as member of the burial society
					members.add(hh.getHead());
				}
			}
		}
		
		// if there aren't enough members yet take some neighbours
		// ### to be implemented ###
		if (members.size() < maxNum){
			System.out.println("burial society is " + (maxNum - members.size()) + " members short");
		}
	}
	
	/** helper function for pickFriends */
	private ArrayList<Value> findChurches(ArrayList<Person> list){
		ArrayList<Value> churches = new ArrayList<Value>();
		for (Person p : list){
			Value church = p.getHousehold().getChurch();
			if (church != null && !churches.contains(church)){
				churches.add(church);
			}
		}
		return churches;
	}
	
	
	/** Initialisation related: create villages, assing households to villages, position households in space. 
	 *  Includes creating churches for villages and affiliate households with them (to just go through 
	 *  the village and household list once).
	 *  
	 *  @param hhClusters kinship clusters of households
	 */
	private void locateHouseholds(ArrayList<ArrayList<NetworkNode>> hhClusters){
		// create villages 
		ArrayList<Village> villages = new ArrayList<Village>(numVillages);
		// split household list according to kinship clusters
		int[] hhIndex = sortHouseholdList(hhClusters);
		for (int i = 0; i < numVillages; i++){
			Village v = createVillage("village-" + i, (hhIndex[i+1]-hhIndex[i]), villages);
			// populate village
			populateVillage(v, hhIndex[i], hhIndex[i+1]);
			// create churches and affiliate households with them
			createChurches (i, hhIndex[i], hhIndex[i+1]);
		}
	}
	
	/** Initialisation related: create churches for a given village and assign the village's households
	 * randomly to the churches. Called from locateHouseholds().
	 * 
	 * @param village village index
	 * @param startIndex index into the household list (start of this village's households)
	 * @param endIndex index into the household list (end of this village's households)
	 */
	private void createChurches(int village, int startIndex, int endIndex){
		System.out.println("++++++++++++ create churches for village " + village);
		// choose number of denominations in the village (randomly between 1 and numDenominations)
		int numChurches = Random.uniform.nextIntFromTo(1, numDenominations);
		// create churches
		Church[] churches = new Church[numChurches];
		for (int i = 0; i < numChurches; i++){
			churches[i] = new Church("church-" + village + "-" + i, "denomination-" + i, this);			
		}
		// assign households to churches (randomly)
		for (int i = startIndex; i < endIndex; i++) {
			Household hh = householdList.get(i);
			if (Random.uniform.nextDouble() < churchParticipationRate){
				churches[Random.uniform.nextIntFromTo(0, numChurches-1)].addHousehold(hh);
			}
		}
	}
	
	/** Initialisation related: position households in space, or to be more precise, position
	 * them randomly inside the area the village takes up. Called from locateHouseholds()
	 * 
	 * @param village
	 * @param startIndex index into the household list (start of this village's households)
	 * @param endIndex index into the household list (end of this village's households)
	 */
	private void populateVillage(Village village, int startIndex, int endIndex){
		// distribute households on random positions inside village
		for (int i = startIndex; i < endIndex; i++) {
			Household hh = householdList.get(i);
			int[] extension = village.getExtension();
			int x, y;
			do {
				x = Random.uniform.nextIntFromTo(extension[0], extension[2]);
				y = Random.uniform.nextIntFromTo(extension[1], extension[3]);
			} while (space.getObjectAt(x, y) != null);
			hh.setXY(x, y);
			hh.setVillage(village.getName());
			space.putObjectAt(x, y, hh);
		}	
	}
	
	
	/** helper function for locateHouseholds(): sorts household list into clusters detected
	 *  in the kinship network
	 */
	private int[] sortHouseholdList(ArrayList<ArrayList<NetworkNode>> clusters){
		// sort household list into clusters and return the indices of where the clusters start
		ArrayList<Household> sortedList = new ArrayList<Household>(this.householdList.size());
		int[] index = new int[clusters.size()+1];
		int i = 0, j = 0;
		for (ArrayList<NetworkNode> cluster : clusters){
			// begin of a cluster
			index[j++] = i;
			for (NetworkNode node : cluster){
				Household hh = findHouseholdOf(node);
				this.householdList.remove(hh);
				sortedList.add(i++, hh);
			}
		}
		index[clusters.size()] = sortedList.size();
		this.householdList = sortedList;
		return index;
	}
	
	/** returns the household to a given network node
	 * 
	 * @param node the network node
	 * @return the associated household
	 */
	private Household findHouseholdOf(NetworkNode node){
		// search for node ID in household list
		int i = 0;
		while (i < householdList.size() && !householdList.get(i).getNode().equals(node)){
			i++;
		}
		if (i < householdList.size()) return householdList.get(i);
		return null;
	}
	
	/** Initialisation related: creates a village for the given number of households and
	 *  positions it in space. The number of households determine the size of the village area.
	 *  
	 * @param name the name of the village
	 * @param numHouseholds the number of households in the village
	 * @param villages the list of existing villages (needed to avoid overlapping villages in space)
	 * @return the new village
	 */
	private Village createVillage(String name, int numHouseholds, ArrayList<Village> villages){
		// determine size of village
		int radius = (int)(Math.ceil(Math.sqrt(numHouseholds)/2.0) * densityFactor);
		// determine location of village on the grid
		int[] villageCentre = determineVillageLocation(radius, villages);
		Village v = new Village(name, villageCentre, radius);
		villages.add(v);
		this.nameVillageMap.put(name, v);
		return v;
	}
	
	/** helper method for createVillage() */
	private int[] determineVillageLocation(int radius, ArrayList<Village> villages){
		int[] centre = new int[2];
		// find a free region on the grid to place the village
		do {
			centre[0] = Random.uniform.nextIntFromTo(radius+1, this.gridSizeX-radius-1);
			centre[1] = Random.uniform.nextIntFromTo(radius+1, this.gridSizeY-radius-1);
		} while (overlapping(centre, radius, villages));
		return centre;
	}
	
	/** helper method for createVillage() */
	private boolean overlapping(int[] centre, int radius, ArrayList<Village> villages){
		// check if new village (determined by centre and radius) will overlap with an existing village
		boolean tooClose = false;
		int[] x = new int[]{centre[0] - radius, centre[0] + radius};
		int[] y = new int[]{centre[1] - radius, centre[1] + radius};
		for (Village v : villages){
			for (int i = 0; i < x.length; i++){
				for (int j = 0; j < y.length; j++){
					tooClose = tooClose | v.coversWithBuffer(x[i], y[j]);
				}
			}
		}
		return tooClose;
	}
	
	/** Initialisation related: create households according to RADAR data */
	private void createHouseholds(){
		// (according to RADAR data, size distribution follows Normal dist with mean = 7, sd = 3)
		Normal hhSizeDist = new Normal(7.0, 3.0, new MersenneTwister(seed));  
		// (age distribution of hh heads follows Normal dist with mean = 56.2, sd = 12.2)
		Normal hhHeadAgeDist = new Normal(56.2, 12.2, new MersenneTwister(seed));
		// age difference between married spouses (man-woman), approx. by Normal dist with mean = 8.43, sd = 6.576
		Normal ageDiffDist = new Normal(8.43, 6.576, new MersenneTwister(seed));
		// female hh heads are 58% widowed, 22.4% single, 16.2% divorced/separated, 3.4% married
		EmpiricalDiscreteDist femaleStatusDist = new EmpiricalDiscreteDist(new double[]{0.224, 0.034, 0.162, 0.58}, new MersenneTwister(seed));
		// male hh heads are 0.8% single, 96.4% married, 0.8% divorced, 2% widowed
		EmpiricalDiscreteDist maleStatusDist = new EmpiricalDiscreteDist(new double[]{0.008, 0.964, 0.008, 0.02}, new MersenneTwister(seed));
		// type of hh member (except head / spouse): child 0.653, grandchild 0.238, other 0.109
		EmpiricalDiscreteDist memberTypeDist = new EmpiricalDiscreteDist(new double[]{0.653, 0.238, 0.109}, new MersenneTwister(seed));
		// general age distribution, can be approximated with Gamma distribution, with alpha = 2.4 and lambda = 0.086)
		Gamma ageDist = new Gamma(2.4, 0.086, new MersenneTwister(seed));
	 
		int size;
		for (int i = 0; i < numHouseholds; i++){
			Household h = new Household(this);
			h.setId("household-" + i);
			nameHouseholdMap.put(h.getId().toString(), h);
			// determine size of hh 
			do {
				size = (int)hhSizeDist.nextDouble();
			} while (size < 1);
			// create members according to empirical data from RADAR
			populateHousehold(size, h, hhHeadAgeDist, ageDiffDist, ageDist, femaleStatusDist, maleStatusDist, memberTypeDist);
			// give some random cash
			h.setCash(Random.uniform.nextDoubleFromTo(500.0, 900.0));
			this.householdList.add(h);
			try {
				engine.definstance("household", h, true);
			} catch (JessException je) {
				System.err.println("An error occurred at line " + je.getLineNumber());
				System.err.println("caused by " + je.getCause());
				System.err.println("Message: " + je.getDetail());
				je.printStackTrace();
			}
		}
		// set person index (necessary for later, when new persons are created and dead persons are removed from personList)
		this.personIndex = personList.size();
	}
	
	/* Initialisation related: choose some agents to be migrants
	 *  according to RADAR data
	 */
	private void pickMigrants() {
		// 13.8% of hh members are migrants, 5.05% female, 8.75% male
		// most of them are single (78% of females, 70% of males)
		// a small proportion (22.8% of females, 13.8% of males) go to school,
		// i.e. cost school fees instead of sending remittances
		// rough age distribution: 17% <= 23 (= going to school), 61% > 23 and <=
		// 40, 22% > 40
		int numFemales = (int) (this.personList.size() * 0.0505);
		int numMales = (int) (this.personList.size() * 0.0875);
		int young = (int) ((numFemales + numMales) * 0.17);
		int old = (int) ((numFemales + numMales) * 0.22);
		int middle = numFemales + numMales - young - old;
		System.out.println("there are " + numFemales + " female and " + numMales
				+ " male migrants");
		System.out.println("young: " + young + ", middle: " + middle + ", old: "
				+ old);
		dump("Migrants:");
		// sort personList into the 3 age categories to pick randomly from
		ArrayList<Person> tempList = new ArrayList<Person>(personList.size());
		int[] index = sortPersonList(23, 40, tempList);
		Person migrant;
		int numFCreated = 0, numMCreated = 0, youngCreated = 0, middleCreated = 0, oldCreated = 0;
		ArrayList<Person> migrants = new ArrayList<Person>(); // to make sure that
																				// no-one is picked
																				// more than once
		while (numFCreated < numFemales | numMCreated < numMales) {
			while (youngCreated < young) {
				migrant = tempList.get(Random.uniform
						.nextIntFromTo(0, index[0] - 1));
				// make sure not to pick only member of a household!
				while ((migrant.isHead() && migrant.household.getMembers().size() == 1)
						|| migrants.contains(migrant)) {
					migrant = tempList.get(Random.uniform.nextIntFromTo(0,
							index[0] - 1));
				}
				migrants.add(migrant);
				if (migrant.getGender().toString().equals("female")
						&& numFCreated < numFemales) {
					migrant.setMigrant(true);
					numFCreated++;
					youngCreated++;
					dump(" * " + migrant);
					// System.out.println("female migrant aged " + migrant.getAge() +
					// ", " + migrant.getMarital_status().toString());
				}
				else if (migrant.getGender().toString().equals("male")
						&& numMCreated < numMales) {
					migrant.setMigrant(true);
					numMCreated++;
					youngCreated++;
					dump(" * " + migrant);
					// System.out.println("male migrant aged " + migrant.getAge() +
					// ", " + migrant.getMarital_status().toString());
				}
			}
			while (middleCreated < middle) {
				migrant = tempList.get(Random.uniform.nextIntFromTo(index[0],
						index[1] - 1));
				// make sure not to pick only member of a household!
				while ((migrant.isHead() && migrant.household.getMembers().size() == 1)
						|| migrants.contains(migrant)) {
					migrant = tempList.get(Random.uniform.nextIntFromTo(0,
							index[0] - 1));
				}
				migrants.add(migrant);
				if (migrant.getGender().toString().equals("female")
						&& numFCreated < numFemales) {
					migrant.setMigrant(true);
					numFCreated++;
					middleCreated++;
					dump(" * " + migrant);
					// System.out.println("female migrant aged " + migrant.getAge() +
					// ", " + migrant.getMarital_status().toString());
				}
				else if (migrant.getGender().toString().equals("male")
						&& numMCreated < numMales) {
					migrant.setMigrant(true);
					numMCreated++;
					middleCreated++;
					dump(" * " + migrant);
					// System.out.println("male migrant aged " + migrant.getAge() +
					// ", " + migrant.getMarital_status().toString());
				}
			}
			while (oldCreated < old) {
				migrant = tempList.get(Random.uniform.nextIntFromTo(index[1],
						tempList.size() - 1));
				// make sure not to pick only member of a household!
				while ((migrant.isHead() && migrant.household.getMembers().size() == 1)
						|| migrants.contains(migrant)) {
					migrant = tempList.get(Random.uniform.nextIntFromTo(0,
							index[0] - 1));
				}
				migrants.add(migrant);
				if (migrant.getGender().toString().equals("female")
						&& numFCreated < numFemales) {
					migrant.setMigrant(true);
					numFCreated++;
					oldCreated++;
					dump(" * " + migrant);
					// System.out.println("female migrant aged " + migrant.getAge() +
					// ", " + migrant.getMarital_status().toString());
				}
				else if (migrant.getGender().toString().equals("male")
						&& numMCreated < numMales) {
					migrant.setMigrant(true);
					numMCreated++;
					oldCreated++;
					dump(" * " + migrant);
					// System.out.println("male migrant aged " + migrant.getAge() +
					// ", " + migrant.getMarital_status().toString());
				}
			}
		}
	}
	
	/** helper method for pickMigrants() */
	private int[] sortPersonList(int lowAge, int highAge, ArrayList<Person> newList){
		int[] index = new int[2];
		index[0] = 0; index[1] = 0;
		for (Person p : personList){
			if (p.getAge() <= lowAge){
				// add to front of list = put in first part of list
				// ignore young children <= 5
				if (p.getAge() > 5) {
					newList.add(0, p);
					index[0]++; index[1]++;
				}
			}
			else if (p.getAge() > highAge){
				// add to back of list = put in last part of list
				newList.add(p);
			}
			else {
				// put in middle part
				newList.add(index[0], p);
				index[1]++;
			}
		}
		return index;
	}
	
	/** Initialisation related: create household members according to empirical data (RADAR)
	 * 
	 * @param numMembers	number of members
	 * @param hh	the household
	 * @param hhHeadAgeDist	distribution for the age of the household head
	 * @param ageDiffDist	distribution for the age difference between head and spouse
	 * @param ageDist	distribution for the age of household members in general
	 * @param femaleStatusDist	distribution for the marital status of females
	 * @param maleStatusDist	distribution for the marital status of males
	 * @param memberTypeDist	distribution for the type of household members (child / grandchild of head or other)
	 */
	private void populateHousehold(int numMembers, Household hh, Normal hhHeadAgeDist, Normal ageDiffDist, Gamma ageDist,
			EmpiricalDiscreteDist femaleStatusDist, EmpiricalDiscreteDist maleStatusDist, EmpiricalDiscreteDist memberTypeDist){
		int n = personList.size(); // last person ID
		int numCreated = 0;

		// start with creating the household head
		Person head = createHouseholdMember("person-" + n, hh);
		head.setAge((int)hhHeadAgeDist.nextDouble());
		int maleness = Random.uniform.nextIntFromTo(0,1);
		if (maleness == 0) {
			head.setGender("female");
			setAsHead(head, femaleStatusDist);
		}
		else {
			head.setGender("male");
			setAsHead(head, maleStatusDist);
		}
		if (askPensionOracle(head)) head.setRegisteredForPension(true); 
		numCreated++; n++;
		
		
		//if head married, give head a spouse
		
		if (head.maritalStatus.toString() == "married"){
			Person spouse = createHouseholdMember("person-" + n, hh);
			++n; ++numCreated;
			int ageDiff = (int)ageDiffDist.nextDouble();
			String genderString = head.gender.toString();
			if (genderString == "female"){
				spouse.setGender("male");
				spouse.setAge(head.getAge() + ageDiff);
			}
			else {
				spouse.setGender("female");
				spouse.setAge(head.getAge() - ageDiff);
			}
			head.setSpouse(spouse);
			spouse.setSpouse(head);
			spouse.setMaritalStatus("married");
			if (askPensionOracle(spouse)) spouse.setRegisteredForPension(true); 
		}
		
		// create rest of members
		// check for very rare case that hhsize is 1 and head is married --> change hhsize to 2
		if (numCreated > numMembers) numMembers = numCreated;

		Person[] members = new Person[numMembers-numCreated];
		int[] memberType = new int[members.length];
		for (int i = 0; i < memberType.length; i++){
			memberType[i] = memberTypeDist.nextInt();
//			System.out.print(" " + memberType[i] /*+ "/" + memberAge[i]*/);
		}
//		System.out.println();
		memberType = sortTypes(memberType);
		int eldestChildAge = 0;
		for (int i = 0; i < members.length; i++){
			members[i] = createHouseholdMember("person-" + (n + i),hh);
			if (Random.uniform.nextIntFromTo(0,1) == 0){
				members[i].setGender("female");
			}
			else {
				members[i].setGender("male");
			}
			int age = (int)ageDist.nextDouble();			
			int compAge = head.getAge();
			if ((head.getGender().toString() == "male") && head.getMaritalStatus().toString() == "married") compAge = head.spouse.getAge();
			if (memberType[i] == CHILD){
				// determine age according to hh head (resp. spouse): females can give birth between 14 and 45
				// --> child must be at least 14 years younger, at most 45 years younger
				while (age > (compAge-14) || age < (compAge-45)){
					age = (int)ageDist.nextDouble();
				}
				members[i].setAge(age);
				if (age > eldestChildAge) eldestChildAge = age;
//				System.out.println("child aged " + age);
				}
			else if (memberType[i] == GRANDCHILD) {
				if (eldestChildAge < 15){
					// change member type to CHILD
					memberType[i] = CHILD;
//					System.out.println("-- changed type from grandchild to child");
					while (age > (compAge-14) || age < (compAge-45)){
						age = (int)ageDist.nextDouble();
					}					
				}
				else {
					// determine age according to eldest child: females can give birth between 14 and 45
					// --> grandchild must be at least 14 years younger, at most 45 years younger
					while (age > (eldestChildAge-14) || age < Math.max((eldestChildAge-45),0)){
						age = (int)ageDist.nextDouble();
					}
				}
				members[i].setAge(age);
//				System.out.println("grandchild aged " + age);
			}
			else if (memberType[i] == OTHER){
				members[i].setAge(age);
//				System.out.println("other aged " + age);
			}
		}
		for (Person pers : members) {
			
			if (askPensionOracle(pers)) pers.setRegisteredForPension(true); 
			if (askChildGrantOracle(pers)) pers.setRegisteredForChildGrant(true);
		}
		createFamilyLinks(hh, members, memberType);
	}
	
	/** Creates a new member for the given household. Called during model initialisation and whenever a baby is born
	 *  uring simulation.
	 *  @param name the members's name
	 *  @param hhold the household
	 *  @return the new person
	 */
	private Person createHouseholdMember(String name, Household hhold){
		Person pers = new Person(name, hhold, this);
		pers.setNode(new NetworkNode(pers.getName()));
		pers.setBirthMonth(Random.uniform.nextIntFromTo(0,11));
		pers.setBirthWeek(Random.uniform.nextIntFromTo(0,3));
		int maleness = Random.uniform.nextIntFromTo(0,1);
		if (maleness == 0) pers.setGender("female");
		else pers.setGender("male");
		pers.setTagEvolutionPropensity (Random.uniform.nextDoubleFromTo(minTagEvolutionPropensity, maxTagEvolutionPropensity));
		pers.createTag();
		personList.add(pers);
		namePersonMap.put(name, pers);
		try {
			engine.definstance("person",pers, true);
		}
		catch(JessException je) {
			System.err.println("An error occurred at line " + je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}
		hhold.addMember(pers);

		return pers;
	}
	
	/** helper function for populateHouseholds() */
	private void setAsHead(Person person, EmpiricalDiscreteDist statusDistribution){
		person.setMaritalStatus(statusDistribution.nextInt());
		person.setHead(true);
	}
	
	/** helper function for populateHouseholds() */
	private void createFamilyLinks(Household hh, Person[] members, int[] memberType){
		// members contains the household members apart from head and spouse (if any)
		// memberType[i] describes the type of member[i]: child, grandchild or other
		
		// new version with network links replaced by direct references
		
		// create family links	
		Person head = hh.getHead();
		Person spouse = null;
		dump(hh.getId() + " has the following members:");
		dump(" * " + head.getName() + " (" + head.getGender() + ", " + head.getAge() + ") -- head");
		// create links with / for rest of household members
		for (int i=0; i < members.length; i++){
			if (memberType[i] == CHILD){
				dump(" * " + members[i].getName() + " (" + members[i].getGender() + ", " + members[i].getAge() + ") -- child of head");
				// create parent/child links
				head.addChild(members[i]);
				if (head.getGender().toString().equals("male")){
					members[i].father = head.getName();
				}
				else {
					members[i].mother = head.getName();
				}
				if (spouse != null){
					spouse.addChild(members[i]);
					if (spouse.getGender().toString().equals("male")){
						members[i].father = spouse.getName();
					}
					else {
						members[i].mother = spouse.getName();
					}
				}
				// create sibling links with all the other children? 
				for (int j=0; j < members.length; j++){
					if (j != i && memberType[j] == CHILD){
						members[i].addSibling(members[j]);
					}
				}
			}
			if (memberType[i] == GRANDCHILD){
				// find a suitable child to be the parent
				int parent = findParent(members, memberType, i);
				if (parent < 0){
//					System.out.println("NO SUITABLE PARENT FOR GRANDCHILD FOUND");
					dump(" * " + members[i].getName() + " (" + members[i].getGender() + ", " + members[i].getAge() + ") -- grandchild of head");
					// now what? direct links to grandparents are not possible anymore
					// a solution would be is-kin endorsement
					// alternative: ignore as it's only a tiny fraction at model initialisation
					//###
				}
				else {
					dump(" * " + members[i].getName() + " (" + members[i].getGender() + ", " + members[i].getAge() + ") -- child of " + members[parent].getName());
					members[parent].addChild(members[i]);
					if (members[parent].getGender().toString().equals("male")){
						members[i].father = members[parent].getName();
					}
					else {
						members[i].mother = members[parent].getName();
					}
				}
			}
			if (memberType[i] == OTHER){
				// create family:other links with head and spouse
				// better: with everyone? 
				dump(" * " + members[i].getName() + " (" + members[i].getGender() + ", " + members[i].getAge() + ") -- other");
				// don't have direct reference to others anymore
				// solution: use is-kin endorsements
				// alternative: just do nothing as it's only a small fraction at model initialisation anyway
				//###
			}
		}
		
		// create sibling links between grandchildren (if of same mother/father)
		for (Person p : members){
			if (p.getChildren().length > 1){
				ArrayList<Person> siblings = retrieveObjects(p.getChildren());
				for (int i = 0; i < siblings.size()-1; i++){
					Person child = siblings.get(i);
					for (int j = i+1; j < siblings.size(); j++){
						Person sibling = siblings.get(j);
						child.addSibling(sibling);
						sibling.addSibling(child);
					}
				}
			}
		}
		
	}
	
	/** Initialisation related: choose some agents to be government employees */
	private void assignEmployments(){
		
		// make sure government employees have education levels >= secondary
		ArrayList<Person> employed = new ArrayList<Person>();
		ArrayList<Person> educated = new ArrayList<Person>();
		// determine potential employees and persons with higher education
		int numAdults = 0;
		for (Person p : personList){
			if (p.age > 16 && p.age <= 60){
				numAdults++;
				if (!p.isAtSchool() && Random.uniform.nextDouble() < adultEmploymentRate){
					employed.add(p);
					String educationLevel = p.getAchievedEducation().toString();
					if (educationLevel.equals("secondary") || educationLevel.equals("tertiary")){
						educated.add(p);
					}
				}
			}
		}
		// educated is a subset of employed which is a subset of adults
		// adjust governmentEmploymentRate (which is for all adults) to the subset of educated employees
		double adjustedGovEmpRate = (numAdults * governmentEmploymentRate) / (double)educated.size();
		// pick government employees
		for (Person p : educated){
			if (Random.uniform.nextDouble() < adjustedGovEmpRate){
				p.assertEmployment("government", Random.uniform.nextIntFromTo(4,8) * standardWage, (60-p.age)*12);
			}
		}
	}
	
	/** helper function for populateHousehold() */
	private int[] sortTypes(int[] types){
		int[] sorted = new int[types.length];
		int index = 0;
		for (int i = 0; i < types.length; i++){
			// copy all the CHILD types
			if (types[i] == CHILD) {
				sorted[index] = types[i];
				index++;
			}
		}
		for (int i = 0; i < types.length; i++){
			// copy all the GRANDCHILD types
			if (types[i] == GRANDCHILD){
				sorted[index] = types[i];
				index++;
			}
		}
		for (int i = 0; i < types.length; i++){
			// copy all OTHER types
			if (types[i] == OTHER){
				sorted[index] = types[i];
				index++;
			}
		}
		return sorted;
	}
	
	/** helper function for createFamilyLinks() */
	private int findParent(Person[] members, int[] type, int childIndex){
		// try to find a daughter (of the head) with suitable age
		ArrayList<Integer> suitableDaughters = new ArrayList<Integer>();
		ArrayList<Integer> suitableSons = new ArrayList<Integer>();
		int childAge = members[childIndex].getAge();
		for (int index = 0; index < members.length; index++){
			if (type[index] == CHILD && members[index].getAge() >= (childAge + 14)){
				if (members[index].gender.toString() == "female") suitableDaughters.add(new Integer(index));
				else suitableSons.add(new Integer(index));
			}
		}
		// if there are suitable daughters, choose randomly amongst them
		if (suitableDaughters.size() > 0){
			return suitableDaughters.get(Random.uniform.nextIntFromTo(0, suitableDaughters.size()-1)).intValue();
		}
		else if (suitableSons.size() > 0){
			return suitableSons.get(Random.uniform.nextIntFromTo(0, suitableSons.size()-1)).intValue();			
		}
		else {
			// bugger!
			return -1;
		}
	}
	
	/** Initialisaiton related: create the neighbour network according to the households' position in space
	 */
	private void createNeighbourNetwork(){
		// use spatial neighbourhood 
		// create neighbour endorsements between members of households
		// within the model's neighbourhoodRadius
		
		for (Household hh : householdList){
			for (Household hh2 : householdList){
				if (hh != hh2){
					double distance =
						Math.sqrt((hh.x - hh2.x) * (hh.x - hh2.x)
								+ (hh.y - hh2.y) * (hh.y - hh2.y));
					if (distance < neighbourhoodRadius){
						for (Person p : hh.members){
							for (Person p2 : hh2.members)
								p.assertNeighbourEndorsement(p2);
						}
							
					}
				}
			}
		}
	}
	
	/** Initialisation related: create the kinship network between households
	 *  as a small-world network
	 * @return the kinship network as a list of clusters
	 */
	private ArrayList<ArrayList<NetworkNode>> createKinshipNetwork(){
		ArrayList<NetworkLink> linkList = new ArrayList<NetworkLink>();
		
		// small world kinship network
		List list = NetworkFactory.createWattsStrogatzNetwork(householdList.size(), 2, 0.1, NetworkNode.class, NetworkLink.class);
		// assign households to nodes and put them into Jess
		for (int i = 0; i < householdList.size(); i++){
			NetworkNode node = (NetworkNode)list.get(i);
			Household household = householdList.get(i);
			node.setId(household.getId());
			household.setNode(node);
			linkList.addAll(node.getOutEdges());
		}
		// put the links into Jess as endorsements
		for (NetworkLink link : linkList){
			link.setType("kinship");
			// call kin endorsement
			endorseKin(findHouseholdOf(link.getFrom()), findHouseholdOf(link.getTo()));
		}
		// determine clusters to populate villages with
//		System.out.println("\n *** determining household clusters");
		return determineClusters(list);		
	}
	
	/** Assert kinship endorsements in Jess between the members of the given households.
	 *  Called during model initialisation and whenever a couple builds a house and moves out
	 *  into their own household during simulation.
	 * 
	 * @param hh1 first household
	 * @param hh2 second household, related to first
	 */
	public void endorseKin(Household hh1, Household hh2){
		dump(hh1.getId() + "\tis kin to\t" + hh2.getId());
		// to replace rule endorse-kin
		// every member of hh1 endorses every member of hh2 with is-kin endorsement
		try {
			Value isKin = new Value("is-kin", RU.SYMBOL);
			for (Value fromMember : hh1.getMemberIDs()){
				for (Value toMember : hh2.getMemberIDs()){
						Fact kinFact = new Fact("endorsed", engine);
						kinFact.setSlotValue("endorsed-object", toMember);
						kinFact.setSlotValue("clauseOwner", fromMember);
						kinFact.setSlotValue("endorsement", isKin);
						kinFact.setSlotValue("tickStamp", new Value(tick, RU.INTEGER));
						engine.assertFact(kinFact);
//						System.out.println("asserted new is-kin endorsement from " + fromMember + " to " + toMember);
				}
			}
		}
		catch(JessException je) {
			System.err.println("An error occurred at line " + je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}	
	}
	
	/** Assert neighbour endorsements in Jess for the given person (a newborn baby).
	 * 
	 * @param p a new person
	 */
	void endorseNeighbours(Person p){
		// new baby needs to endorse every member of neighbour households with is-neighbour endorsements
		// and vice versa
		endorseAllNeighbours(p, p.getHousehold());
	}
	
	/** Assert neighbour endorsements in Jess for the given person (a miner renting a shack
	 *  from the given household)
	 *  @param p a miner
	 *  @param hh his landlord household
	 */
	public void endorseNeighbours(Person p, Household hh){
		// miner endorses (a) every member of the landlord household and 
		// (b) every member of their neighbour households with is-neighbour endorsements
		for (Person p2 : hh.members){
			p.assertNeighbourEndorsement(p2);
			p2.assertNeighbourEndorsement(p);
		}
		endorseAllNeighbours(p, hh);
	}
	
	/** helper function: assert neighbour endorsements for the given person
	 * and every member of the given household's neighbour households.
	 * @param p the person
	 * @param hh the household
	 */
	private void endorseAllNeighbours(Person p, Household hh){
		Value village = hh.getVillage();
		for (Household hh2 : householdList){
			if (hh2.getVillage().equals(village) && hh != hh2){
				double distance = Math.sqrt((hh.x - hh2.x) * (hh.x - hh2.x)
														+ (hh.y - hh2.y) * (hh.y - hh2.y));
				if (distance < neighbourhoodRadius){
					for (Person p2 : hh2.members) {
						p.assertNeighbourEndorsement(p2);
						p2.assertNeighbourEndorsement(p);
					}
				}
			}
		}		
	}
	
	/** Assert kinship endorsements in Jess for the given person (a newborn baby).
	 * 
	 * @param baby a new person
	 */
	void endorseKin(Person baby){
		// new baby needs to endorse every member of kinship households with is-kin endorsements
		// and vice versa
		ArrayList links = baby.getHousehold().getNode().getOutEdges();
		for (Object o : links){
			NetworkLink link = ((NetworkLink)o);
			if (link.getType().equals("kinship")){
				Household hh = findHouseholdOf(link.getTo());
				try {
					Value isKin = new Value("is-kin", RU.SYMBOL);
					for (Value toMember : hh.getMemberIDs()){
						Fact kinFact = new Fact("endorsed", engine);
						kinFact.setSlotValue("endorsed-object", toMember);
						kinFact.setSlotValue("clauseOwner", baby.getName());
						kinFact.setSlotValue("endorsement", isKin);
						kinFact.setSlotValue("tickStamp", new Value(tick, RU.INTEGER));
						engine.assertFact(kinFact);				
						kinFact = new Fact("endorsed", engine);
						kinFact.setSlotValue("endorsed-object", baby.getName());
						kinFact.setSlotValue("clauseOwner", toMember);
						kinFact.setSlotValue("endorsement", isKin);
						kinFact.setSlotValue("tickStamp", new Value(tick, RU.INTEGER));
						engine.assertFact(kinFact);				
					}
				}
				catch(JessException je) {
					System.err.println("An error occurred at line " + je.getLineNumber());
					System.err.println("caused by " + je.getCause());
					System.err.println("Message: " + je.getDetail());
					je.printStackTrace();
				}	
			}
		}
		
	}
	
	/** Perform cluster analysis on kinship network: sort nodes (= households) into as many clusters
	 * as there are villages. Called from createKinshipNetwork().
	 * 
	 * @param list the kinship network as a list of nodes
	 * @return list of clusters
	 */
	private ArrayList<ArrayList<NetworkNode>> determineClusters(List list){
		// put network info into matrix with 0 = no link and 1 = link between nodes
		int[][] matrix = new int[list.size()][list.size()];
		for (int i = 0; i < matrix.length; i++){
			for (int j = 0; j < matrix[i].length; j++){
				matrix[i][j] = 0;
			}
		}
		for (Object o : list){
			NetworkNode node = (NetworkNode)o;
			ArrayList<NetworkNode> linkedNodes = getLinkedNodes(node);
			int from = getNodeIndex(node);
			for (NetworkNode n : linkedNodes){
				int to = getNodeIndex(n);
				matrix[from][to] = 1;
			}
		}
		
		// compute distances (1 - p_ij) between all the nodes using the ... coefficient for p_ij
		// with additional weight on existing link between i and j 
		double[][] distances = new double[list.size()][list.size()];
		for (int i = 0; i < matrix.length; i++){
			for (int j = 0; j < matrix.length; j++){
				if (i == j) distances[i][j] = 0.0;
				else {
					int a = 0, b = 0, c = 0, iLink = 0, jLink = 0;
					for (int k = 0; k < matrix[i].length; k++){
						if (matrix[i][k] == 1 && matrix[j][k] == 1) {
							a++; 
						}
						else if (matrix[i][k] == 1 && matrix[j][k] == 0) {
							b++; 
							if (k == j) jLink = 1;
						}
						else if (matrix[i][k] == 0 && matrix[j][k] == 1) {
							c++;
							if (k == i) iLink = 1;
						}
					}
					if ((a + b + c) == 0) distances[i][j] = 1.0;
					else distances[i][j] = 1 - ((a + iLink + jLink) / (double)(a + b + c));
				}
			}
		}

		// hierachical algorithm: start with each node in its own cluster
		// than merge those clusters with smallest distance until condition is met 
		// -- distance between clusters can be computed in several ways, we'll use so called
		// average linkage: mean distance between elements of each cluster
		// -- condition being either that the remaining clusters are too far apart (distance measure)
		// or that a desired number of clusters has been reached
		// -- we could use the latter to determine which households go into which village?

		ArrayList<ArrayList<NetworkNode>> clusters = new ArrayList<ArrayList<NetworkNode>>();
		for (Object o : list) {
			ArrayList<NetworkNode> cluster = new ArrayList<NetworkNode>();
			cluster.add((NetworkNode)o);
			clusters.add(cluster);
		}

		double minDist;
		do {
			minDist = Integer.MAX_VALUE;
			ArrayList<NetworkNode> closestA = null, closestB = null;			
			// determine the two "closest" clusters
			for (int i = 0; i < clusters.size()-1; i++){
				ArrayList<NetworkNode> clusterA = clusters.get(i);
				for (int j = i+1; j < clusters.size(); j++){
					ArrayList<NetworkNode> clusterB = clusters.get(j);
					double distance = computeDistance(clusterA, clusterB, distances);
					if (distance < minDist){
						minDist = distance; 
						closestA = clusterA;
						closestB = clusterB;
					}
				}
			}
			// merge the two clusters
			clusters.remove(closestA);
			clusters.remove(closestB);
			closestA.addAll(closestB);
			clusters.add(closestA);			
		} while (clusters.size() > numVillages);  /*(minDist < 0.7);*/
				
//		System.out.println("done (determining clusters)");
		return clusters;
	}
	
	/** helper method for determineClusters() */
	private double computeDistance(ArrayList<NetworkNode> clusterA, ArrayList<NetworkNode> clusterB, double[][] distances){
		double n = clusterA.size() * clusterB.size();
		double sum = 0;
		for (NetworkNode nodeA : clusterA){
			for (NetworkNode nodeB : clusterB){
				sum += distances[getNodeIndex(nodeA)][getNodeIndex(nodeB)];
			}
		}
		if (n > 0 ) return sum/n;
		return 1.0;
	}
	
	/** helper method for determineClusters() */
	private int getNodeIndex(NetworkNode node){
		// retrieve id and turn it into an integer = strip off the word "household-"
		String id = node.id.toString().substring(10);
		int intID = Integer.valueOf(id).intValue();
		return intID;
	}
	
	
	/** helper method for determineClusters() */
	private ArrayList<NetworkNode> getLinkedNodes(NetworkNode node){
		ArrayList<NetworkNode> linkedNodes = new ArrayList<NetworkNode>();
		ArrayList<NetworkLink> outEdges = node.getOutEdges();
		for (NetworkLink link : outEdges){
			linkedNodes.add(link.getTo());
		}
		return linkedNodes;
	}
	
	/** Initialisation related: create a mine and some mine workers from outside the village
	 */
	public void createMine(){
		dump("creating the mine");
		this.mine = new Mine("mine-0", this.neededSkilled, this.neededUnskilled, this);
		// ignore skilled workers for now; they'll probably not interact with the village anyway
		// create some in-migrants to work in the mine
		// -- fill about half of the available positions
		// -- make the workers male, between 23 and 36 years old
		// -- make some of them HIV positive
		// -- create friendship links between some of them (as if they come from the same village)
		dump("start creating miners from outside the village");
		this.minerSource = new MinerSource();
		ArrayList<Person> miners = minerSource.createMiners(neededUnskilled/2);
		// employ the miners at the mine
		for (Person miner : miners){
			mine.employUnskilledWorker(miner, tick);
		}
	}
	
		
	//---- model logic related --------------------------------------------------------

	/** procedural update of the model state, called each tick */
	void update(){
		// does procedural updates of model state
		if (month == 0 && week == 0) checkDeath();
		updateAgents();
		if (month == 0 && week == 0) {
			// at beginning of year determine births for that year
			calculateNextBirths();
		}
		for (int i = 0; i < nextBirths[month][week]; i++){
			// create a new person and find parents for him/her
			createNewPerson();			
		}
		if (week == 0) {
			// at the beginning of each month the grants are distributed
			supplyGrants();
			// households calculate their monthly food expenditure
			for (Household hh : householdList){
				hh.setMonthlyFoodCost(hh.calculateMonthlyFoodCost());
			}
		}
		// update from Jess facts
		jessLink.update();
	}
	
	/** Supply households with state grants: 
	 *  child grants for children < 7 and pensions for registered pensioners. 
	 *  For efficiency reasons, this is done in Java instead of Jess. The respective
	 *  sum is added directly to a household's cash.
	 */
	private void supplyGrants(){
		// supply households with state grants
		// registered children receive a child grant
		// registered pensioners receive a pension
		for (Household hh : householdList){
			double grants = 0;
			for (Person p : hh.getMembers()){
				if (p.isRegisteredForChildGrant()){
					grants += childGrant;
				}
				else if (p.isRegisteredForPension()){
					grants += statePension;
				}
			}
			if (grants > 0) {
//				System.out.println (hh.getId() + " receives " + grants + " in grants");
				dump(hh.getId().toString() + " receives " + grants + " in grants");
				hh.addCash(grants);
				if (chosenHH.contains(hh.getId().toString())){
					hhEconomyRecorder.recordIn(hh.getId().toString(), "grants", grants);
				}
				villageEconomyRecorder.record("outside", hh.getId().toString(), grants);
				hhIncomeRecorder.record(hh.getId().toString(), grants);
			}
		}
	}
	
	/** Creates a new agent during simulation, i.e. this models the birth of a person */
	void createNewPerson(){
		// find possible mothers
		ArrayList<Person> possibleMothers = getPossibleMothers();
		// for now: choose randomly from possible mothers
		Person mother = possibleMothers.get(Random.uniform.nextIntFromTo(0, possibleMothers.size()-1));
		// create new person
		Person baby = createHouseholdMember("person-" + personIndex++, mother.getHousehold());
		baby.setAge(0);
		baby.setBirthMonth(month);
		baby.setBirthWeek(week);
		checkDeath(baby);
		mother.setLastDelivery(tick);
		// check if child will receive child grant
		if (askChildGrantOracle(baby)){
			baby.setRegisteredForChildGrant(true);
		}
		// create family links
		createFamilyLinks(baby, mother, mother.getHousehold());
		// create kin links
		endorseKin(baby);
		// create neighbour links
		endorseNeighbours(baby);
		dump("NEW PERSON BORN INTO " + mother.getHousehold().getId() + " for " + mother.getName() + ", age " + mother.getAge());
		// if mother still at school, let her drop out
		if (mother.isAtSchool()){
			mother.setAtSchool(false);
		}
		// HIV/AIDS transmission
		if (mother.isHIVPositive() && probabilityp(motherChildHIVTransmissionRate)){	
			baby.setInfectedSince(tick);
		}
		// add new person to GUI (if necessary) and record birth in Jess
		// p has been born and has to be added to all gui (friendship, stokvel)
		// --> p has already been added to personList and namePersonMap in createHHMember()
		if (showGUI){
			NetworkNode node = baby.getNode();
			friendNetwork.addNode(node);
			stokvelNetwork.addNode(node);
		}
		// record birth in Jess
		try {
			Fact birthFact = new Fact("birth", engine);
			birthFact.setSlotValue("name", baby.getName());
			birthFact.setSlotValue("household", baby.getHouseholdID());
			birthFact.setSlotValue("tickStamp", new Value(tick, RU.INTEGER));
			engine.assertFact(birthFact);
		}
		catch(JessException je) {
			System.err.println("An error occurred at line " + je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}	
	}
	
	private Person pickMother(ArrayList<Person> possibleMothers){
		// sort possible mothers in ascending order of age
		
		return null;
	}
	
	/** creates family links between new person and members of the household
	 * 
	 * @param baby the new person
	 * @param mother their mother
	 * @param household their household
	 */
	private void createFamilyLinks(Person baby, Person mother, Household household){
		// create family links for new baby
		// (version with network nodes and links replaced by direct references)
		
		// 1. with mother
		baby.mother = mother.getName();
		// 2. with father if mother is married
		Person father = mother.getSpouse();
		if (father != null){
			baby.father = father.getName();
			father.addChild(baby);
		}
		// 3. with all siblings
		ArrayList<Person> siblings = retrieveObjects(mother.getChildren());
		for (Person sibling : siblings){
			baby.addSibling(sibling);
			sibling.addSibling(baby);
		}
		mother.addChild(baby); // add baby here so that it's not in the children list when doing the siblings		
	}
	
	/** helper function: determines possible mothers for a baby.
	 * 
	 * @return list of possible mothers
	 */
	private ArrayList<Person> getPossibleMothers(){
		// for now: just find all women of child-bearing age (15-45) who haven't given birth during the last 12 months
		// and -- if HIV infected -- whose infection isn't too advanced yet (< 8 years) ### ask Shah for source
		ArrayList<Person> mothers = new ArrayList<Person>();
		for (Person p : personList){
			if (p.getGender().toString().equals("female") && 
					p.getAge() >= 15 && p.getAge() <= 45 &&
					(p.getInfectedSince() < 0 || (tick - p.getInfectedSince() < 48*8) && !p.getHasAIDS()) &&
					(p.getLastDelivery() == -1 || p.getLastDelivery() < (tick - 48)))
			{
				mothers.add(p);
			}
		}
		return mothers;
	}
	
	/** Determines number and dates of births for the upcoming year */
	void calculateNextBirths(){
		// use birth rate and size of population to calculate number of births this year
		int numBirths = (int)(birthRate / 1000.0 * personList.size());
		// distribute them uniformly over the weeks of the year
		resetNextBirths();
		for (int i = 0; i < numBirths; i++){
			int month = Random.uniform.nextIntFromTo(0, 11);
			int week = Random.uniform.nextIntFromTo(0, 3);
			nextBirths[month][week]++;
		}
	}
	
	/** helper function: resets the data structure keeping track of upcoming births */
	private void resetNextBirths(){
		for (int i = 0; i < nextBirths.length; i++){
			for (int j = 0; j < nextBirths[i].length; j++){
				nextBirths[i][j] = 0;
			}
		}
	}
	
	/** Procedural update of agents (persons), once each tick. 
	 * Calls update() on each Person object, then removes all dead agents and left miners
	 */
	void updateAgents(){
		ArrayList<Person> died = new ArrayList<Person>();
		ArrayList<Person> left = new ArrayList<Person>();
		for (Person p : personList){
			p.update(this.tick);
			if (!p.isAlive()) died.add(p);
			if (p.left) left.add(p);
		}		
		// remove all dead agents
		for (Person p : died){
			removePerson(p);
		}
		// remove all miners who left the area
		ArrayList<String> names = new ArrayList<String>(left.size());
		for (Person p : left){
			removeMiner(p);
			numMiners--;
			names.add(p.getName().toString());
		}
		// remove their neighbour links
		removeNeighbourLinks(names);
	}
	
	
	/** Determines number and dates of deaths at the beginning of a year using the WHO life tables.
	 *  HIV-positive persons are treated with priority, i.e. they will be chosen first in their age group.
	 */
	void checkDeath(){
		// sort persons into age groups (only persons whose age is lower bound of an age interval)
		System.out.println("checkDeath: sorting persons into age groups");
		HashMap<Integer, ArrayList<Person>> males = new HashMap<Integer, ArrayList<Person>>();
		HashMap<Integer, ArrayList<Person>> females = new HashMap<Integer, ArrayList<Person>>();
		for (Person p : personList){
			WHOLifeTable table = (p.gender.toString().equalsIgnoreCase("female"))? femalesTable : malesTable;
			HashMap<Integer, ArrayList<Person>> list = (p.gender.toString().equalsIgnoreCase("female"))? females : males;
			if (table.isLowerBound(p.age + 1)){
				int index = table.findAgeInterval(p.age + 1);
				ArrayList<Person> indexList;
				if (list.containsKey(index)){
					indexList = list.get(index);
				}
				else {
					indexList = new ArrayList<Person>();
				}
				// put HIV positives due to develop AIDS in front, others in back of list
				if (p.getHasAIDS()){ 
					indexList.add(0, p);
				}
				else {
					indexList.add(p);
				}
				list.put(index, indexList);
			}
		}
		System.out.println("done. Now determining deaths for males...");
		// for each age group in males and females, determine the persons going to die in that interval
		for (Iterator<Integer> it = males.keySet().iterator(); it.hasNext(); ){
			int index = it.next();
			ArrayList<Person> ageGroup = males.get(index);
			checkDeath(ageGroup, malesTable);
		}
		System.out.println("done. Now determining deaths for females...");
		for (Iterator<Integer> it = females.keySet().iterator(); it.hasNext(); ){
			int index = it.next();
			ArrayList<Person> ageGroup = females.get(index);
			checkDeath(ageGroup, femalesTable);
		}
	}
	
	/** helper function for checkDeath():  determines deaths for given age group
	 *  according to given life table.
	 *  First sorts age group so that people with AIDS are in front, followed by
	 *  HIV-positives, followed by healthy people.
	 *  Then determines number of deaths (by drawing n samples from a random uniform
	 *  distribution and checking how many of them are smaller than the death probability
	 *  for the age group).
	 *  Then picks the first m people in the list and assigns a random death week 
	 *  within the interval of the age group.
	 */
	void checkDeath(ArrayList<Person> ageGroup, WHOLifeTable table){
		// determine number of AIDS patients
		int index = 0;
		while (index < ageGroup.size() && ageGroup.get(index).getHasAIDS()){
			index++;
		}
		// index now points to position of first person after AIDS patients
		// sort rest of ageGroup into HIV positives and others by inserting HIV positives at position index
		int i = index;
		while (i < ageGroup.size()){
			if (ageGroup.get(i).isHIVPositive()){
				ageGroup.add(index, ageGroup.remove(i));
			}
			i++;
		}
		// ageGroup now has AIDS patients in front, followed by HIV positives, followed by others
		
		// determine number of deaths 
		int age = ageGroup.get(0).age + 1;
		double deathProbability = table.getDeathProbability(age);
		int numDeaths = 0;
		for (i = 0; i < ageGroup.size(); i++){
			if (Random.uniform.nextDouble() <= deathProbability){
				numDeaths++;
			}
		}
		// now take first numDeaths persons out of ageGroup to die 
		for (i = 0; i < numDeaths; i++){
			// choose a random week from within this interval
			int timeToDie = Random.uniform.nextIntFromTo(1,table.getAgeIntervalSize(age)*48);
			// assign it to person as tick to die in
			Person p = ageGroup.get(i);
			p.setDeathWeek(tick + p.birthMonth*4 + p.birthWeek + timeToDie - 1);
		}		
	}
	
	/** Old version of checkDeath(). Still called for new persons to check for infant
	 * mortality.
	 * @param p a newborn agent
	 */
	void checkDeath(Person p){
		// use WHO life tables to determine probability of (natural?) death at current age
		// http://www3.who.int/whosis/life/life_tables/life_tables_process.cfm?country=zaf&language=en#
		WHOLifeTable table = (p.gender.toString().equalsIgnoreCase("female"))? femalesTable : malesTable;
		// need only calculate anything when person's age is lower bound of an age interval
		int age = p.getAge();
		if (!table.isLowerBound(age)) return;
		// retrieve probability
		double prob = table.getDeathProbability(age);
		if (Random.uniform.nextDouble() <= prob){
			// person will die in current age interval
			// choose a random week from within this interval
			int timeToDie = Random.uniform.nextIntFromTo(1,table.getAgeIntervalSize(age)*48);
			// assign it to person as tick to die in
			p.setDeathWeek(this.tick + timeToDie - 1);
		}
	}
	
	/** Determines if the given person will receive a pension. This is done randomly with
	 *  probabilities 0.58 for females over 60 and 0.5 for males over 65 (RADAR data).
	 * 
	 * @param p a senior
	 * @return true if person will receive a pension, false otherwise
	 */
	public boolean askPensionOracle(Person p){
		// from RADAR data we know that 58% of females >= 60 and 50% of males >= 65 receive state pensions
		if ((p.gender.toString() == "female") && (p.getAge() >= pensionAgeLimitFemale)){
			if (Random.uniform.nextDouble() < pensionProportionFemale) return true;
		}
		if ((p.gender.toString()=="male") && (p.getAge() >= pensionAgeLimitMale)){
			if (Random.uniform.nextDouble() < pensionProportionMale) return true;
		}
		return false;
	}
	
	/** Determines if the given person will receive a child grant. This is done randomly with
	 *  the proportion set as a model parameter.
	 */
	public boolean askChildGrantOracle(Person p){
		if (p.getAge() < childGrantAgeLimit && probabilityp(childGrantProportion)){
			return true;
		}
		return false;
	}
	
	/** Removes a dead person from the model and records his/her death.
	 * 
	 * @param p the dead person
	 */
	public void removePerson(Person p){
		// p has died and has to be removed from the person list and all gui (friendship, stokvel)
		personList.remove(p);
		deathRecorder.record(p);
		if (showGUI){
			NetworkNode node = p.getNode();
			friendNetwork.removeNode(node);
			stokvelNetwork.removeNode(node);
		}
		// remove p from household
		p.getHousehold().removeMember(p);
		// the links p had to others plus the links others had to p have to be deleted
		removeLinks(p);
		// remove shadow fact from Jess --> no, has to be done in Jess at the end of the tick
	}
	
	/** Removes an external miner from the model
	 * @param p the miner who is leaving the area
	 */
	public void removeMiner(Person p){
		personList.remove(p);
		if (showGUI){
			NetworkNode node = p.getNode();
			friendNetwork.removeNode(node);
			stokvelNetwork.removeNode(node);
		}
		// the links p had to others plus the links others had to p have to be deleted
		removeLinks(p);
		for (Person partner : p.currentPartners){
			partner.removeCurrentPartner(p);
		}
		p.removeCurrentPartners();
		for (Person potentialFriend : personList){
			if (potentialFriend.hasFriend(p)){
				potentialFriend.removeFriend(p);
				dump(potentialFriend.getName() + " removes " + p.getName() + " from the list of friends");
			}
		}
		p.friends.clear();
		// remove shadow fact from Jess
//		try {
//			engine.undefinstance(p);
//		}
//		catch (JessException je){
//			System.err.println("Jess engine.undefinstance() for " + p.getName() + " didn't work!");
//			System.err.println("An error occurred at line " + je.getLineNumber());
//			System.err.println("caused by " + je.getCause());
//			System.err.println("Message: " + je.getDetail());
//			je.printStackTrace();
//		}
	}
	
	/** Helper function: removes the links of the given dead person. 
	 * @param p the dead person
	 */
	private void removeLinks(Person p){
		NetworkNode node = p.getNode();
		ArrayList<NetworkLink> toBeDeleted = new ArrayList<NetworkLink>();
		for (Object o : node.getOutEdges()){
			NetworkLink link = (NetworkLink)o;
			// leave family links intact so they can be used to access the extended family
			if (!link.getType().startsWith("family:")){
//				System.out.println("remove " + link.toString());
				NetworkNode other = link.getTo();
				other.removeInEdge(link);
				toBeDeleted.add(link);
			}
		}
		for (NetworkLink link : toBeDeleted){
			node.removeOutEdge(link);
		}
		toBeDeleted = new ArrayList<NetworkLink>();
		for (Object o : node.getInEdges()){
			NetworkLink link = (NetworkLink)o;
			// leave family links intact so they can be used to access the extended family
			if (!link.getType().startsWith("family:")){
//				System.out.println("remove " + link.toString());
				NetworkNode other = link.getFrom();
				other.removeOutEdge(link);
				toBeDeleted.add(link);
			}
		}
		for (NetworkLink link : toBeDeleted){
			node.removeInEdge(link);
		}
		toBeDeleted = null;
	}
	
	
	/** Removes the given dissolved household from the model
	 * 
	 * @param hh the household to be removed
	 */
	public void removeHousehold(Household hh){
		// hh has dissolved and has to be removed from the household list and all gui (kinship, space)
		householdList.remove(hh);
		space.putObjectAt(hh.getX(), hh.getY(), null);
		if (showGUI){
			NetworkNode node = hh.getNode();
			kinshipNetwork.removeNode(node);
		}
		// the links hh had to others plus the links others had to hh have to be deleted
		removeLinks(hh);
		// remove shadow fact from Jess
		try {
			engine.undefinstance(hh);
		}
		catch(JessException je) {
			System.err.println("An error occurred at line " + je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}	
	}

	/** helper method for removeHousehold(): 
	 * removes the links of the given dissolved household
	 * 
	 * @param hh the dissolved household
	 */
	private void removeLinks(Household hh){
		NetworkNode node = hh.getNode();
		ArrayList<NetworkLink> toBeDeleted = new ArrayList<NetworkLink>();
		for (Object o : node.getOutEdges()){
			NetworkLink link = (NetworkLink)o;
			NetworkNode other = link.getTo();
			other.removeInEdge(link);
			toBeDeleted.add(link);
		}
		for (NetworkLink link : toBeDeleted){
			node.removeOutEdge(link);
		}
		toBeDeleted = new ArrayList<NetworkLink>();
		for (Object o : node.getInEdges()){
			NetworkLink link = (NetworkLink)o;
			NetworkNode other = link.getFrom();
			other.removeOutEdge(link);
			toBeDeleted.add(link);
		}
		for (NetworkLink link : toBeDeleted){
			node.removeInEdge(link);
		}
		toBeDeleted = null;
	}
	
	
	/** The main method: runs the model */
	public static void main(String[] args) {
		uchicago.src.sim.engine.SimInit init = new uchicago.src.sim.engine.SimInit();
		SouthAfricaModel model = new SouthAfricaModel();
		init.loadModel (model, null, false);
	}
	

	/** returns true with the given probability p.
	 * 
	 * @param p the given propability
	 * @return true if the random number sampled from a uniform distribution is lower than p,
	 * 		otherwise false
	 */
	public boolean probabilityp(double p){
		boolean result = false;
		if (Random.uniform.nextDoubleFromTo(0.0, 1.0) < p)
			result = true;
		return result;
	}

	/** samples a random uniform distribution in the interval from-to. 
	 * Provided for use in Jess rules.
	 * 
	 * @param from lower bound
	 * @param to upper bound
	 * @return random number (int) between from and to
	 */
	public int randomInt(int from, int to){
		// for use in Jess
		return Random.uniform.nextIntFromTo(from, to);
	}
	
	/** Make a happy couple out of a male and female agent. 
	 *  The bride moves into the household of the groom.
	 * 
	 * @param maleName name of the male
	 * @param femaleName name of the female
	 */
	public void marry(Value maleName, Value femaleName){
		Person groom = namePersonMap.get(maleName);
		Person bride = namePersonMap.get(femaleName);
		Household groomHH = groom.getHousehold();
		Household brideHH = bride.getHousehold();
		dump("MARRIAGE!");
		dump(groom.getName() + ", age " + groom.getAge() + ", of " + groomHH.getId() +
				" will marry " + bride.getName() + ", age " + bride.getAge() + ", of " + brideHH.getId());
		// set them up as spouses
		groom.marry(bride);
		bride.marry(groom);
		// link the households as kin?? --> NO according to storyline validation
//		if (!groomHH.isKinOf(brideHH)){
//			NetworkLink link = new NetworkLink(groomHH.getNode(), brideHH.getNode());
//			link.setType("kinship");
//			link = new NetworkLink(brideHH.getNode(), groomHH.getNode());
//			link.setType("kinship");
//			endorseKin(groomHH, brideHH);
//			endorseKin(brideHH, groomHH);
//		}
		// have the bride move into the groom's household
		// ### what if the bride's the only one left in her household? Wouldn't they rather move into her house?
		// ### or more generally: if the bride's head of her household and the groom is only a member of his household	
		// --> they won't move into her house because that would be like paying lobola to him
		// --> the house needs to be sold/need to find someone to live there
		brideHH.removeMember(bride);
		bride.setHousehold(groomHH); // adds bride to groomHH automatically
		dump(bride.getName() + " moves into household of " + groom.getName());
		// ### have to adjust her neighbour links!!
	}
	

	/** wrapper for the marry() method: turns the parameters of type String into
	 * parameters of type Value.
	 * 
	 * @param maleNameString	name of the male agent as a String
	 * @param femaleNameString  name of the female agent as a String
	 */
	public void marry(String maleNameString, String femaleNameString){
		try {
			marry(new Value(maleNameString, RU.SYMBOL), new Value(femaleNameString, RU.SYMBOL));
		}
		catch (JessException je){
			System.err.println("An error occurred at line " + je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}
	}

	/** Create a new household with the given person as head. This happens when a
	 * couple has got enough money to move out from the husband's parents' household. 
	 * 
	 * @param head the household head
	 * @return name of the new household
	 * @throws JessException
	 */
	public Value buildNewHouse(Person head) throws JessException {
		// create a new household
		Household oldHH = head.getHousehold();
		Village village = this.nameVillageMap.get(oldHH.getVillage().toString());
		String hhID = "household-" + this.householdList.size();
		Household hh = new Household(this, hhID, oldHH.getChurch(), village);
		this.nameHouseholdMap.put(hhID, hh);
		this.householdList.add(hh);		
		// head and spouse (and possible children) are moving out
		oldHH.removeMember(head);
		head.setHousehold(hh);
		head.setHead(true);
		if (head.spouse != null){
			oldHH.removeMember(head.spouse);
			head.spouse.setHousehold(hh);
		}
		ArrayList<Person> children = retrieveObjects(head.getChildren());
		ArrayList<Person> herChildren = retrieveObjects(head.spouse.getChildren());
		for (Person child : herChildren){
			if (!children.contains(child))
				children.add(child);
		}
		for (Person child : children){
			oldHH.removeMember(child);
			child.setHousehold(hh);
		}
		
		StringBuffer text = new StringBuffer("Moving into the new house are:");
		for (Person p : hh.getMembers()){
			text.append(" ");
			text.append(p.getName().toString());
		}
		dump(text.toString());
		
		// create kinship links on the hh level ####
		// several options:
		// (1) keep new hh in kinship network of husband --> copy kinship links of oldHH
		//     (since the households of husband and wife are kin through their marriage, this would at least
		//      give the new household one kinship link to the wife's kinship network (namely: her old hh)
		// 	Q: should the wife endorse her husband's family with is-kin when she moves into his hh???
		// (2) have new hh be part of both kinship networks --> copy kinship links of oldHH and wife's parents' hh
		// 	Q: should this be reflected on the endorsement level, i.e. husband endorses wife's family and vice versa?
		// (3) use the intersection of both is-kin endorsement sets = only make kinship links to both parents' hhs
		// (4) have the new hh keep a subset of both kinship networks
		//		Q: selection criteria? Any reason why this would be so? Yes, some relations, eg. cousins on parents' level, become more distant for children
		
		// Validation exercise in village reveals: it's option 1. 
		// The question remains though if it's correct to link husband's and wife's old households as kin
		
		NetworkNode hhNode = hh.getNode();
		NetworkNode oldHHNode = oldHH.getNode();
		for (Object o : oldHHNode.getOutEdges()){
			NetworkLink link = (NetworkLink)o;
			if (link.getType().equals("kinship")){
				NetworkNode toNode = link.getTo();
				new NetworkLink(hhNode, toNode, "kinship");  // adds link to nodes automatically
				new NetworkLink(toNode, hhNode, "kinship");
				Household toHH = findHouseholdOf(toNode);
				endorseKin(hh, toHH);
				endorseKin(toHH, hh);
			}
		}
		// add a link to the old hh and vice versa
		new NetworkLink(hhNode, oldHHNode, "kinship");
		new NetworkLink(oldHHNode, hhNode, "kinship");
		endorseKin(hh, oldHH);
		endorseKin(oldHH, hh);
		// add to gui
		if (showGUI){
			kinshipNetwork.addNode(hhNode);
		}
		
		// have to delete the old neighbour links for all members of the new household
		removeNeighbourLinks(hh.getMemberIDs());

		// create new neighbour links
		for (Household hh2 : householdList){
			if (hh2.getVillage().equals(village.getName()) && hh != hh2){
				double distance = Math.sqrt((hh.x - hh2.x) * (hh.x - hh2.x)
														+ (hh.y - hh2.y) * (hh.y - hh2.y));
				if (distance < neighbourhoodRadius){
					for (Person p : hh.members){
						for (Person p2 : hh2.members) {
							p.assertNeighbourEndorsement(p2);
							p2.assertNeighbourEndorsement(p);
						}
					}
				}
			}
		}
		// put household into Jess
		try {
			engine.definstance("household", hh, true);
		}
		catch (JessException je){
			System.err.println("An error occurred at line " + je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();			
		}
		// add hh to village economy recorder
		villageEconomyRecorder.addHousehold(hhID);
		return new Value(hhID, RU.SYMBOL);		
	}
	
	/** helper function: wrapper for same method with ArrayList parameter */
	void removeNeighbourLinks(Value[] list){
		ArrayList<String> temp = new ArrayList<String>(list.length);
		for (Value elem : list){
			temp.add(elem.toString());
		}
		removeNeighbourLinks(temp);
	}
	
	/** helper function: removes neighbour links (is-neighbour endorsements) for 
	 * the given list of persons (members of a household, group of leaving miners)
	 * @param list list of names
	 */
	void removeNeighbourLinks(ArrayList<String> list){
		// find the neighbour endorsement facts
		ArrayList<Fact> disposalList = new ArrayList<Fact>();
		Iterator factList = engine.listFacts();
		Context context = engine.getGlobalContext();
		while (factList.hasNext()){
			try {
				Fact f = (Fact)factList.next();
				if (f.getName().endsWith("endorsed")){
					String endorsement = f.getSlotValue("endorsement").stringValue(context);
					if (endorsement.equalsIgnoreCase("is-neighbour")){
						// it's a neighbour endorsement 
						// --> see if clauseOwner or endorsed-object are in the list of names
						String owner = f.getSlotValue("clauseOwner").stringValue(context);
						String endorsee = f.getSlotValue("endorsed-object").stringValue(context);
						if (list.contains(owner) || list.contains(endorsee)){
							disposalList.add(f);
						}
					}
				}
			} catch (JessException je) {
				System.err.println("An error occurred at line " + je.getLineNumber());
				System.err.println("caused by " + je.getCause());
				System.err.println("Message: " + je.getDetail());
				je.printStackTrace();
			}
		}	
		// retract the neighbour endorsements
		for (Fact f : disposalList){
			try {
				engine.retract(f);
			} catch (JessException je) {
				System.err.println("An error occurred at line " + je.getLineNumber());
				System.err.println("caused by " + je.getCause());
				System.err.println("Message: " + je.getDetail());
				je.printStackTrace();
			}
		}		
	}

	/** returns the Person objects for the given list of names. 
	 *  For use with family relations and friends.
	 *  @param names list of agents' names
	 *  @return list of Person objects
	 */
	ArrayList<Person> retrieveObjects(Value[] names){
		ArrayList<Person> list = new ArrayList<Person>(names.length);
		for (Value name : names){
			list.add(namePersonMap.get(name.toString()));
		}
		return list;
	}
	
	
	//---- model output related -----------------------------------------------------

	/** Output the given text to console (if showOutput is true) and/or 
	 *  trace file (if printToFile is true).
	 *  
	 * @param text the text to be output
	 */
	public void dump(String text){
		if (showOutput){
			System.out.println(text);
		}
		if (printToFile){
			printOut.println(text);
		}
	}

	/** returns the number of households asking for food during the current month
	 * 
	 * @return number of households asking for food
	 */
	public int getNumHHAskingForFood() {
		return numHHAskingForFood;
	}

	/** sets the number of households asking for food during the current month
	 * 
	 * @param numHHAskingForFood number of households asking for food
	 */
	protected void setNumHHAskingForFood(int numHHAskingForFood) {
		this.numHHAskingForFood = numHHAskingForFood;
	}

	/** returns the number of households with a lack of food even after having asked
	 * relatives/neighbours for help
	 * 
	 * @return number of households without enough food
	 */
	public int getNumHHWithoutEnoughFood() {
		return numHHWithoutEnoughFood;
	}

	/** sets the number of households with a lack of food even after having asked
	 * relatives/neighbours for help
	 * 
	 * @param numHHWithoutEnoughFood number of households without enough food
	 */
	protected void setNumHHWithoutEnoughFood(int numHHWithoutEnoughFood) {
		this.numHHWithoutEnoughFood = numHHWithoutEnoughFood;
	}

	/** increases the number of households asking for food by 1 */
	public void incNumHHAskingForFood(){
		this.numHHAskingForFood++;
	}
	
	/** increases the number of households without enough food by 1 */
	public void incNumHHWithoutEnoughFood(){
		this.numHHWithoutEnoughFood++;
	}
	
	/** returns the highest amount of cash one of the households owns in the model
	 * 
	 * @return max amount of cash
	 */
	public double getMaxHHCash(){
		double max = 0;
		for (Household hh : this.householdList){
			if (hh.cash > max) max = hh.cash;
		}
		return max;
	}

	/** returns the smallest amount of cash one of the households owns in the model
	 * 
	 * @return min amount of cash
	 */
	public double getMinHHCash(){
		double min = 10000;
		for (Household hh : this.householdList){
			if (hh.cash < min) min = hh.cash;
		}
		return min;
	}	
	
	/** returns the average amount of cash over all households
	 * 
	 * @return average amount of cash
	 */
	public double getAvgHHCash(){
		double sum = 0;
		for (Household hh : this.householdList){
			sum += hh.cash;
		}
		return (sum / (double)householdList.size());
	}
	
	/** resets the output variables recorded on a monthly basis */
	public void resetOutputVariables(){
		setNumHHAskingForFood(0);
		setNumHHWithoutEnoughFood(0);
		hhIncomeRecorder.reset();
	}
	
	/** returns the number of HIV-positive persons
	 * 
	 * @return number of HIV-positive agents
	 */
	public int getNumHIVPositives(){
		int sum = 0;
		for (Person p : personList){
			if (p.isHIVPositive()){
				sum++;
			}
		}
		return sum;
	}
	
	/** returns the number of female HIV-positive persons
	 * 
	 * @return number of female HIV-positive agents
	 */
	public int getNumFemaleHIVPositives(){
		int sum = 0;
		for (Person p : personList){
			if (p.gender.toString().equalsIgnoreCase("female") && p.isHIVPositive()){
				sum++;
			}
		}
		return sum;
	}
	
	/** returns the number of HIV-positive miners (in-migrants)
	 * 
	 * @return number of HIV-positive miners
	 */
	public int getNumMinerHIVPositives(){
		int sum = 0;
		for (Person p : personList){
			if (p.name.toString().startsWith("miner") && p.isHIVPositive()){
				sum++;
			}
		}
		return sum;
	}
	
	
	/** returns the number of persons
	 * 
	 * @return number of agents
	 */
	public int getNumPersons(){
		return personList.size();
	}
	
	/** calculates the HIV incidence for the last month
	 * 
	 * @return number of persons newly infected within the last month
	 */
	public int getHIVIncidence(){
		if (this.tick == 0) return 0;
		int sum = 0;
		// sum up all persons newly infected within the last month
		for (Person p : personList){
			if (p.getInfectedSince() >= 0 && p.getInfectedSince() >= (this.tick - 4)){
				sum++;
			}
		}
		return sum;
	}
	
	/** returns the number of agents who are members of a stokvel
	 * 
	 * @return number of stokvel members
	 */
	public int getNumStokvelMembers(){
		int sum = 0;
		for (Person p : personList){
			if (p.stokvels.size() > 0){
				sum++;
			}
		}
		return sum;
	}
	
	/** returns the number of agents who are members of a burial society
	 * 
	 * @return number of burial society members
	 */
	public int getNumBSMembers(){
		int sum = 0;
		for (Person p : personList){
			if (p.burialSocieties.size() > 0){
				sum++;
			}
		}
		return sum;		
	}
	
	/** compiles the sexual network to be recorded
	 * 
	 * @return the sexual network as a hash map (list of partners per agent)
	 */
	public HashMap<String, String[]> compileSexualNetwork(){
		HashMap<String, String[]> network = new HashMap<String, String[]>();
		for (Person p : personList){
			ArrayList<String> partners = new ArrayList<String>();
			if (p.currentPartners != null && p.currentPartners.size() > 0){
				for (Person cp : p.currentPartners){
					partners.add(cp.getName().toString());
				}
			}
			if (p.spouse != null){
				partners.add(p.spouse.getName().toString());
			}
			if (partners.size() > 0){
				String[] list = new String[partners.size()];
				list = partners.toArray(list);
				network.put(p.getName().toString(), list);
			}
		}
		return network;
	}
	
	/** calculates the cluster coefficient of the friend network applying the algorithm
	 * from Repast
	 * 
	 * @return cluster coefficient
	 */
	public double calculateClusterCoefficient(){
		return NetworkUtilities.calcClustCoef(personList, "friend");
//		// obtain node list
//		ArrayList<NetworkNode> nodeList = new ArrayList<NetworkNode>(personList.size());
//		for (Person p : personList){
//			nodeList.add(p.getNode());
//		}
//		return NetworkUtilities.calcClustCoef(nodeList, "friend");
	}

	/** records the given event for the given stokvel
	 * 
	 * @param stokvel name of a stokvel
	 * @param event type of event
	 */
	public void recordStokvelEvent(String stokvel, String event){
		stokvelRecorder.recordEvent(stokvel, event);
	}
	
	/** records the given event regarding the given member for the given stokvel
	 * 
	 * @param stokvel name of a stokvel
	 * @param event type of event
	 * @param member agent (stokvel member)
	 */
	public void recordStokvelEvent(String stokvel, String event, Person member){
		stokvelRecorder.recordEvent(stokvel, event, member);
	}

	/** returns the HIV prevalence amongst in-migrating mine workers
	 * 
	 * @return HIV prevalence 
	 */
	public double getMineHIVPrevalence() {
		return mineHIVPrevalence;
	}

	/** sets the HIV prevalence amongst in-migrating mine workers
	 * 
	 * @param mineHIVPrevalence HIV prevalence
	 */
	public void setMineHIVPrevalence(double mineHIVPrevalence) {
		this.mineHIVPrevalence = mineHIVPrevalence;
	}

	/** returns the tick when mining starts in the simulation
	 * 
	 * @return start tick of mining
	 */
	public int getMiningTick() {
		return miningTick;
	}

	/** sets the tick when mining starts in the simulation
	 * 
	 * @param miningTick start tick of mining
	 */
	public void setMiningTick(int miningTick) {
		this.miningTick = miningTick;
	}

	/** returns the number of skilled workers needed in the mine
	 * 
	 * @return number of skilled workers
	 */
	public int getNeededSkilled() {
		return neededSkilled;
	}

	/** sets he number of skilled workers needed in the mine
	 * 
	 * @param neededSkilled number of skilled workers
	 */
	public void setNeededSkilled(int neededSkilled) {
		this.neededSkilled = neededSkilled;
	}

	/** returns the number of unskilled workers needed in the mine
	 * 
	 * @return number of unskilled workers
	 */
	public int getNeededUnskilled() {
		return neededUnskilled;
	}

	/** sets he number of unskilled workers needed in the mine
	 * 
	 * @param neededUnskilled number of unskilled workers
	 */
	public void setNeededUnskilled(int neededUnskilled) {
		this.neededUnskilled = neededUnskilled;
	}
	
	/** returns the number of employed unskilled workers in the mine
	 * 
	 * @return number of unskilled workers
	 */
	public int getNumUnskilled(){
		if (this.mine != null){
			return mine.getNumUnskilled();
		}
		return 0;
	}
	
	/** returns the number of in-migrant miners in the area
	 * @return number of miners
	 */
	public int getNumMiners(){
		return this.numMiners;
	}
	
	/** returns the number of in-migrant miners employed at the mine
	 * 
	 * @return number of employed in-migrant miners
	 */
	public int getNumEmployedMiners(){
		if (this.mine != null){
			return mine.getNumberOfMigrantWorkers();
		}
		return 0;
	}
	
	/** returns the current number of households in the model. Used for
	 * model output.
	 * @return current number of households
	 */
	public int getCurrentNumHouseholds(){
		if (this.householdList == null) return 0;
		return this.householdList.size();
	}
	
	/** returns the current number of migrants from the villages. Used for model output.
	 * @return current number of migrants
	 */
	public int getNumMigrants(){
		int num = 0;
		for (Person p : this.personList){
			if (p.isMigrant() || p.isOnVisit()){
				num++;
			}
		}
		return num;
	}
	
	/** returns the number of migrants currently on visit in the village(s).
	 * @return current number of visiting migrants
	 */
	public int getNumMigrantsOnVisit(){
		int num = 0;
		for (Person p : this.personList){
			if (p.isOnVisit()){
				num++;
			}
		}
		return num;
	}
	
	/** checks if the specified agents are fit for marriage, i.e. if the female
	 * is out of school and old enough to marry (>= 16) and if they've been 
	 * together long enough (min. 2 months). This function is called from the Jess
	 * rule findSuitablePartner.
	 * 
	 * @param male the name of the male partner
	 * @param female the name of the female partner
	 * @return true if the partners are fit for marriage, otherwise false.
	 */
	public boolean fitForMarriage(String male, String female){
		// check if the female agent is old enough to marry and out of school
		// and if male and female have been together long enough
		Person femalePerson = namePersonMap.get(female);
		Person malePerson = namePersonMap.get(male);
		if (femalePerson.isAtSchool() || femalePerson.getAge() < 16) {
			return false;
		}
		if (malePerson.currentPartnerDuration(femalePerson) < 8) {
			return false;
		}
		return true;
	}
	
	/** wrapper for the fitForMarriage method */
	public Value fitForMarriage (Value male, Value female){
		boolean result = fitForMarriage(male.toString(), female.toString());
		if (result) return Funcall.TRUE;
		return Funcall.FALSE;
	}
	

	//---- getter and setter methods ----------------------------------------------
	//---- with propagation of property change for Jess (since model is a shadow fact)
	
	/** returns the current month 
	 * @return month as int between 0 and 11
	 */
	public int getMonth() {
		return month;
	}

	/** sets the current month */
	public void setMonth(int m) {
		int old = month;
		month = m;
		pcs.firePropertyChange("month", old, month);	
	}

	/** returns the current week
	 * @return week as int between 0 and 3
	 */
	public int getWeek() {
		return week;
	}

	/** sets the current week */
	public void setWeek(int wk) {
		int old = week;
		week = wk;
		pcs.firePropertyChange("week", old, week);	

	}

	/** returns the current year
	 * @return year as int starting from 0
	 */
	public int getYear() {
		return year;
	}

	/** sets the current year */
	public void setYear(int yr) {
		int old = year;
		year = yr;
		pcs.firePropertyChange("year", old, year);	

	}

	/** adds a property change listener (for Jess) */
	public void addPropertyChangeListener(PropertyChangeListener p) {
		pcs.addPropertyChangeListener(p);
	}

	/** removes a property change listener (for Jess) */
	public void removePropertyChangeListener(PropertyChangeListener p) {
		pcs.removePropertyChangeListener(p);
	}

	/** returns the current tick 
	 * @return tick as int starting from 0
	 */
	public int getTick() {
		return tick;
	}

	/** sets the current tick */
	public void setTick(int t) {
		int old = tick;
		tick = t;
		pcs.firePropertyChange("tick", old, tick);	
	}

	/** computes and sets the current date values (year, month, week) */
	private void setDateValues(){
		setYear(tick / 48);
		setMonth((tick - (48 * year)) / 4);
		setWeek(tick % 4);
	}
	
	//---- getter and setter methods for model parameters -------------------

	/** returns the number of households with which to initialise the model
	 * @return the initial number of households
	 */
	public int getNumHouseholds() {
		return numHouseholds;
	}

	/** sets the number of households with which to initialise the model
	 * 
	 * @param numHouseholds the initial number of households
	 */
	public void setNumHouseholds(int numHouseholds) {
		this.numHouseholds = numHouseholds;
	}

	/** returns the seed to be used with random generators
	 * 
	 * @return the seed
	 */
	public int getSeed() {
		return seed;
	}

	/** sets the seed to be used with random generators
	 * 
	 * @param seed the seed
	 */
	public void setSeed(int seed) {
		this.seed = seed;
	}
	
	/** returns if the GUI is to be displayed or not
	 * 
	 * @return true if the GUI is on, false if the GUI is off
	 */
	public boolean getShowGUI(){
		return this.showGUI;
	}
	
	/** sets if the GUI is to be displayed or not
	 * 
	 * @param showGUI true to turn the GUI on, false to turn it off
	 */
	public void setShowGUI(boolean showGUI){
		this.showGUI = showGUI;
	}

	/** returns the adult employment rate
	 * 
	 * @return adult employment rate
	 */
	public double getAdultEmploymentRate() {
		return adultEmploymentRate;
	}

	/** sets the adult employment rate
	 * 
	 * @param rate proportion of adult agents with a job
	 */
	public void setAdultEmploymentRate(double rate) {
		adultEmploymentRate = rate;
	}

	/** returns the government employment rate, i.e.
	 * the proportion of adult agents with a government job
	 * @return government employment rate
	 */ 
	public double getGovernmentEmploymentRate() {
		return governmentEmploymentRate;
	}

	/** sets the government employment rate, i.e.
	 * the proportion of adult agents with a government job
	 * @param governmentEmploymentRate government employment rate
	 */ 
	public void setGovernmentEmploymentRate(double governmentEmploymentRate) {
		this.governmentEmploymentRate = governmentEmploymentRate;
	}
	
	/** returns the size in dimension x of the 2D grid used as a spatial model
	 * 
	 * @return the number of cells in x 
	 */
	public int getGridSizeX() {
		return gridSizeX;
	}

	/** sets the size in dimension x of the 2D grid used as a spatial model
	 * 
	 * @param gridSizeX the number of cells in x 
	 */
	public void setGridSizeX(int gridSizeX) {
		this.gridSizeX = gridSizeX;
	}

	/** returns the size in dimension y of the 2D grid used as a spatial model
	 * 
	 * @return the number of cells in y
	 */
	public int getGridSizeY() {
		return gridSizeY;
	}

	/** sets the size in dimension y of the 2D grid used as a spatial model
	 * 
	 * @param gridSizeY the number of cells in y 
	 */
	public void setGridSizeY(int gridSizeY) {
		this.gridSizeY = gridSizeY;
	}

	/** returns the number of villages
	 * 
	 * @return the number of villages
	 */
	public int getNumVillages() {
		return numVillages;
	}

	/** sets the number of villages
	 * 
	 * @param numVillages the number of villages
	 */
	public void setNumVillages(int numVillages) {
		this.numVillages = numVillages;
	}

	/** returns the density factor determining how densely the villages
	 *  are populated with households (1.0 = very compact, every cell filled
	 *  with a house)
	 * 
	 * @return the density factor
	 */
	public double getDensityFactor() {
		return densityFactor;
	}

	/** sets the density factor determining how densely the villages
	 *  are populated with households (1.0 = very compact, every cell filled
	 *  with a house)
	 * 
	 * @param densityFactor the density factor
	 */
	public void setDensityFactor(double densityFactor) {
		if (densityFactor < 1.0) densityFactor*=-1;
		this.densityFactor = densityFactor;
	}

	/** returns the birth rate (number of births per 1000 persons
	 * in the population)
	 * 
	 * @return the birth rate
	 */
	public double getBirthRate() {
		return birthRate;
	}

	/** sets the birth rate (number of births per 1000 persons
	 * in the population)
	 * 
	 * @param birthRate the birth rate
	 */
	public void setBirthRate(double birthRate) {
		this.birthRate = birthRate;
	}

	/** returns the neighbourhood radius 
	 * 
	 * @return the neighbourhood radius
	 */
	public double getNeighbourhoodRadius() {
		return neighbourhoodRadius;
	}

	/** sets the neighbourhood radius 
	 * 
	 * @param neighbourhoodRadius the neighbourhood radius
	 */
	public void setNeighbourhoodRadius(double neighbourhoodRadius) {
		this.neighbourhoodRadius = neighbourhoodRadius;
	}

	/** returns the tag length (number of tags)
	 * 
	 * @return the number of tags per agent
	 */
	public int getTagLength() {
		return tagLength;
	}

	/** sets the tag length (number of tags)
	 * 
	 * @param tagLength the number of tags per agent
	 */
	public void setTagLength(int tagLength) {
		this.tagLength = tagLength;
	}

	/** returns the tag base (number of values per tag)
	 * 
	 * @return the tag base
	 */
	public int getTagBase() {
		return tagBase;
	}

	/** sets the tag base (number of values per tag)
	 * 
	 * @param tagBase the tag base
	 */
	public void setTagBase(int tagBase) {
		this.tagBase = tagBase;
	}

	/** returns the overall church participation rate
	 * 
	 * @return the church participation rate
	 */
	public double getChurchParticipationRate() {
		return churchParticipationRate;
	}

	/** sets the overall church participation rate
	 * 
	 * @param churchParticipationRate the church participation rate
	 */
	public void setChurchParticipationRate(double churchParticipationRate) {
		this.churchParticipationRate = churchParticipationRate;
	}

	/** returns the overall number of church denominations
	 * 
	 * @return the number of denominations
	 */
	public int getNumDenominations() {
		return numDenominations;
	}

	/** sets the overall number of church denominations
	 * 
	 * @param numDenominations the number of denominations
	 */
	public void setNumDenominations(int numDenominations) {
		this.numDenominations = numDenominations;
	}

	/** returns the upper bound for agents' tag evolution propensity
	 * 
	 * @return the maximal tag evolution propensity
	 */
	public double getMaxTagEvolutionPropensity() {
		return maxTagEvolutionPropensity;
	}

	/** sets the upper bound for agents' tag evolution propensity
	 * 
	 * @param maxTagEvolutionPropensity the maximal tag evolution propensity
	 */
	public void setMaxTagEvolutionPropensity(double maxTagEvolutionPropensity) {
		this.maxTagEvolutionPropensity = maxTagEvolutionPropensity;
	}

	/** returns the lower bound for agents' tag evolution propensity
	 * 
	 * @return the minimal tag evolution propensity
	 */
	public double getMinTagEvolutionPropensity() {
		return minTagEvolutionPropensity;
	}

	/** sets the lower bound for agents' tag evolution propensity
	 * 
	 * @param minTagEvolutionPropensity the minimal tag evolution propensity
	 */
	public void setMinTagEvolutionPropensity(double minTagEvolutionPropensity) {
		this.minTagEvolutionPropensity = minTagEvolutionPropensity;
	}
	
	/** returns if the output to file is set on or off
	 * 
	 * @return true if turned on, false if turned off
	 */
	public boolean getPrintToFile(){
		return this.printToFile;
	}
	
	/** turns the output to file on or off
	 * 
	 * @param printToFile true to turn it on, false to turn it off
	 */
	public void setPrintToFile(boolean printToFile){
		this.printToFile = printToFile;
	}

	/** returns the output path
	 * 
	 * @return the output path
	 */
	public String getOutputPath() {
		return outputPath;
	}

	/** sets the output path
	 * 
	 * @param outputPath output path 
	 */
	public void setOutputPath(String outputPath) {
		this.outputPath = outputPath;
	}
	
	/** returns if the output to console is set on or off
	 * 
	 * @return true if output to console is turned on, false otherwise
	 */
	public boolean isShowOutput() {
		return showOutput;
	}

	/** turns the output to console on or off
	 * 
	 * @param showOutput true to turn the output on, false to turn it off
	 */
	public void setShowOutput(boolean showOutput) {
		this.showOutput = showOutput;
	}
	
	/** returns the tick when a model run stops
	 * 
	 * @return stop time of the model run
	 */
	public int getStopTime() {
		return stopTime;
	}

	/** sets the tick when a model run stops
	 * 
	 * @param stopTime the stop time of a model run
	 */
	public void setStopTime(int stopTime) {
		this.stopTime = stopTime;
	}

	/** returns the mean inter-arrival time for in-migrating mine workers. Used
	 * for the arrival distribution in the miner source (mean inter-arrival time = 1 / lambda)
	 * 
	 * @return mean inter-arrival time (in ticks)
	 */
	public double getInterArrivalTime() {
		return interArrivalTime;
	}

	/** sets the mean inter-arrival time for in-migrating mine workers. Used
	 * for the arrival distribution (neg. exponential) in the miner source 
	 * (mean inter-arrival time = 1 / lambda)
	 * 
	 * @param interArrivalTime mean inter-arrival time (in ticks)
	 */
	public void setInterArrivalTime(double interArrivalTime) {
		this.interArrivalTime = interArrivalTime;
	}

	/** returns the mean number of in-migrant mine workers arriving at one time.
	 * Used for the number distribution (normal) in the miner source
	 * 
	 * @return mean number of in-migrants
	 */
	public int getMeanNumArrived() {
		return meanNumArrived;
	}

	/** sets the mean number of in-migrant mine workers arriving at one time.
	 * Used for the number distribution (normal) in the miner source
	 * 
	 * @param meanNumArrived mean number of in-migrants
	 */
	public void setMeanNumArrived(int meanNumArrived) {
		this.meanNumArrived = meanNumArrived;
	}

	/** returns the standard deviation of the number of in-migrant mine workers arriving at one time.
	 * Used for the number distribution (normal) in the miner source
	 * 
	 * @return standard deviation of the number of in-migrants
	 */
	public int getSdNumArrived() {
		return sdNumArrived;
	}

	/** sets the standard deviation of the number of in-migrant mine workers arriving at one time.
	 * Used for the number distribution (normal) in the miner source
	 * 
	 * @param sdNumArrived standard deviation of the number of in-migrants
	 */
	public void setSdNumArrived(int sdNumArrived) {
		this.sdNumArrived = sdNumArrived;
	}

	/** returns the lower limit for the maximum number of friends
	 * 
	 * @return lower limit of maximum number of friends
	 */
	public int getLowerMaxNumFriends() {
		return lowerMaxNumFriends;
	}

	/** sets the lower limit for the maximum number of friends
	 * 
	 * @param lowerMaxNumFriends lower limit of maximum number of friends
	 */
	public void setLowerMaxNumFriends(int lowerMaxNumFriends) {
		this.lowerMaxNumFriends = lowerMaxNumFriends;
	}

	/** returns the upper limit for the maximum number of friends
	 * 
	 * @return upper limit of maximum number of friends
	 */
	public int getUpperMaxNumFriends() {
		return upperMaxNumFriends;
	}

	/** sets the upper limit for the maximum number of friends
	 * 
	 * @param upperMaxNumFriends upper limit of maximum number of friends
	 */
	public void setUpperMaxNumFriends(int upperMaxNumFriends) {
		this.upperMaxNumFriends = upperMaxNumFriends;
	}

	/** returns the upper limit for the maximum number of concurrent sexual partners
	 * 
	 * @return upper limit of maximum number of partners
	 */
	public int getUpperMaxNumPartners() {
		return upperMaxNumPartners;
	}

	/** sets the upper limit for the maximum number of concurrent sexual partners
	 * 
	 * @param upperMaxNumPartners upper limit of maximum number of partners
	 */
	public void setUpperMaxNumPartners(int upperMaxNumPartners) {
		this.upperMaxNumPartners = upperMaxNumPartners;
	}
	
	/** returns the lower limit for the maximum number of concurrent sexual partners
	 * 
	 * @return lower limit of maximum number of partners
	 */
	public int getLowerMaxNumPartners() {
		return lowerMaxNumPartners;
	}

	/** sets the lower limit for the maximum number of concurrent sexual partners
	 * 
	 * @param lowerMaxNumPartners lower limit of maximum number of partners
	 */
	public void setLowerMaxNumPartners(int lowerMaxNumPartners) {
		this.lowerMaxNumPartners = lowerMaxNumPartners;
	}

	/** returns the maximum value for the endorsement base used in the computation of
	 * overall endorsement values. Each agent is assigned a random base between maximum
	 * and minimum value specified as model parameters.
	 * 
	 * @return maximum value for the endorsement base
	 */
	public double getMaxEndorsementBase() {
		return maxEndorsementBase;
	}

	/** sets the maximum value for the endorsement base used in the computation of
	 * overall endorsement values. Each agent is assigned a random base between maximum
	 * and minimum value specified as model parameters.
	 * 
	 * @param maxEndorsementBase maximum value for the endorsement base
	 */
	public void setMaxEndorsementBase(double maxEndorsementBase) {
		this.maxEndorsementBase = maxEndorsementBase;
	}

	/** returns the minimum value for the endorsement base used in the computation of
	 * overall endorsement values. Each agent is assigned a random base between maximum
	 * and minimum value specified as model parameters.
	 * 
	 * @return minimum value for the endorsement base
	 */
	public double getMinEndorsementBase() {
		return minEndorsementBase;
	}

	/** sets the minimum value for the endorsement base used in the computation of
	 * overall endorsement values. Each agent is assigned a random base between maximum
	 * and minimum value specified as model parameters.
	 * 
	 * @param minEndorsementBase minimum value for the endorsement base
	 */
	public void setMinEndorsementBase(double minEndorsementBase) {
		this.minEndorsementBase = minEndorsementBase;
	}

	/** returns the minimum number of endorsement classes. Each agent is assigned a random
	 * number between min and max. The number of endorsement classes defines the range of 
	 * values for the individual weights each agent assigns with the different endorsement
	 * tokens (with <code>1 <= weight_i <= number of classes</code>).
	 * 
	 * @return minimum number of endorsement classes
	 */
	public int getMinEndorsementClasses() {
		return minEndorsementClasses;
	}

	/** returns the maximum number of endorsement classes. Each agent is assigned a random
	 * number between min and max. The number of endorsement classes defines the range of 
	 * values for the individual weights each agent assigns with the different endorsement
	 * tokens (with <code>1 <= weight_i <= number of classes</code>).
	 * 
	 * @return maximum number of endorsement classes
	 */
	public int getMaxEndorsementClasses() {
		return maxEndorsementClasses;
	}

	/** sets the maximum number of endorsement classes. Each agent is assigned a random
	 * number between min and max. The number of endorsement classes defines the range of 
	 * values for the individual weights each agent assigns with the different endorsement
	 * tokens (with <code>1 <= weight_i <= number of classes</code>).
	 * 
	 * @param maxEndorsementClasses maximum number of endorsement classes
	 */
	public void setMaxEndorsementClasses(int maxEndorsementClasses) {
		this.maxEndorsementClasses = maxEndorsementClasses;
	}


	/** sets the minimum number of endorsement classes. Each agent is assigned a random
	 * number between min and max. The number of endorsement classes defines the range of 
	 * values for the individual weights each agent assigns with the different endorsement
	 * tokens (with <code>1 <= weight_i <= number of classes</code>).
	 * 
	 * @param minEndorsementClasses minimum number of endorsement classes
	 */
	public void setMinEndorsementClasses(int minEndorsementClasses) {
		this.minEndorsementClasses = minEndorsementClasses;
	}

	/** returns the standard monthly wage (for piece jobs in the village).
	 * 
	 * @return standard monthly wage
	 */
	public double getStandardWage() {
		return standardWage;
	}

	/** sets the standard monthly wage (for piece jobs in the village).
	 * 
	 * @param standardWage standard monthly wage
	 */
	public void setStandardWage(double standardWage) {
		this.standardWage = standardWage;
	}

	/** returns the monthly child grant for children up to age 7
	 * 
	 * @return child grant
	 */
	public double getChildGrant() {
		return childGrant;
	}

	/** sets the monthly child grant for children up to age 7
	 * 
	 * @param childGrant child grant
	 */
	public void setChildGrant(double childGrant) {
		this.childGrant = childGrant;
	}

	/** returns the monthly state pension for seniors.
	 * 
	 * @return monthly state pension
	 */
	public double getStatePension() {
		return statePension;
	}

	/** sets the monthly state pension for seniors.
	 * 
	 * @param statePension monthly state pension
	 */
	public void setStatePension(double statePension) {
		this.statePension = statePension;
	}

	/** returns the monthly food cost for a child <= 7 years old
	 * 
	 * @return monthly food cost for a child
	 */
	public double getFoodCostChild() {
		return foodCostChild;
	}

	/** sets the monthly food cost for a child <= 7 years old
	 * 
	 * @param foodCostChild monthly food cost for a child
	 */
	public void setFoodCostChild(double foodCostChild) {
		this.foodCostChild = foodCostChild;
	}

	/** returns the monthly food cost for a female adult
	 * 
	 * @return monthly food cost for a female
	 */
	public double getFoodCostFemaleAdult() {
		return foodCostFemaleAdult;
	}

	/** sets the monthly food cost for a female adult
	 * 
	 * @param foodCostFemaleAdult monthly food cost for a female 
	 */ 
	public void setFoodCostFemaleAdult(double foodCostFemaleAdult) {
		this.foodCostFemaleAdult = foodCostFemaleAdult;
	}

	/** returns the monthly food cost for a male adult
	 * 
	 * @return monthly food cost for a male
	 */
	public double getFoodCostMaleAdult() {
		return foodCostMaleAdult;
	}

	/** sets the monthly food cost for a male adult
	 * 
	 * @param foodCostMaleAdult monthly food cost for a male 
	 */ 
	public void setFoodCostMaleAdult(double foodCostMaleAdult) {
		this.foodCostMaleAdult = foodCostMaleAdult;
	}

	/** returns the similarity age range
	 * 
	 * @return similarity age range
	 */
	public double getSimilarityAgeRange() {
		return similarityAgeRange;
	}

	/** sets the similarity age range
	 * 
	 * @param similarityAgeRange similarity age range
	 */
	public void setSimilarityAgeRange(double similarityAgeRange) {
		this.similarityAgeRange = similarityAgeRange;
	}

	/** returns the average cost of a burial
	 * 
	 * @return burial cost
	 */
	public double getBurialCost() {
		return burialCost;
	}

	/** sets the average cost of a burial
	 * 
	 * @param burialCost the burial cost
	 */
	public void setBurialCost(double burialCost) {
		this.burialCost = burialCost;
	}

	/** returns the probability to encounter a random sexual partner
	 * 
	 * @return random sexual partner probability
	 */
	public double getRandomPartnerProbability() {
		return randomPartnerProbability;
	}

	/** sets the probability to encounter a random sexual partner
	 * 
	 * @param randomPartnerProbability random sexual partner probability
	 */
	public void setRandomPartnerProbability(double randomPartnerProbability) {
		this.randomPartnerProbability = randomPartnerProbability;
	}

	/** returns the mean for the age distribution of in-migrants.
	 * 
	 * @return mean age of in-migrants
	 */
	public int getMeanAgeArrived() {
		return meanAgeArrived;
	}

	/** sets the mean for the age distribution of in-migrants.
	 * 
	 * @param meanAgeArrived mean age of in-migrants
	 */
	public void setMeanAgeArrived(int meanAgeArrived) {
		this.meanAgeArrived = meanAgeArrived;
	}

	/** returns the standard deviation for the age distribution of in-migrants.
	 * 
	 * @return standard deviation of age of in-migrants
	 */
	public int getSdAgeArrived() {
		return sdAgeArrived;
	}

	/** sets the standard deviation for the age distribution of in-migrants.
	 * 
	 * @param sdAgeArrived standard deviation of age of in-migrants
	 */
	public void setSdAgeArrived(int sdAgeArrived) {
		this.sdAgeArrived = sdAgeArrived;
	}

	/** returns the proportion of households with a shack to rent out if necessary
	 * 
	 * @return proportion of shacks
	 */
	public double getShackProportion() {
		return shackProportion;
	}

	/** sets the proportion of households with a shack to rent out if necessary
	 * 
	 * @param shackProportion proportion of shacks in the village
	 */
	public void setShackProportion(double shackProportion) {
		this.shackProportion = shackProportion;
	}

	/** returns the lower bound for remittances from migrants to their household
	 * in the village
	 * 
	 * @return lower bound for remittances
	 */
	public int getMinRemittance() {
		return minRemittance;
	}

	/** sets the lower bound for remittances from migrants to their household
	 * in the village
	 * 
	 * @param minRemittance lower bound for remittances
	 */
	public void setMinRemittance(int minRemittance) {
		this.minRemittance = minRemittance;
	}

	/** returns the upper bound for remittances from migrants to their household
	 * in the village
	 * 
	 * @return upper bound for remittances
	 */
	public int getMaxRemittance() {
		return maxRemittance;
	}

	/** sets the upper bound for remittances from migrants to their household
	 * in the village
	 * 
	 * @param maxRemittance upper bound for remittances
	 */
	public void setMaxRemittance(int maxRemittance) {
		this.maxRemittance = maxRemittance;
	}

	/** returns the probability for sending remittance home for any migrant in any month.
	 * 
	 * @return probability for remittance 
	 */
	public double getRemittanceProbability() {
		return remittanceProbability;
	}

	/** sets the probability for sending remittance home for any migrant in any month.
	 * 
	 * @param remittanceProbability probability for remittance
	 */
	public void setRemittanceProbability(double remittanceProbability) {
		this.remittanceProbability = remittanceProbability;
	}

	/** returns the yearly college/university fee.
	 * 
	 * @return college fee
	 */
	public double getCollegeFee() {
		return collegeFee;
	}

	/** sets the yearly college/university fee.
	 * 
	 * @param collegeFee college fee
	 */
	public void setCollegeFee(double collegeFee) {
		this.collegeFee = collegeFee;
	}

	/** returns the yearly school fee for non-compulsory secondary school (pupils > 15 years).
	 * 
	 * @return school fee
	 */
	public double getSchoolFee() {
		return schoolFee;
	}

	/** sets the yearly school fee for non-compulsory secondary school (pupils > 15 years).
	 * 
	 * @param schoolFee school fee
	 */
	public void setSchoolFee(double schoolFee) {
		this.schoolFee = schoolFee;
	}

	/** returns the upper bound for the lobola a groom has to pay the bride's household.
	 * 
	 * @return upper bound for lobola
	 */
	public int getMaxLobola() {
		return maxLobola;
	}

	/** sets the upper bound for the lobola a groom has to pay the bride's household.
	 * 
	 * @param maxLobola upper bound for lobola
	 */
	public void setMaxLobola(int maxLobola) {
		this.maxLobola = maxLobola;
	}

	/** returns the lower bound for the lobola a groom has to pay the bride's household.
	 * 
	 * @return lower bound for lobola
	 */
	public int getMinLobola() {
		return minLobola;
	}

	/** sets the lower bound for the lobola a groom has to pay the bride's household.
	 * 
	 * @param minLobola lower bound for lobola
	 */
	public void setMinLobola(int minLobola) {
		this.minLobola = minLobola;
	}

	/** returns the proportion of female seniors over the age limit who receive state pensions.
	 * 
	 *  @return proportion of female seniors with pensions
	 */
	public double getPensionProportionFemale() {
		return pensionProportionFemale;
	}

	/** sets the proportion of female seniors over the age limit who receive state pensions.
	 * 
	 * @param pensionProportionFemale proportion of female seniors with pensions
	 */
	public void setPensionProportionFemale(double pensionProportionFemale) {
		this.pensionProportionFemale = pensionProportionFemale;
	}

	/** returns the proportion of male seniors over the age limit who receive state pensions.
	 * 
	 * @return proportion of male seniors with pensions
	 */
	public double getPensionProportionMale() {
		return pensionProportionMale;
	}

	/** sets the proportion of male seniors over the age limit who receive state pensions.
	 * 
	 * @param pensionProportionMale proportion of male seniors with pensions
	 */
	public void setPensionProportionMale(double pensionProportionMale) {
		this.pensionProportionMale = pensionProportionMale;
	}

	/** returns the proportion of children under the age limit who receive a child grant.
	 * 
	 * @return proportion of children with child grants
	 */
	public double getChildGrantProportion() {
		return childGrantProportion;
	}

	/** sets the proportion of children under the age limit who receive a child grant.
	 * 
	 * @param childGrantProportion proportion of children with child grants
	 */
	public void setChildGrantProportion(double childGrantProportion) {
		this.childGrantProportion = childGrantProportion;
	}

	/** returns the age limit for child grants. Children under the age of this limit may receive
	 * a child grant from the state.
	 * 
	 * @return age limit for child grants
	 */
	public int getChildGrantAgeLimit() {
		return childGrantAgeLimit;
	}

	/** sets the age limit for child grants. Children under the age of this limit may receive
	 * a child grant from the state.
	 * 
	 * @param childGrantAgeLimit age limit for child grants
	 */
	public void setChildGrantAgeLimit(int childGrantAgeLimit) {
		this.childGrantAgeLimit = childGrantAgeLimit;
	}

	/** returns the age limit for state pensions for female seniors. Female seniors over the age
	 * of this limit may receive a state pension.
	 * 
	 * @return female age limit for state pensions
	 */
	public int getPensionAgeLimitFemale() {
		return pensionAgeLimitFemale;
	}

	/** sets the age limit for state pensions for female seniors. Female seniors over the age
	 * of this limit may receive a state pension.
	 * 
	 * @param pensionFemaleAgeLimit female age limit for state pensions
	 */
	public void setPensionAgeLimitFemale(int pensionFemaleAgeLimit) {
		this.pensionAgeLimitFemale = pensionFemaleAgeLimit;
	}

	/** returns the age limit for state pensions for male seniors. Male seniors over the age
	 * of this limit may receive a state pension.
	 * 
	 * @return male age limit for state pensions
	 */
	public int getPensionAgeLimitMale() {
		return pensionAgeLimitMale;
	}

	/** sets the age limit for state pensions for male seniors. Male seniors over the age
	 * of this limit may receive a state pension.
	 * 
	 * @param pensionMaleAgeLimit male age limit for state pensions
	 */
	public void setPensionAgeLimitMale(int pensionMaleAgeLimit) {
		this.pensionAgeLimitMale = pensionMaleAgeLimit;
	}

	/** returns the rate for mother-to-child HIV transmission.
	 * 
	 * @return mother-to-child HIV transmission rate
	 */
	public double getMotherChildHIVTransmissionRate() {
		return motherChildHIVTransmissionRate;
	}

	/** sets the rate for mother-to-child HIV transmission.
	 * 
	 * @param motherChildHIVTransmissionRate mother-to-child HIV transmission rate
	 */
	public void setMotherChildHIVTransmissionRate(
			double motherChildHIVTransmissionRate) {
		this.motherChildHIVTransmissionRate = motherChildHIVTransmissionRate;
	}

	/** returns the average cost to build a new house in the village.
	 * 
	 * @return average house cost
	 */
	public int getHouseCost() {
		return houseCost;
	}

	/** sets the average cost to build a new house in the village.
	 * 
	 * @param houseCost house cost
	 */
	public void setHouseCost(int houseCost) {
		this.houseCost = houseCost;
	}

	/** returns the upper limit for monthly rent of a shack in the village.
	 * 
	 * @return upper limit of the shack rent
	 */
	public int getMaxShackRent() {
		return maxShackRent;
	}

	/** sets the upper limit for monthly rent of a shack in the village.
	 * 
	 * @param maxShackRent upper limit of the shack rent
	 */
	public void setMaxShackRent(int maxShackRent) {
		this.maxShackRent = maxShackRent;
	}

	/** returns the lower limit for monthly rent of a shack in the village.
	 * 
	 * @return lower limit of the shack rent
	 */
	public int getMinShackRent() {
		return minShackRent;
	}

	/** sets the lower limit for monthly rent of a shack in the village.
	 * 
	 * @param minShackRent lower limit of the shack rent
	 */
	public void setMinShackRent(int minShackRent) {
		this.minShackRent = minShackRent;
	}

	
}
