package org.cfpm.caves.za;

import java.util.ArrayList;
import java.util.HashMap;

import uchicago.src.sim.util.SimUtilities;
//import uchicago.src.sim.network.DefaultEdge;
import java.beans.*;

import jess.JessException;
import jess.Rete;
import jess.Value;
import jess.WatchConstants;

/** A stokvel is a savings club with a monthly fee and either a monthly rotation of payouts (ROSCA) or 
 *  a payout of accumulated savings at the end of a circle (ASCA). For the sake of simplicity, this class
 *  implements an ASCA-type stokvel. Thus it's easier to have grace periods like in burial societies. 
  * 
 * @author Ruth
 *
 */
public class Stokvel extends SavingsClub {
	
	//--- stokvel attributes --------------------------------

	/** list of start dates for all members */
	ArrayList<Integer> startDates;
	/** next date (tick) of pay out (depends on number of members) */
	int payOutDate;
		
	
	//--- constructors ----------------------------------------------
	
	/** Creates a new stokvel for the given members
	 * @param members list of members
	 * @param model the model
	 */
	public Stokvel(ArrayList<Person> members, SouthAfricaModel model) {
		super("stokvel", 0, members, model);
		// set monthly fee (draw random sample from uniform distribution (30-70 Rand)
		this.monthlyFee = model.randomInt(30,70);
		// set date for payout (a cycle lasts number of members months, payout is at the end of a month)
		this.payOutDate = model.tick + (3 - model.week) + (4 * members.size());
		// put it into Jess (needs to be here as Jess can't handle inheritance with shadow facts)
		try {
			model.engine.watch(WatchConstants.FACTS);
			model.engine.definstance("stokvel", this, true);
			model.engine.unwatch(WatchConstants.FACTS);
		} catch (JessException je) {
			System.err.println("An error occurred at line "	
					+ je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}
		// record it for model output
		model.stokvelRecorder.add(this);
	}
	

	//--- methods -------------------------------------------------------
	
	/** lets the given person join this stokvel
	 * @param p the new member
	 */
	public void addMember(Person p){
		super.addMember(p);
		if (startDates == null) startDates = new ArrayList<Integer>();
		startDates.add(model.tick);
		// add stokvel to member's societies
		p.joinStokvel(this);
	}
	
	/** lets the given person leave this stokvel
	 * @param p the leaving member
	 */
	public void removeMember(Person p){
		// remove stokvel from members's societies
		p.leaveStokvel(this);
		int index = this.members.indexOf(p);
		super.removeMember(p);
		startDates.remove(index);
		// what if people leave during a cycle? 
		// they should get their savings back (Found this in http://datafirst.cssr.uct.ac.za/resource/papers/keswell_0358.pdf)
	}
	
	/** pay out accumulated savings to the members
	 * 
	 * @return list of each member's share of the current holdings
	 */
	public double[] payOut(){
		// pay out accumulated savings
		double[] amounts = new double[members.size()];
		// compute share of holdings for every member
		double sum = 0;
		for (int i = 0; i < amounts.length; i++){
			amounts[i] = (model.tick - startDates.get(i)) / 4.0;	// number of months paid in
			sum += amounts[i];
		}
		for (int i = 0; i < amounts.length; i++){
			amounts[i] = (amounts[i] / sum) * holdings;
		}
		reset();
		return amounts;
	}
	
	/** internal function: resets all variables for the start of a new cycle */
	private void reset(){
		double old = holdings;
		holdings = 0;
		pcs.firePropertyChange("holdings", old, holdings);
		int[] oldStartDates = getStartDates();
		this.startDates = new ArrayList<Integer>(members.size());
		for (int i = 0; i < members.size(); i++){
			startDates.add(model.tick);
		}
		pcs.firePropertyChange("startDates", oldStartDates, getStartDates());
		setPayOutDate(model.tick + (3 - model.week) + (4 * members.size()));
	}
	
	/** returns the next payout date
	 * 
	 * @return date of next payout
	 */
	public int getPayOutDate(){
		return this.payOutDate;
	}
	
	/** sets the next payout date
	 * 
	 * @param date date of next payout
	 */
	public void setPayOutDate(int date){
		int old = this.payOutDate;
		this.payOutDate = date;
		pcs.firePropertyChange("payOutDate", old, this.payOutDate);
	}
	
	/** returns the start dates of all members 
	 * 
	 * @return list of dates (ticks) when membership started
	 */
	public int[] getStartDates(){
		int[] dates = new int[this.startDates.size()];
		for (int i = 0; i < dates.length; i++){
			dates[i] = this.startDates.get(i);
		}
		return dates;
	}
	
	/** disbands this stokvel */
	public void disband(){
		// last payout is done from Jess before calling this method
		// just remove stokvel from each member's list of stokvels
		for (Person p : members){
			p.leaveStokvel(this);
		}
		// clear all variables so this object can be garbage collected
		holdings = 0;
		startDates = null;
		monthlyFee = 0;
		payOutDate = 0;
		members = null;
		memberIDs = null;
		onGrace = null;
		name = null;
		model = null;
	}


}
