package org.cfpm.caves.za;

import jess.*;

/** This class represents a village in the model. It maintains its area in space
 *  so that new households can be placed within the boundaries of the village.
 *  Villages are placed randomly in space but don't overlap.
 * 
 * @author Ruth
 *
 */
public class Village {
	
	//--- constants -----------------------------------------------------
	/** buffer around villages to ensure they don't overlap */
	static final int buffer = 5;
	
	//--- village attributes --------------------------------------------
	
	/** name of the village */
	Value name;
	/** bounding box of the village area on the 2D grid */
	int[] extension = new int[4];
	
	//--- constructors --------------------------------------------------
	
	/** Creates a new village of the given name. Its area is specified by centre and radius.
	 *  @param nameString name of the village
	 *  @param centre x,y coordinates of the centre
	 *  @param radius the radius as number of cells 
	 */
	public Village(String nameString, int[] centre, int radius) {
		setExtension(new int[]{centre[0] - radius, centre[1] - radius, centre[0] + radius, centre[1] + radius});
		setName(nameString);
	}
	
	/** Creates a new village of the given name. Its area is specified by the bounding box.
	 *  @param nameString name of the village
	 *  @param x0 x coordinate of the "upper left" corner
	 *  @param y0 y coordinate of the "upper left" corner
	 *  @param x1 x coordinate of the "lower right" corner
	 *  @param y1 y coordinate of the "lower right" corner
	 */
	public Village(String nameString, int x0, int y0, int x1, int y1){
		setExtension(new int[]{x0, y0, x1, y1});
		setName(nameString);
	}
	
	//--- methods -------------------------------------------------------
	
	/** sets the village's spatial extension (bounding box)
	 * @param extension the bounding box as x,y coordinates of the upper left and lower right corners
	 */
	public void setExtension(int[] extension){
		this.extension = extension;
	}
	
	/** sets the village's name
	 * 
	 * @param nameString name of the village
	 */
	public void setName(String nameString){
		try {
			name = new Value(nameString, RU.SYMBOL);
		}
		catch(JessException je){
			System.err.println("An error occurred at line " + je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}
	}
	
	/** returns the name of the village
	 * 
	 * @return the village's name
	 */
	public Value getName(){
		return this.name;
	}
	
	/** returns the village's spatial extension
	 * 
	 * @return bounding box as x,y coordinates of the upper left and lower right corners
	 */
	public int[] getExtension(){
		return this.extension;
	}
	
	/** checks if the specified location is covered by the village
	 * 
	 * @param x x-coordinate
	 * @param y y-coordinate
	 * @return true if inside the village, false otherwise
	 */
	public boolean covers(int x, int y){
		if (x >= extension[0] && x <= extension[2] &&
			 y >= extension[1] && y <= extension[3])
		{
			return true;
		}
		return false;
	}
	
	/** checks if the specified location falls within the village area plus buffer.
	 * 
	 * @param x x-coordinate
	 * @param y y-coordinate
	 * @return true if inside the buffered village, false otherwise
	 */
	public boolean coversWithBuffer(int x, int y){
		if (x >= (extension[0]-buffer) && x <= (extension[2]+buffer) &&
			 y >= (extension[1]-buffer) && y <= (extension[3]+buffer))
		{
			return true;
		}
		return false;
		
	}

}
