package org.cfpm.caves.za.util;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.io.*;
import java.util.StringTokenizer;

import uchicago.src.sim.network.*;
import org.cfpm.caves.za.*;

/**  Utility functions for networks.
 * 
 * @author Ruth
 *
 */
public class NetworkUtilities {
	
	 /** 
	 * Calculates the clustering coefficient (avg. density of "ego networks")
	 * in graph for links of the specified type.
	 * 
	 * Adapted from Repast: NetUtilities and changed to be able to work with a multi layer network
	 * expressed as direct references to Person objects.
	 * @param nodes the network for which the ClustCoef will be calculated
	 * @param linkType the type of link determining the network layer (friend, stokvel or burial)
	 */
	public static double calcClustCoef(ArrayList<Person> nodes, String linkType) {
		
		// construct matrix for network layer
		HashMap<String, ArrayList<String>> network = new HashMap<String, ArrayList<String>>();
		for (Person p : nodes){
			ArrayList<Person> linked = p.retrieveLinkedPersons(linkType);
			ArrayList<String> lNames = new ArrayList<String>(linked.size());
			for (Person lPerson : linked){
				String name = lPerson.getName().toString();
				if (!lNames.contains(name)) lNames.add(name);	// make sure every linked node is only listed once
			}
			network.put(p.getName().toString(), lNames);
		}
		
		double clustCoef = 0.0;
		//loop over all nodes and compute c_i
		for (String iNode : network.keySet()){
			int iDens = 0;
			ArrayList<String> jNodes = network.get(iNode);
			int iDegree = jNodes.size();
			for (String jNode : jNodes){
				ArrayList<String> kNodes = network.get(jNode);
				try{
				for (String kNode : kNodes){
					if (jNodes.contains(kNode) && !jNode.equals(kNode)){	// make sure not a loop
						iDens++;
					}
				}
				}catch (NullPointerException e){
					System.out.println("NullPointerException in calcClusterCoefficient");
					System.out.println("iNode = " + iNode + ", jNode = " + jNode + ", kNodes = " + kNodes);
					e.printStackTrace();
				}
			}
			//trap condition of no arcs (density  = 0)
			if (iDens > 0) {
				//cluster coef of i = (actual #)/(max possible number of edges in ngh)
				clustCoef += (double) iDens
						/ (double) (iDegree * (iDegree - 1));
			}
		}
		clustCoef = clustCoef / nodes.size();
		return clustCoef;
	}
	
	
	
	/** 
	 * Calculates the clustering coefficient (avg. density of "ego networks")
	 * in graph for links of the specified type.
	 * 
	 * Taken from Repast: NetUtilities and changed to be able to work with a multi layer network.
	 * @param nodes the network for which the ClustCoef will be calculated
	 * @param linkType the type of link determining the network layer
	 */
	public static double calcClustCoef(List nodes, String linkType) {
		//will directionality be a problem?  What if k -> i but not i -> k?
		//possibility of multiple edges to one node?
		//IGNORE arcs to i

		//loop over all nodes
		int nNodes = nodes.size();
		double clustCoef = 0.0;
		HashSet jNodes = new HashSet(); //set of all nodes (j) connected to i
		for (int i = 0; i < nNodes; i++) {
			jNodes.clear();
			int iDens = 0;
			Node iNode = (Node) nodes.get(i);
			ArrayList edges = iNode.getOutEdges();
			int iDegree = edges.size();
			for (int j = 0; j < edges.size(); j++) {
				// only consider edges of specified type
				if (((Edge) edges.get(j)).getType().equalsIgnoreCase(linkType)) {
					jNodes.add((Node) ((Edge) edges.get(j)).getTo());
				}
			}
			if (jNodes.contains(iNode)) {
				jNodes.remove(iNode);
			}
			Iterator jIter = jNodes.iterator();
			while (jIter.hasNext()) {
				Node jNode = (Node) jIter.next();
				ArrayList jEdges = jNode.getOutEdges();
				for (int k = 0; k < jEdges.size(); k++) {
					// only consider edges of specified type
					if (((Edge) jEdges.get(k)).getType().equalsIgnoreCase(linkType)){
						Node kNode = ((Edge) jEdges.get(k)).getTo();
						if (jNodes.contains(kNode) && !jNode.equals(kNode)) //make sure not a loop
						{
							iDens++;
						}
					}
				}
			}
			//trap condition of no arcs (density  = 0)
			if (iDens > 0) {
				//cluster coef of i = (actual #)/(max possible number of edges in ngh)
				clustCoef += (double) iDens
						/ (double) (iDegree * (iDegree - 1));
			}
		}
		clustCoef = clustCoef / nNodes;
		return clustCoef;
	}
	
//	public static void extractLinkStats(String traceFile, String outputFile, String path) throws IOException {
//		BufferedReader trace = new BufferedReader(new FileReader(path + traceFile));
//        BufferedWriter output = new BufferedWriter(new FileWriter(path + outputFile));
//		output.write("tick\treciprocal\tone-directional\tproportion\tkin\tnon-kin\tproportion\n");
//		while (trace.ready()){
//			StringBuffer outLine = new StringBuffer();
//			String line = trace.readLine();
//			// find beginning of tick
//			while (!line.startsWith("***") && trace.ready()){
//				line = trace.readLine();
//			}
//			System.out.println(line);
//			// extract tick from line "*** TICK ## ***": it's the third element
//			StringTokenizer tok = new StringTokenizer(line, " ");
//			tok.nextElement();
//			tok.nextElement();
//			outLine.append(tok.nextElement() + "\t");
//			// find start of reciprocal/one-directional links 
//			line = trace.readLine();
//			while(!line.startsWith("There are") && trace.ready()){
//				line = trace.readLine();
//			}
//			System.out.println(line);
//			// extract reciprocal (third element) and one-directional (seventh element) values
//			tok = new StringTokenizer(line, " ");
//			tok.nextElement(); tok.nextElement();
//			outLine.append(tok.nextElement() + "\t");
//			tok.nextElement(); tok.nextElement(); tok.nextElement();
//			outLine.append(tok.nextElement() + "\t");
//			// extract proportion (second element)
//			line = trace.readLine();
//			System.out.println(line);
//			tok = new StringTokenizer(line, " ");
//			tok.nextElement();
//			outLine.append(tok.nextElement() + "\t");
//			// extract kin (3rd element) and non-kin (9th element) values
//			line = trace.readLine();
//			System.out.println(line);
//			tok = new StringTokenizer(line, " ");
//			tok.nextElement(); tok.nextElement();
//			outLine.append(tok.nextElement() + "\t");
//			int i = 3;
//			while (tok.hasMoreElements() && i < 8) {
//				i++;
//				tok.nextElement();  
//			}
//			if (tok.hasMoreElements()) outLine.append(tok.nextElement() + "\t");
//			// extract proportion (second element)
//			if (trace.ready()) {
//				line = trace.readLine();
//				System.out.println(line);
//				tok = new StringTokenizer(line, " ");
//				tok.nextElement();
//				outLine.append(tok.nextElement());
//			}
//			// write out
//			output.write(outLine.toString() + "\n");
//			output.flush();
//		}
//		output.flush();
//		output.close();
//		trace.close();
//	}
//	
//	public static void main(String[] args) throws IOException {
//		NetworkUtilities.extractLinkStats("trace.txt", "friendship-network-stats.txt", 
//				"/data/experiments/southAfricaModel/run-1191534831932/");
//	}

}
